@@ -1,9 +1,32 @@
-/*  $Id: init.pl,v 1.48 1998/07/06 10:27:37 jan Exp $
+/*  $Id: init.pl,v 1.88 2002/02/01 16:49:10 jan Exp $
 
-    Copyright (c) 1990 Jan Wielemaker. All rights reserved.
-    jan@swi.psy.uva.nl
+    Part of SWI-Prolog
 
-    Purpose: Get the Ball Rolling ...
+    Author:        Jan Wielemaker
+    E-mail:        jan@swi.psy.uva.nl
+    WWW:           http://www.swi-prolog.org
+    Copyright (C): 1985-2002, University of Amsterdam
+
+    This program is free software; you can redistribute it and/or
+    modify it under the terms of the GNU General Public License
+    as published by the Free Software Foundation; either version 2
+    of the License, or (at your option) any later version.
+
+    This program is distributed in the hope that it will be useful,
+    but WITHOUT ANY WARRANTY; without even the implied warranty of
+    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+    GNU General Public License for more details.
+
+    You should have received a copy of the GNU Lesser General Public
+    License along with this library; if not, write to the Free Software
+    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+
+    As a special exception, if you link this library with other files,
+    compiled with a Free Software compiler, to produce an executable, this
+    library does not by itself cause the resulting executable to be covered
+    by the GNU General Public License. This exception does not however
+    invalidate any other reasons why the executable file might be covered by
+    the GNU General Public License.
 */
 
 /*
@@ -19,9 +42,12 @@ If you want  to  debug  this  module,  put  a  '$:-'  trace.   directive
 somewhere.   The  tracer will work properly under boot compilation as it
 will use the C defined write predicate  to  print  goals  and  does  not
 attempt to call the Prolog defined trace interceptor.
+
+Please note that of version  3.3,  $:-   is  no  longer  an operator and
+therefore you have to use braces.
 */
 
-'$:-' format('Loading boot file ...~n', []).
+'$:-'(format('Loading boot file ...~n', [])).
 
 		/********************************
 		*    LOAD INTO MODULE SYSTEM	*
@@ -33,13 +59,6 @@ attempt to call the Prolog defined trace interceptor.
 		*          DIRECTIVES           *
 		*********************************/
 
-op(_, _, []) :- !.
-op(Priority, Type, [Name|Rest]) :- !,
-	$op(Priority, Type, Name),
-	op(Priority, Type, Rest).
-op(Priority, Type, Name) :-
-	$op(Priority, Type, Name).
-
 dynamic((Spec, More)) :- !,
 	dynamic(Spec),
 	dynamic(More).
@@ -120,6 +139,8 @@ $show_childs(Name, Arity) :-
 	ignore/1,
 	block/3,
 	catch/3,
+	call_cleanup/2,
+	call_cleanup/3,
 	apply/2.
 
 %   ->/2, ;/2, |/2 and \+/1 are normally compiled. These predicate catch them
@@ -200,14 +221,25 @@ fail(Label) :-
 %	throw(+Exception)
 %
 %	ISO compliant exception handling.  $throw/1 is compiled to
-%	rhe virtual instruction B_THROW.  See pl-wam.c for details.
+%	the virtual instruction B_THROW.  See pl-wam.c for details.
 
-catch(Goal, _Catcher, _Recover) :-
-	Goal.
+catch(_Goal, _Catcher, _Recover) :-
+	$catch.
 
 throw(Exception) :-
 	$throw(Exception).
 
+%	call_cleanup(:Goal, +Catcher, :Cleanup)
+%
+%	Call Cleanup as Goal finished (deterministic success, failure,
+%	exception or cut).  $call_cleanup translated to I_CALLCLEANUP.
+
+call_cleanup(Goal, Cleanup) :-
+	call_cleanup(Goal, _Catcher, Cleanup).
+
+call_cleanup(_Goal, _Catcher, _Cleanup) :-
+	$call_cleanup.
+
 :-
 	$hide((';'), 2),
 	$hide(('|'), 2),
@@ -256,28 +288,29 @@ $prefix_module(Module, _, Head, Module:Head).
 %	to give a DWIM warning. Otherwise fail. C will print an error
 %	message.
 
-:- flag($verbose_autoload, _, off).
-:- flag($enable_autoload, _, on).
+:- set_prolog_flag(autoload, true).
+:- set_prolog_flag(verbose_autoload, false).
 :- flag($autoloading, _, 0).
 
 $undefined_procedure(Module, Name, Arity, Action) :-
 	$prefix_module(Module, user, Name/Arity, Pred),
 	user:exception(undefined_predicate, Pred, Action), !.
 $undefined_procedure(Module, Name, Arity, retry) :-
-	flag($enable_autoload, on, on),
+	current_prolog_flag(autoload, true),
 	$find_library(Module, Name, Arity, LoadModule, Library),
 	functor(Head, Name, Arity),
 	flag($autoloading, Old, Old+1),
+	print_message(silent, autoload(Module:Name/Arity, Library)),
 	(   Module == LoadModule
-	->  ignore(ensure_loaded(Library))
+	->  ensure_loaded(Module:Library)
 	;   (   $c_current_predicate(_, LoadModule:Head)
 	    ->	Module:import(LoadModule:Head)
-	    ;	ignore(Module:use_module(Library, [Name/Arity]))
+	    ;	use_module(Module:Library, [Name/Arity])
 	    )
 	),
 	flag($autoloading, _, Old),
 	$c_current_predicate(_, Module:Head).
-$undefined_procedure(_, _, _, fail).
+$undefined_procedure(_, _, _, error).
 
 $calleventhook(Term) :-
 	(   notrace(user:prolog_event_hook(Term))
@@ -287,110 +320,45 @@ $calleventhook(Term) :-
 
 :- $hide($calleventhook, 1).
 
+%	 handle debugger 'w', 'p' and <N> depth options.
+
+$set_debugger_print_options(write) :- !,
+	set_prolog_flag(debugger_print_options,
+			[quoted(true)]).
+$set_debugger_print_options(print) :- !,
+	set_prolog_flag(debugger_print_options,
+			[quoted(true), portray(true), max_depth(10)]).
+$set_debugger_print_options(Depth) :-
+	current_prolog_flag(debugger_print_options, Options0),
+	delete(Options0, max_depth(_), Options),
+	set_prolog_flag(debugger_print_options,
+			[max_depth(Depth)|Options]).
+
 
 		/********************************
 		*        SYSTEM MESSAGES        *
 		*********************************/
 
-%	$ttyformat(+Format, [+ArgList])
-%	Format on the user stream.  Used to print system messages.
-
-$ttyformat(Format) :-
-	$ttyformat(Format, []).
-$ttyformat(Format, Args) :-
-	format(user_error, Format, Args).
-
-%	$confirm(Format, Args)
+%	$confirm(Spec)
 %
-%	Ask the user to confirm a question.
+%	Ask the user to confirm a question.  Spec is a term as used for
+%	print_message/2.
 
-$confirm(Format, Args) :-
-	$ttyformat(Format, Args),
-	$ttyformat('? '),
+$confirm(Spec) :-
+	print_message(query, Spec),
 	between(0, 5, _),
-	    (   get_single_char(Answer),
-		memberchk(Answer, [0'y, 0'Y, 0'j, 0'J, 0'n, 0'N, 0' ,10])
-	    ->  !, $confirm_(Answer)
-	    ;   $ttyformat('Please answer ''y'' or ''n''~n'),
+	    get_single_char(Answer),
+	    (	memberchk(Answer, "yYjJ \n")
+	    ->	!,
+	        print_message(query, if_tty(yes))
+	    ;	memberchk(Answer, "nN")
+	    ->	!,
+	        print_message(query, if_tty(no)),
+		fail
+	    ;	print_message(help, query(confirm)),
 		fail
 	    ).
-
-$confirm_(Answer) :-
-	memberchk(Answer, [0'y, 0'Y, 0'j, 0'J, 0' ,10]), !,
-	(   $tty
-	->  $ttyformat('yes~n')
-	;   true
-	).
-$confirm_(_) :-
-	$tty,
-	$ttyformat('no~n'),
-	fail.
-
-%	$warning(+Format, [+ArgList])
-%	Format a standard warning to the user and start the tracer.
-
-$warning(Format) :-
-	$warning(Format, []).
-$warning(Format, Args) :-
-	source_location(File, Line), !,
-	(   feature(report_error, true)
-	->  sformat(Msg, Format, Args),
-	    (   user:exception(warning, warning(File, Line, Msg), _)
-	    ->  true
-	    ;   format(user_error, '[WARNING: (~w:~d)~n~t~8|~w]~n',
-		       [File, Line, Msg])
-	    )
-	;   true
-	).
-$warning(Format, Args) :-
-	(   feature(report_error, true)
-	->  format(user_error, '[WARNING: ', []), 
-	    format(user_error, Format, Args), 
-	    format(user_error, ']~n', [])
-	;   true
-	),
-	(   feature(debug_on_error, true)
-	->  trace
-	;   true
-	).
-
-
-%	$warn_undefined(+Goal, +Dwims)
-%	Tell the user that the predicate implied by `Goal' does not exists,
-%	If there are alternatives (DWIM) tell the user about them.
-
-:- module_transparent
-	$warn_undefined/2,
-	$write_alternatives/1,
-	$predicate_name/2.
-
-$warn_undefined(Goal, Dwims) :-
-	$predicate_name(Goal, Name),
-	$ttyformat('[WARNING: Undefined predicate: `~w''', [Name]),
-	(   Dwims == []
-	;   $ttyformat('~nHowever there are definitions for:'), 
-	    $write_alternatives(Dwims)
-	), !,
-	$ttyformat(']~n').
-
-$write_alternatives([]) :- !.
-$write_alternatives([Dwim|Rest]) :-
-	$predicate_name(Dwim, Name), 
-	$ttyformat('~n~t~8|~w', [Name]), 
-	$write_alternatives(Rest).
-
-%	$predicate_name(+Head, -String)
-%	Convert `Head' into a predicate name.
-
-$predicate_name(Goal, String) :-
-	$strip_module(Goal, Module, Head), 
-	functor(Head, Name, Arity), 
-	(   memberchk(Module, [user, system])
-	->  sformat(String, '~w/~w',	[Name, Arity])
-	;   sformat(String, '~w:~w/~w',	[Module, Name, Arity])
-	).
-
-
+	    
 :- dynamic
 	user:portray/1.
 :- multifile
@@ -407,24 +375,47 @@ $predicate_name(Goal, String) :-
 user:(file_search_path(library, Dir) :-
 	library_directory(Dir)).
 user:file_search_path(swi, Home) :-
-	feature(home, Home).
+	current_prolog_flag(home, Home).
 user:file_search_path(foreign, swi(ArchLib)) :-
-	feature(arch, Arch),
-	concat('lib/', Arch, ArchLib).
+	current_prolog_flag(arch, Arch),
+	atom_concat('lib/', Arch, ArchLib).
 user:file_search_path(foreign, swi(lib)).
+user:file_search_path(user_profile, '.').
+user:file_search_path(user_profile, UserHome) :-
+	catch(expand_file_name(~, [UserHome]), _, fail).
+user:file_search_path(user_profile, SwiHome) :-
+	current_prolog_flag(windows, true), 		% single user machine
+	current_prolog_flag(home, SwiHome).
+
+
+%	expand_file_search_path(+Spec, -Expanded)
+%
+%	Expand a search path.  The system uses depth-first search upto a
+%	specified depth.  If this depth is exceeded an exception is raised.
+%	TBD: bread-first search?
 
 expand_file_search_path(Spec, Expanded) :-
+	catch($expand_file_search_path(Spec, Expanded, 0, []),
+	      loop(Used),
+	      throw(error(loop_error(Spec), file_search(Used)))).
+
+$expand_file_search_path(Spec, Expanded, N, Used) :-
 	functor(Spec, Alias, 1),
 	user:file_search_path(Alias, Exp0),
-	expand_file_search_path(Exp0, Exp1),
+	NN is N + 1,
+	(   NN > 16
+	->  throw(loop(Used))
+	;   true
+	),
+	$expand_file_search_path(Exp0, Exp1, NN, [Alias=Exp0|Used]),
 	arg(1, Spec, Base),
 	$make_path(Exp1, Base, Expanded).
-expand_file_search_path(Spec, Spec) :-
+$expand_file_search_path(Spec, Spec, _, _) :-
 	atomic(Spec).
 
 $make_path(Dir, File, Path) :-
-	concat(_, /, Dir), !,
-	concat(Dir, File, Path).
+	atom_concat(_, /, Dir), !,
+	atom_concat(Dir, File, Path).
 $make_path(Dir, File, Path) :-
 	$concat_atom([Dir, '/', File], Path).
 
@@ -433,30 +424,78 @@ $make_path(Dir, File, Path) :-
 		*         FILE CHECKING         *
 		*********************************/
 
+%	absolute_file_name(+Term, +Args, -AbsoluteFile)
+
+absolute_file_name(Spec, Args, Path) :-
+	(   select(extensions(Exts), Args, Conditions)
+	->  true
+	;   select(file_type(Type), Args, Conditions)
+	->  $file_type_extensions(Type, Exts)
+	;   Conditions = Args,
+	    Exts = ['']
+	),
+	(   select(solutions(Sols), Conditions, C1)
+	->  true
+	;   Sols = first,
+	    C1 = Conditions
+	),
+	(   select(file_errors(FileErrors), C1, C2)
+	->  true
+	;   FileErrors = error,
+	    C2 = C1
+	),
+	(   $chk_file(Spec, Exts, C2, Path)
+	*-> (   Sols == first
+	    ->  !
+	    ;   true
+	    )
+	;   (   FileErrors == fail
+	    ->  fail
+	    ;   throw(error(existence_error(source_sink, Spec), _))
+	    )
+	).
+
+$file_type_extensions(Type, Exts) :-
+	'$current_module'('$bags', _File), !,
+	findall(Ext, user:prolog_file_type(Ext, Type), Exts0),
+	append(Exts0, [''], Exts).
+$file_type_extensions(prolog, [pl, '']). % findall is not yet defined ...
+
+%	user:prolog_file_type/2
+%
+%	Define type of file based on the extension.  This is used by
+%	absolute_file_name/3 and may be used to extend the list of
+%	extensions used for some type.
+
+:- multifile(user:prolog_file_type/2),
+   dynamic(user:prolog_file_type/2).
+
+user:prolog_file_type(pl,	prolog).
+user:prolog_file_type(Ext,	prolog) :-
+	current_prolog_flag(associate, Ext),
+	Ext \== pl.
+user:prolog_file_type(qlf,	qlf).
+user:prolog_file_type(Ext,	executable) :-
+	current_prolog_flag(shared_object_extension, Ext).
+
 %	File is a specification of a Prolog source file. Return the full
 %	path of the file.
 
-$check_file(0, _) :- !, fail.			% deal with variables
-$check_file(user, user) :- !.
-$check_file(File, Absolute) :-
-	flag($compiling, database, database), !,
-	$chk_file(File, ['.qlf', '.pl', ''], exists, Absolute).
-$check_file(File, Absolute) :-
-	$chk_file(File, ['.pl', ''], exists, Absolute).
-
 $chk_file(Spec, Extensions, Cond, FullName) :-
 	$canonise_extensions(Extensions, Exts),
 	$dochk_file(Spec, Exts, Cond, FullName).
 
 $dochk_file(Spec, Extensions, Cond, FullName) :-
+	compound(Spec),
 	functor(Spec, Alias, 1),
 	user:file_search_path(Alias, _), !,
-	$chk_alias_file(Spec, Extensions, Cond, FullName).
+	working_directory(CWD, CWD),
+	$chk_alias_file(Spec, Extensions, Cond, CWD, FullName).
 $dochk_file(Term, Ext, Cond, FullName) :-	% allow a/b, a-b, etc.
 	\+ atomic(Term), !,
 	term_to_atom(Term, Raw),
 	atom_chars(Raw, S0),
-	delete(S0, 0' , S1),
+	delete(S0, ' ', S1),
 	atom_chars(Atom, S1),
 	$dochk_file(Atom, Ext, Cond, FullName).
 $dochk_file(File, Exts, Cond, FullName) :-
@@ -477,20 +516,39 @@ $dochk_file(File, Exts, Cond, FullName) :-
 	$absolute_file_name(Extended, FullName).
 
 :- dynamic
-	$search_path_file_cache/4.
+	$search_path_file_cache/5.
 :- volatile
-	$search_path_file_cache/4.
+	$search_path_file_cache/5.
 
-$chk_alias_file(Spec, Exts, Cond, FullFile) :-
-	$search_path_file_cache(Spec, Cond, FullFile, Exts).
-$chk_alias_file(Spec, Exts, Cond, FullFile) :-
+:- set_prolog_flag(verbose_file_search, false).
+
+$chk_alias_file(Spec, Exts, Cond, CWD, FullFile) :-
+	$search_path_file_cache(Spec, Cond, CWD, FullFile, Exts),
+	$file_condition(Cond, FullFile),
+	$search_message(file_search(cache(Spec, Cond), FullFile)).
+$chk_alias_file(Spec, Exts, Cond, CWD, FullFile) :-
 	expand_file_search_path(Spec, Expanded),
 	$extend_file(Expanded, Exts, LibFile),
-	$file_condition(Cond, LibFile),
-	$absolute_file_name(LibFile, FullFile),
-	\+ $search_path_file_cache(Spec, Cond, FullFile, Exts),
-	asserta($search_path_file_cache(Spec, Cond, FullFile, Exts)).
+	(   $file_condition(Cond, LibFile),
+	    $absolute_file_name(LibFile, FullFile),
+	    \+ $search_path_file_cache(Spec, Cond, CWD, FullFile, Exts),
+	    asserta($search_path_file_cache(Spec, Cond, CWD, FullFile, Exts))
+	->  $search_message(file_search(found(Spec, Cond), FullFile))
+	;   $search_message(file_search(tried(Spec, Cond), LibFile)),
+	    fail
+	).
+
+$search_message(Term) :-
+	current_prolog_flag(verbose_file_search, true), !,
+	print_message(informational, Term).
+$search_message(_).
+
+
 	
+%	$file_condition(+Condition, +Path)
+%
+%	Verify Path satisfies Condition. 
+
 $file_condition([], _) :- !.
 $file_condition([H|T], File) :- !,
 	$file_condition(H, File),
@@ -541,9 +599,9 @@ $canonise_extensions(E, [CE]) :-
 
 $canonise_extension('', '') :- !.
 $canonise_extension(DotAtom, DotAtom) :-
-	concat('.', _, DotAtom), !.
+	atom_concat('.', _, DotAtom), !.
 $canonise_extension(Atom, DotAtom) :-
-	concat('.', Atom, DotAtom).
+	atom_concat('.', Atom, DotAtom).
 
 
 		/********************************
@@ -586,44 +644,56 @@ $ifcompiling(G) :-
 preprocessor(Old, New) :-
 	flag($preprocessor, Old, New).
 
-$open_source(File, Goal) :-
-	preprocessor(none, none), !,
-	seeing(Old), see(File),
-	$open_source_call(File, Goal, True),
-	seen, see(Old),
+$open_source(user, user_input, Goal) :- !,
+	$push_input_context,
+	$open_source_call(user_input, user, Goal, True),
+	$pop_input_context,
 	True == yes.
-$open_source(File, Goal) :-
+$open_source(File, In, Goal) :-
+	preprocessor(none, none), !,
+	$push_input_context,
+	open(File, read, In),
+	$open_source_call(In, File, Goal, True),
+	close(In),
+	$pop_input_context,
+	True == yes.
+$open_source(File, In, Goal) :-
 	preprocessor(Pre, Pre),
 	(   $substitute_atom('%f', File, Pre, Command)
-	->  seeing(Old), see(pipe(Command)),
-	    $open_source_call(File, Goal, True),
-	    seen, see(Old), !,
+	->  $push_input_context,
+	    open(pipe(Command), read, In),
+	    $open_source_call(In, File, Goal, True),
+	    close(In),
+	    $pop_input_context,
 	    True == yes
-	;   $warning('Illegal preprocessor specification: `~w''', [Pre]),
-	    fail
+	;   throw(error(domain_error(preprocessor, Pre), _))
 	).
 
 
-$open_source_call(File, Goal, Status) :-
+:- flag($load_input, _, -1).
+
+$open_source_call(File, In, Goal, Status) :-
 	flag($compilation_level, Level, Level+1),
-	ignore(user:$start_compilation(File, Level)),
-	(   Goal
+	flag($load_input, OldIn, In),
+	ignore(user:'$start_compilation'(File, Level)),
+	(   catch(Goal, E,
+		  (print_message(error, E),
+		   fail))
 	->  Status = yes
 	;   Status = no
 	),
-	ignore(user:$end_compilation(File, Level)),
+	ignore(user:'$end_compilation'(File, Level)),
+	flag($load_input, _, OldIn),
 	flag($compilation_level, _, Level).
 
 
-$substitute_atom(Old, New, Org, Result) :-
-	name(Old, OS),
-	name(New, NS),
-	name(Org, OrgS),
-	append(Before, Rest, OrgS),
-	append(OS, After, Rest), !,
-	append(Before, NS, R1),
-	append(R1, After, R2), !,
-	name(Result, R2).
+%	$substitute_atom(+From, +To, +In, -Out)
+
+$substitute_atom(From, To, In, Out) :-
+	sub_atom(In, B, _, A, From),
+	sub_atom(In, 0, B, _, Before),
+	sub_atom(In, _, A, 0, After),
+	concat_atom([Before, To, After], Out).
 
 
 		/********************************
@@ -676,15 +746,16 @@ use_module(Files, Import) :-
 consult(List) :-
 	load_files(List).
 
-%	Compilation extensions
-
-$compiler_extension('.qlf', $qload_file).
-$compiler_extension('',  $consult_file).
+%	$consult_goal(+Path, -Goal)
+%
+%	Determine how to load the indicated file
 
 $consult_goal(Path, Goal) :-
-	$compiler_extension(Ext, Goal),
-	concat(_, Ext, Path), !.
-
+	(   file_name_extension(_, Ext, Path),
+	    user:prolog_file_type(Ext, qlf)
+	->  Goal = $qload_file
+	;   Goal = $consult_file
+	).
 
 %	$consult_file(+File, +Options)
 %	
@@ -701,12 +772,18 @@ load_files(Files) :-
 	load_files(Files, []).
 load_files(Files, Options) :-
 	$strip_module(Files, Module, TheFiles),
-        $load_files(TheFiles, Module, Options).
+        with_mutex('$load', $load_files(TheFiles, Module, Options)).
 
 $load_files([], _, _) :- !.
 $load_files([H|T], Module, Options) :- !,
-	$load_file(Module:H, Options),
+	$load_files(H, Module, Options),
 	$load_files(T, Module, Options).
+$load_files(Spec, Module, Options) :-
+	atom(Spec),
+	$get_option(expand(Expand), Options, true),
+	Expand == true, !,
+	expand_file_name(Spec, Files),
+	$load_files(Files, Module, [expand(false)|Options]).
 $load_files(File, Module, Options) :-
 	$load_file(Module:File, Options).
 
@@ -727,23 +804,51 @@ $noload(changed, FullFile) :-
         time_file(FullFile, Modified),
         Modified @=< LoadTime, !.
 
+%	$qlf_file(+PlFile, -LoadFile)
+%
+%	Return the QLF file if it exists.  Might check for modification
+%	time, version, etc.
+
+$qlf_file(user, user) :- !.
+$qlf_file(FullFile, QlfFile) :-
+	flag($compiling, database, database),
+	file_name_extension(Base, PlExt, FullFile),
+	user:prolog_file_type(PlExt, prolog),
+	user:prolog_file_type(QlfExt, qlf),
+	file_name_extension(Base, QlfExt, QlfFile),
+	access_file(QlfFile, read).
+$qlf_file(FullFile, FullFile).
+
 :- flag($load_silent, _, false).
 
 $load_file(Spec, Options) :-
 	statistics(heapused, OldHeap),
 	statistics(cputime, OldTime),
  
+        $strip_module(Spec, Module, File),
+	(   File == user
+	->  true
+	;   absolute_file_name(File,
+			       [ file_type(prolog),
+				 access(read)
+			       ],
+			       FullFile)
+	),
+	    
 	$get_option(imports(Import), Options, all),
 	$get_option(must_be_module(IsModule), Options, false),
 	flag($load_silent, DefSilent, DefSilent),
 	$get_option(silent(Silent), Options, DefSilent),
 	flag($load_silent, _, Silent),
 	$get_option(if(If), Options, true),
+	$get_option(autoload(Autoload), Options, false),
 
-        $strip_module(Spec, Module, File),
-	    
-	(   once($chk_file(File, ['.pl', ''], exists, FullFile)),
-	    $noload(If, FullFile)
+	(   Autoload == false
+	->  flag($autoloading, AutoLevel, AutoLevel)
+	;   flag($autoloading, AutoLevel, AutoLevel+1)
+	),
+
+	(   $noload(If, FullFile)
 	->  (   $current_module(LoadModule, FullFile)
 	    ->  $import_list(Module, LoadModule, Import)
 	    ;   (   Module == user
@@ -751,52 +856,52 @@ $load_file(Spec, Options) :-
 		;   $load_file(Spec, [if(true)|Options])
 		)
 	    )
-	;   (   $check_file(File, Absolute)
-	    *-> true
-	    ;   $warning('No such file: ~w', Spec),
-		fail
-	    ),
-
-	    $calleventhook(load_file(Absolute, start)),
-	    (   $consult_goal(Absolute, Goal),
-	        $apply(Goal, [Absolute, Module, Import, IsModule, Action, LM])
-	    ->  true
-	    ;   $warning('Failed to load file: ~w', Spec),
-		$calleventhook(load_file(Absolute, false)),
-		fail
-	    ),
-	    $calleventhook(load_file(Absolute, true)),
+	;   $qlf_file(FullFile, Absolute),
 
+	    flag($compilation_level, Level, Level),
 	    (   Silent == false,
-		(flag($autoloading, 0, 0) ; flag($verbose_autoload, on, on))
-	    ->  statistics(heapused, Heap),
-		statistics(cputime, Time),
-		HeapUsed is Heap - OldHeap,
-		TimeUsed is Time - OldTime,
-		$confirm_file(File, Absolute, ConfirmFile),
-		$confirm_module(LM, ConfirmModule),
+		(   flag($autoloading, 0, 0)
+		;   current_prolog_flag(verbose_autoload, true)
+		)
+	    ->	MessageLevel = informational
+	    ;	MessageLevel = silent
+	    ),
 
-		$ttyformat('~N~w ~w~w, ~2f sec, ~D bytes.~n',
-			   [ConfirmFile, Action, ConfirmModule,
-			    TimeUsed, HeapUsed])
-	    ;   true
-	    )
+	    $print_message(silent /*MessageLevel*/,
+			   load_file(start(Level,
+					   file(File, Absolute)))),
+	    (   $consult_goal(Absolute, Goal),
+		call(Goal, Absolute, Module, Import, IsModule, Action, LM)
+	    ->  true
+	    ;   print_message(error, load_file(failed(Spec))),
+		fail
+	    ),
+
+	    statistics(heapused, Heap),
+	    statistics(cputime, Time),
+	    HeapUsed is Heap - OldHeap,
+	    TimeUsed is Time - OldTime,
+
+	    $print_message(MessageLevel,
+			   load_file(done(Level,
+					  file(File, Absolute),
+					  Action,
+					  LM,
+					  TimeUsed,
+					  HeapUsed)))
 	),
+	flag($autoloading, _, AutoLevel),
 	flag($load_silent, _, DefSilent).
 
 
-$confirm_file(library(_), Absolute, Absolute) :- !.
-$confirm_file(File, _, File).
+$print_message(Level, Term) :-
+	$current_module('$messages', _), !,
+	print_message(Level, Term).
+$print_message(_Level, _Term).
 
-$confirm_module(user, '') :- !.
-$confirm_module(Module, Message) :-
-	atom(Module), !,
-	concat(' into ', Module, Message).
-$confirm_module(_, '').
-
-$read_clause(Clause) :-				% get the first non-syntax
-	repeat,					% error
-	    read_clause(Clause), !.
+$print_message_fail(E) :-
+	print_message(error, E),
+	fail.
 
 $consult_file(Absolute, Module, Import, IsModule, What, LM) :-
 	$set_source_module(Module, Module), !, % same module
@@ -819,9 +924,8 @@ $consult_file_2(Absolute, Module, Import, IsModule, What, LM) :-
 	$assert_load_context_module(Absolute, OldModule),
 
 	$style_check(OldStyle, OldStyle),	% Save style parameters
-	$open_source(Absolute, (		% Load the file
-	    $read_clause(First),
-	    $load_file(First, Absolute, Import, IsModule, LM))),
+	$open_source(Absolute, In,
+		     $load_file(In, Absolute, Import, IsModule, LM)),
 	$style_check(_, OldStyle),		% Restore old style
 	$set_source_module(_, OldModule).	% Restore old module
 
@@ -845,52 +949,50 @@ $assert_load_context_module(File, Module) :-
 $assert_load_context_module(File, Module) :-
 	recordz($load_context_module, File/Module, _).
 
-%   $load_file(+FirstTerm, +Path, +Import, +IsModule, -Module)
+%   $load_file(+FirstTerm, +In, +Path, +Import, +IsModule, -Module)
 %
-%   $load_file5 does the actual loading. The first term has already been
+%   $load_file/6 does the actual loading. The first term has already been
 %   read as this may be the module declaraction.
 
-$load_file((?- module(Module, Public)), File, all, _, Module) :- !,
-	$load_module(Module, Public, all, File).
-$load_file((:- module(Module, Public)), File, all, _, Module) :- !,
-	$load_module(Module, Public, all, File).
-$load_file((?- module(Module, Public)), File, Import, _, Module) :- !,
-	$load_module(Module, Public, Import, File).
-$load_file((:- module(Module, Public)), File, Import, _, Module) :- !,
-	$load_module(Module, Public, Import, File).
-$load_file(_, File, _, true, _) :- !,
-	$warning('use_module: ~w is not a module file', [File]),
-	fail.
-$load_file(end_of_file, _, _, _, Module) :- !,		% empty file
+$load_file(In, File, Import, IsModule, Module) :-
+	(   peek_char(In, #)
+	->  skip(In, 10)
+	;   true
+	),
+	read_clause(In, First),
+	$load_file(First, In, File, Import, IsModule, Module).
+
+$load_file((?- module(Module, Public)), In, File, all, _, Module) :- !,
+	$load_module(Module, Public, all, In, File).
+$load_file((:- module(Module, Public)), In, File, all, _, Module) :- !,
+	$load_module(Module, Public, all, In, File).
+$load_file((?- module(Module, Public)), In, File, Import, _, Module) :- !,
+	$load_module(Module, Public, Import, In, File).
+$load_file((:- module(Module, Public)), In, File, Import, _, Module) :- !,
+	$load_module(Module, Public, Import, In, File).
+$load_file(_, _, File, _, true, _) :- !,
+	throw(error(domain_error(module_file, File), _)).
+$load_file(end_of_file, _, _, _, _, Module) :- !, 	% empty file
 	$set_source_module(Module, Module).
-$load_file(FirstClause, File, _, false, Module) :- !,
+$load_file(FirstClause, In, File, _, false, Module) :- !,
 	$set_source_module(Module, Module),
 	$ifcompiling($qlf_start_file(File)),
 	ignore($consult_clause(FirstClause, File)),
-	repeat,
-	    read_clause(Clause),
-	    $consult_clause(Clause, File), !,
+	$consult_stream(In, File),
 	$ifcompiling($qlf_end_part).
 
-
 $reserved_module(system).
 $reserved_module(user).
 
-$load_module(Reserved, _, _, _) :-
+$load_module(Reserved, _, _, _, _) :-
 	$reserved_module(Reserved), !,
-	$warning('Cannot load into module "~w": reserved module name',
-		 [Reserved]),
-	fail.
-$load_module(Module, Public, Import, File) :-
+	throw(error(permission_error(load, module, Reserved), _)).
+$load_module(Module, Public, Import, In, File) :-
 	$set_source_module(OldModule, OldModule),
 	$declare_module(Module, File),
 	$export_list(Module, Public),
 	$ifcompiling($qlf_start_module(Module)),
-
-	repeat,
-	    read_clause(Clause),
-	    $consult_clause(Clause, File), !,
-
+	$consult_stream(In, File),
 	Module:$check_export,
 	$ifcompiling($qlf_end_part),
 	$import_list(OldModule, Module, Import).
@@ -915,23 +1017,43 @@ $import_all([Head|Rest], Context, Source) :-
 
 
 $export_list(_, []) :- !.
-$export_list(Module, [Name/Arity|Rest]) :- !,
-	functor(Term, Name, Arity),
-	export(Module:Term),
+$export_list(Module, [Name/Arity|Rest]) :-
+	functor(Term, Name, Arity), !,
+	catch(export(Module:Term), E, print_message(error, E)),
 	$export_list(Module, Rest).
 $export_list(Module, [Term|Rest]) :-
-	$warning('Illegal predicate specification in public list: `~w''',
-		 [Term]),
+	print_message(error, type_error(predicate_indicator, Term)),
 	$export_list(Module, Rest).
 
-$consult_clause(Clause, File) :-
-	expand_term(Clause, Expanded),
-	(   $store_clause(Expanded, File)
-	->  Clause == end_of_file
-	;   fail
-	).
+%	$consult_stream(+Stream, +File)
+%
+%	Read and record all clauses until the rest of the file.
 
-$execute_directive(Goal) :-
+$consult_stream(In, File) :-
+	repeat,
+	catch($consult_stream2(In, File),
+	      E,
+	      (	  print_message(error, E),
+		  fail
+	      )), !.
+
+	      
+$consult_stream2(In, File) :-
+	repeat,
+	    read_clause(In, Clause),
+	    expand_term(Clause, Expanded),
+	    $store_clause(Expanded, File),
+	    Clause == end_of_file, !.
+
+$consult_clause(Clause, File) :-
+	catch((expand_term(Clause, Expanded),
+	       $store_clause(Expanded, File)),
+	       E,
+	       $print_message_fail(E)).
+
+$execute_directive(include(File), F) :- !,
+	$expand_include(File, F).
+$execute_directive(Goal, _) :-
 	compiling, !,
 	$add_directive_wic2(Goal, Type),
 	(   Type == call		% suspend compiling into .qlf file
@@ -943,7 +1065,7 @@ $execute_directive(Goal) :-
 	    )
 	;   $execute_directive2(Goal)
 	).
-$execute_directive(Goal) :-
+$execute_directive(Goal, _) :-
 	$execute_directive2(Goal).
 
 $execute_directive2(Goal) :-
@@ -951,10 +1073,7 @@ $execute_directive2(Goal) :-
 	catch(Module:Goal, Term, $exception_in_directive(Term)), !.
 $execute_directive2(Goal) :-
 	$set_source_module(Module, Module),
-	(   Module == user
-	->  $warning('Directive failed: ~w', [Goal])
-	;   $warning('Directive failed: ~w:~w', [Module, Goal])
-        ),
+	print_message(warning, goal_failed(directive, Module:Goal)),
 	fail.
 
 $exception_in_directive(Term) :-
@@ -975,8 +1094,7 @@ $add_directive_wic2(Goal, Type) :-
 $add_directive_wic2(Goal, _) :-
 	(   flag($compiling, qlf, qlf)	% no problem for qlf files
 	->  true
-	;   $warning('Cannot compile mixed loading/calling directives: ~w',
-		     [Goal])
+	;   print_message(error, mixed_directive(Goal))
 	).
 	
 $common_goal_type((A,B), Type) :- !,
@@ -1009,34 +1127,74 @@ $load_goal(use_module(_, _)) :- flag($compiling, wic, wic).
 
 :- user:dynamic(term_expansion/2).
 :- user:multifile(term_expansion/2).
+:- user:dynamic(goal_expansion/2).
+:- user:multifile(goal_expansion/2).
 
 expand_term(Term, Expanded) :-		% local term-expansion
 	$term_expansion_module(Module),
-	Module:term_expansion(Term, Expanded), !.
-expand_term(Term, Expanded) :-
-	$translate_rule(Term, Expanded), !.
+	Module:term_expansion(Term, Expanded0), !,
+	$expand_clauses(Expanded0, Expanded).
+expand_term(Head --> Body, Expanded) :-
+	$translate_rule(Head --> Body, Expanded0), !,
+	$expand_clauses(Expanded0, Expanded).
+expand_term(Term0, Term) :-
+	$goal_expansion_module(_), !,
+	$expand_clauses(Term0, Term).
 expand_term(Term, Term).
 
+
 $store_clause([], _) :- !.
 $store_clause([C|T], F) :- !,
 	$store_clause(C, F),
 	$store_clause(T, F).
 $store_clause(end_of_file, _) :- !.
-$store_clause((:- Goal), _) :- !,
-	$execute_directive(Goal).
-$store_clause((?- Goal), _) :- !,
-	$execute_directive(Goal).
+$store_clause((:- Goal), F) :- !,
+	$execute_directive(Goal, F).
+$store_clause((?- Goal), F) :- !,
+	$execute_directive(Goal, F).
 $store_clause((_, _), _) :- !,
-	$warning('Full stop in clause body? (attempt to define ,/2)').
-$store_clause((_:-B), _) :-
-	nonvar(B), B = (_:-_), !,
-	$warning('Clause not closed by `.''? (attempt to call :-/2)').
-$store_clause($source_location(File, Line):Term, _) :-
+	print_message(error, cannot_redefine_comma),
+	fail.
+$store_clause($source_location(File, Line):Term, _) :- !,
 	$record_clause(Term, File:Line, Ref),
-        $ifcompiling($qlf_assert_clause(Ref, development)).
+	$qlf_assert_clause(Ref).
 $store_clause(Term, File) :-
 	$record_clause(Term, File, Ref),
-        $ifcompiling($qlf_assert_clause(Ref, development)).
+        $qlf_assert_clause(Ref).
+
+$qlf_assert_clause(_) :-
+	flag($compiling, database, database), !.
+$qlf_assert_clause(Ref) :-
+	$qlf_assert_clause(Ref, development).
+
+
+		 /*******************************
+		 *	     INCLUDE		*
+		 *******************************/
+
+$expand_include(File, FileInto) :-
+	absolute_file_name(File,
+			   [ file_type(prolog),
+			     access(read)
+			   ], Path),
+	$push_input_context,
+	open(Path, read, In),
+	read_clause(In, Term0),
+	$read_include_file(Term0, In, Terms),
+	close(In),
+	$pop_input_context,
+	$consult_clauses(Terms, FileInto).
+
+$read_include_file(end_of_file, _, []) :- !.
+$read_include_file(T0, In, [T0|T]) :-
+	read_clause(In, T1),
+	$read_include_file(T1, In, T).
+
+$consult_clauses([], _).
+$consult_clauses([H|T], File) :-
+	$consult_clause(H, File),
+	$consult_clauses(T, File).
+
 
 		 /*******************************
 		 *	 FOREIGN INTERFACE	*
@@ -1050,6 +1208,89 @@ $store_clause(Term, File) :-
 	$foreign_registered/2.
 
 
+		 /*******************************
+		 *   GOAL_EXPANSION/2 SUPPORT	*
+		 *******************************/
+
+$expand_clauses(X, X) :-
+	var(X), !.
+$expand_clauses([H0|T0], [H|T]) :- !,
+	$expand_clauses(H0, H),
+	$expand_clauses(T0, T).
+$expand_clauses((Head :- Body), (Head :- ExpandedBody)) :-
+	nonvar(Body), !,
+	expand_goal(Body,  ExpandedBody).
+$expand_clauses(Head, Head).
+
+expand_goal(A, B) :-
+        $do_expand_body(A, B0),
+	$tidy_body(B0, B).
+
+$do_expand_body(G, G) :-
+        var(G), !.
+$do_expand_body((A,B), (EA,EB)) :- !,
+        $do_expand_body(A, EA),
+        $do_expand_body(B, EB).
+$do_expand_body((A;B), (EA;EB)) :- !,
+        $do_expand_body(A, EA),
+        $do_expand_body(B, EB).
+$do_expand_body((A->B), (EA->EB)) :- !,
+        $do_expand_body(A, EA),
+        $do_expand_body(B, EB).
+$do_expand_body((A*->B), (EA*->EB)) :- !,
+        $do_expand_body(A, EA),
+        $do_expand_body(B, EB).
+$do_expand_body((\+A), (\+EA)) :- !,
+        $do_expand_body(A, EA).
+$do_expand_body(not(A), not(EA)) :- !,
+        $do_expand_body(A, EA).
+$do_expand_body(call(A), call(EA)) :- !,
+        $do_expand_body(A, EA).
+$do_expand_body(once(A), once(EA)) :- !,
+        $do_expand_body(A, EA).
+$do_expand_body(ignore(A), ignore(EA)) :- !,
+        $do_expand_body(A, EA).
+$do_expand_body(forall(A, B), forall(EA, EB)) :- !,
+        $do_expand_body(A, EA),
+        $do_expand_body(B, EB).
+$do_expand_body(A, B) :-
+        $goal_expansion_module(M),
+        M:goal_expansion(A, B0), !,
+	$do_expand_body(B0, B).
+$do_expand_body(A, A).
+
+%	Delete extraneous true's that result from goal_expansion(..., true)
+%
+%	Is the really necessary?  Should we only do it if -O is effective?
+
+$tidy_body(A, A) :-
+	current_prolog_flag(optimise, false), !.
+$tidy_body(A, A) :-
+        var(A), !.
+$tidy_body((A,B), (A, TB)) :-
+        var(A), !,
+        $tidy_body(B, TB).
+$tidy_body((A,B), (TA, B)) :-
+        var(B), !,
+        $tidy_body(A, TA).
+$tidy_body(((A,B),C), R) :- !,
+	$tidy_body((A,B,C), R).
+$tidy_body((true,A), R) :- !,
+        $tidy_body(A, R).
+$tidy_body((A,true), R) :- !,
+        $tidy_body(A, R).
+$tidy_body((A,B), (TA, TB)) :- !,
+        $tidy_body(A, TA),
+        $tidy_body(B, TB).
+$tidy_body((A;B), (TA; TB)) :- !,
+        $tidy_body(A, TA),
+        $tidy_body(B, TB).
+$tidy_body((A->B), (TA->TB)) :- !,
+        $tidy_body(A, TA),
+        $tidy_body(B, TB).
+$tidy_body(A, A).
+
+
 		/********************************
 		*        GRAMMAR RULES          *
 		*********************************/
@@ -1128,8 +1369,8 @@ $t_body((C->T|E), S, SR, (Ct->Tt;Et)) :- !,
 $t_body((C->T), S, SR, (Ct->Tt)) :- !,
 	$t_body(C, S, SR1, Ct),
 	$t_body(T, SR1, SR, Tt).
-$t_body((\+ C), S, SR, (\+ Ct)) :- !,
-	$t_body(C, S, SR, Ct).
+$t_body((\+ C), S, S, (\+ Ct)) :- !,
+	$t_body(C, S, _, Ct).
 $t_body(T, S, SR, Tt) :-
 	$extend([S, SR], T, Tt).
 
@@ -1139,6 +1380,8 @@ $t_fill(S, SR, S1, T, (T, SR=S)) :-
 $t_fill(_S, SR, SR, T, T).
 
 
+$extend(More, M:OldT, M:NewT) :- !,
+	$extend(More, OldT, NewT).
 $extend(More, OldT, NewT) :-
 	OldT =.. OldL,
 	append(OldL, More, NewL),
@@ -1164,15 +1407,14 @@ phrase(RuleSet, Input, Rest) :-
 		*     WIC CODE COMPILER         *
 		*********************************/
 
-/*  This  entry  point  is  called  from  pl-main.c  if  the  -c  option
-    (intermediate  code  compilation) is given.  It's job is simple: get
-    the output file  and  input  files,  open  the  output  file,  setup
-    intermediate  code  compilation  flag  and  finally just compile the
-    input files.
-*/
+/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
+This entry point is called from pl-main.c  if the -c option (compile) is
+given. It compiles all files and finally calls qsave_program to create a
+saved state.
+- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
 $compile_wic :-
-	$argv(Argv),			% gets main() argv as a list of atoms
+	current_prolog_flag(argv, Argv),
 	$get_files_argv(Argv, Files),
 	$translate_options(Argv, Options),
 	$option(compileout, Out, Out),
@@ -1186,8 +1428,8 @@ $get_files_argv([_|Rest], Files) :-
 
 $translate_options([], []).
 $translate_options([O|T0], [Opt|T]) :-
-	atom_chars(O, [0'-,0'-|Rest]),
-	$split(Rest, "=", Head, Tail), !,
+	atom_chars(O, [-,-|Rest]),
+	$split(Rest, [=], Head, Tail), !,
 	atom_chars(Name, Head),
 	name(Value, Tail),
 	Opt =.. [Name, Value],
@@ -1213,6 +1455,26 @@ append([], L, L).
 append([H|T], L, [H|R]) :-
 	append(T, L, R).
 
+%	select(?Elem, ?List1, ?List2)
+%	Is true when List1, with Elem removed results in List2.
+
+select(X, [X|Tail], Tail).
+select(Elem, [Head|Tail], [Head|Rest]) :-
+	select(Elem, Tail, Rest).
+
+
+		 /*******************************
+		 *   HANDLE TRACER 'L'-COMMAND	*
+		 *******************************/
+
+:- multifile
+	user:prolog_list_goal/1.
+
+'$prolog_list_goal'(Goal) :-
+	user:prolog_list_goal(Goal), !.
+'$prolog_list_goal'(Goal) :-
+	user:listing(Goal).
+
 
 		 /*******************************
 		 *	       HALT		*
@@ -1229,12 +1491,14 @@ halt :-
 
 at_halt(Spec) :-
 	$strip_module(Spec, Module, Goal),
-	assert(system:$at_halt(Module:Goal)).
+	asserta(system:$at_halt(Module:Goal)).
 
 $run_at_halt :-
-	$at_halt(Goal),
-	Goal,
-	fail ; true.
+	(   $at_halt(Goal),
+	    catch(Goal, E, print_message(error, E)),
+	    fail
+	;   true
+	).
 
 
 		/********************************
@@ -1246,17 +1510,17 @@ $run_at_halt :-
 	$load_additional_boot_files/0.
 
 $load_wic_files(Module, Files) :-
-	$execute_directive($set_source_module(OldM, Module)),
+	$execute_directive($set_source_module(OldM, Module), []),
 	$style_check(OldS, 2'1111),
 	flag($compiling, OldC, wic),
 	consult(Files),
-	$execute_directive($set_source_module(_, OldM)),
-	$execute_directive($style_check(_, OldS)),
+	$execute_directive($set_source_module(_, OldM), []),
+	$execute_directive($style_check(_, OldS), []),
 	flag($compiling, _, OldC).
 
 
 $load_additional_boot_files :-
-	$argv(Argv),
+	current_prolog_flag(argv, Argv),
 	$get_files_argv(Argv, Files),
 	(   Files \== []
 	->  format('Loading additional boot files~n'),
@@ -1265,14 +1529,18 @@ $load_additional_boot_files :-
 	;   true
         ).
 
-
-'$:-'	
-	format('Loading Prolog startup files~n', []),
-	source_location(File, _Line),
-	file_directory_name(File, Dir),
-	concat(Dir, '/load.pl', LoadFile),
-	$load_wic_files(system, [LoadFile]),
-	format('SWI-Prolog boot files loaded~n', []),
-	flag($compiling, OldC, wic),
-	$execute_directive($set_source_module(_, user)),
-	flag($compiling, _, OldC).
+'$:-'((format('Loading Prolog startup files~n', []),
+       source_location(File, _Line),
+       file_directory_name(File, Dir),
+       atom_concat(Dir, '/load.pl', LoadFile),
+       $load_wic_files(system, [LoadFile]),
+       (   current_prolog_flag(windows, true)
+       ->  atom_concat(Dir, '/menu.pl', MenuFile),
+	   $load_wic_files(system, [MenuFile])
+       ;   true
+       ),
+       format('SWI-Prolog boot files loaded~n', []),
+       flag($compiling, OldC, wic),
+       $execute_directive($set_source_module(_, user), []),
+       flag($compiling, _, OldC)
+      )).
