@@ -1,16 +1,38 @@
-/*  $Id: syspred.pl,v 1.35 1998/08/06 13:25:22 jan Exp $
+/*  $Id: syspred.pl,v 1.55 2002/02/01 16:49:10 jan Exp $
 
-    Copyright (c) 1990 Jan Wielemaker. All rights reserved.
-    jan@swi.psy.uva.nl
+    Part of SWI-Prolog
 
-    Purpose: Prolog system predicate definitions
+    Author:        Jan Wielemaker
+    E-mail:        jan@swi.psy.uva.nl
+    WWW:           http://www.swi-prolog.org
+    Copyright (C): 1985-2002, University of Amsterdam
+
+    This program is free software; you can redistribute it and/or
+    modify it under the terms of the GNU General Public License
+    as published by the Free Software Foundation; either version 2
+    of the License, or (at your option) any later version.
+
+    This program is distributed in the hope that it will be useful,
+    but WITHOUT ANY WARRANTY; without even the implied warranty of
+    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+    GNU General Public License for more details.
+
+    You should have received a copy of the GNU Lesser General Public
+    License along with this library; if not, write to the Free Software
+    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+
+    As a special exception, if you link this library with other files,
+    compiled with a Free Software compiler, to produce an executable, this
+    library does not by itself cause the resulting executable to be covered
+    by the GNU General Public License. This exception does not however
+    invalidate any other reasons why the executable file might be covered by
+    the GNU General Public License.
 */
 
 :- module($syspreds,
 	[ leash/1
 	, visible/1
 	, style_check/1
-	, please/3
 	, (spy)/1
 	, (nospy)/1
 	, trace/1
@@ -19,9 +41,7 @@
 	, debugging/0
 	, concat_atom/2
 	, term_to_atom/2
-	, atom_to_term/3
 	, int_to_atom/2
-	, gensym/2
 	, dwim_match/2
 	, source_file/1
 	, prolog_load_context/2
@@ -30,8 +50,6 @@
 	, predicate_property/2
 	, $predicate_property/2
 	, clause_property/2
-	, clause/2
-	, clause/3
 	, recorda/2
 	, recordz/2
 	, recorded/2
@@ -53,7 +71,6 @@
 	, arithmetic_function/1
         , default_module/2
 	, absolute_file_name/2
-	, absolute_file_name/3
 	, require/1
 	, call_with_depth_limit/3
 	]).	
@@ -104,21 +121,17 @@ $map_style_check(atom,     	    2'0000001).
 $map_style_check(singleton, 	    2'0000010).
 $map_style_check(dollar,   	    2'0000100).
 $map_style_check((discontiguous),   2'0001000).
-$map_style_check(string,	    2'0010000).
-$map_style_check(dynamic,	    2'0100000).
+$map_style_check(dynamic,	    2'0010000).
 
+style_check(+string) :- !,
+	set_prolog_flag(double_quotes, string).
+style_check(-string) :- !,
+	set_prolog_flag(double_quotes, codes).
 style_check(Spec) :-
 	$style_check(Old, Old),
 	$map_bits($map_style_check, Spec, Old, New),
 	$style_check(_, New).
 
-please(autoload, Old, New) :- !,
-	flag($enable_autoload, Old, New).
-please(verbose_autoload, Old, New) :- !,
-	flag($verbose_autoload, Old, New).
-please(Key, Old, New) :-
-	$please(Key, Old, New).
-
 :- module_transparent
 	trace/1,
 	trace/2,
@@ -126,6 +139,19 @@ please(Key, Old, New) :-
 	spy/1,
 	nospy/1.
 
+%	prolog:debug_control_hook(+Action)
+%
+%	Allow user-hooks in the Prolog debugger interaction.  See the calls
+%	below for the provided hooks.  We use a single predicate with action
+%	argument to avoid an uncontrolled poliferation of hooks.
+%
+%	TBD: What hooks to provide for trace/[1,2]
+
+:- multifile
+	prolog:debug_control_hook/1.	% +Action
+:- module_transparent
+	prolog:debug_control_hook/1.
+
 trace(Preds) :-
 	trace(Preds, +all).
 
@@ -134,18 +160,27 @@ trace([H|T], Ps) :- !,
 	trace(H, Ps),
 	trace(T, Ps).
 trace(Pred, Ports) :-
-	debug,
+	set_prolog_flag(debug, true),
 	$find_predicate(Pred, Preds),
 	Preds \== [],
 	(   member(Head, Preds),
-		$define_predicate(Head),
-	        $trace(Ports, Head),
-	        show_trace_point(Head),
+	        (   Head = _:_
+		->  QHead0 = Head
+		;   QHead0 = user:Head
+		),
+		$define_predicate(QHead0),
+	        (   predicate_property(QHead0, imported_from(M))
+		->  QHead0 = _:Plain,
+		    QHead = M:Plain
+		;   QHead = QHead0
+		),
+	        $trace(Ports, QHead),
+	        trace_ports(QHead, Tracing),
+	        print_message(informational, trace(QHead, Tracing)),
 	    fail
 	;   true
 	).
 
-
 trace_alias(all,  [trace_call, trace_redo, trace_exit, trace_fail]).
 trace_alias(call, [trace_call]).
 trace_alias(redo, [trace_redo]).
@@ -179,18 +214,17 @@ tag_list([H0|T0], F, [H1|T1]) :-
 	H1 =.. [F, H0],
 	tag_list(T0, F, T1).
 
-
 spy([]) :- !.
 spy([H|T]) :- !,
 	spy(H),
 	spy(T).
+spy(Spec) :-
+	prolog:debug_control_hook(spy(Spec)), !.
 spy(Spec) :-
 	$find_predicate(Spec, Preds),
 	member(Head, Preds),
 	    $define_predicate(Head),
 	    $spy(Head),
-	    $predicate_name(Head, Name),
-	    $ttyformat('Spy point on ~w~n', [Name]),
 	fail.
 spy(_).
 
@@ -198,64 +232,52 @@ nospy([]) :- !.
 nospy([H|T]) :- !,
 	nospy(H),
 	nospy(T).
+nospy(Spec) :-
+	prolog:debug_control_hook(nospy(Spec)), !.
 nospy(Spec) :-
 	$find_predicate(Spec, Preds),
 	member(Head, Preds),
 	    $nospy(Head),
-	    $predicate_name(Head, Name),
-	    $ttyformat('Spy point removed from ~w~n', [Name]),
 	fail.
 nospy(_).
 
 nospyall :-
-	current_predicate(_, Module:Head),
-	    $nospy(Module:Head),
+	prolog:debug_control_hook(nospyall),
+	fail.
+nospyall :-
+	spy_point(Head),
+	    $nospy(Head),
 	fail.
 nospyall.
 
 debugging :-
-	$debugging, !,
-	format('Debug mode is on; spy points (see spy/1) on:~n'),
-	$show_spy_points,
-	format('Trace points (see trace/1) on:~n'),
-	show_trace_points.
-
+	prolog:debug_control_hook(debugging), !.
 debugging :-
-	format('Debug mode is off~n').
+	current_prolog_flag(debug, true), !,
+	print_message(informational, debugging(on)),
+	findall(H, spy_point(H), SpyPoints),
+	print_message(informational, spying(SpyPoints)),
+	findall(trace(H,P), trace_point(H,P), TracePoints),
+	print_message(informational, tracing(TracePoints)).
+debugging :-
+	print_message(informational, debugging(off)).
 
-$show_spy_points :-
+spy_point(Module:Head) :-
 	current_predicate(_, Module:Head),
 	$get_predicate_attribute(Module:Head, spy, 1),
-	\+ predicate_property(Module:Head, imported_from(_)),
-	$predicate_name(Module:Head, Name),
-	format('~t~8|~w~n', [Name]),
-	fail.
-$show_spy_points.
+	\+ predicate_property(Module:Head, imported_from(_)).
 
-show_trace_points :-
+trace_point(Module:Head, Ports) :-
 	current_predicate(_, Module:Head),
-	$get_predicate_attribute(Module:Head, trace_any, 1),
-	\+ predicate_property(Module:Head, imported_from(_)),
-	show_trace_point(Module:Head),
-	fail.
-show_trace_points.
+	    $get_predicate_attribute(Module:Head, trace_any, 1),
+	    \+ predicate_property(Module:Head, imported_from(_)),
+	    trace_ports(Module:Head, Ports).
 
-:- module_transparent
-	show_trace_point/1,
-	show_trace_ports/1.
-
-show_trace_point(Head) :-
-	$predicate_name(Head, Name),
-	format('~t~8|~w:', [Name]),
-	show_trace_ports(Head),
-	nl.
-
-show_trace_ports(Head) :-
-	trace_alias(Port, [AttName]),
-	$get_predicate_attribute(Head, AttName, 1),
-	format(' ~w', [Port]),
-	fail.
-show_trace_ports(_).
+trace_ports(Head, Ports) :-
+	findall(Port,
+		(trace_alias(Port, [AttName]),
+		 $get_predicate_attribute(Head, AttName, 1)),
+		Ports).
 
 
 		/********************************
@@ -263,30 +285,20 @@ show_trace_ports(_).
 		*********************************/
 
 concat_atom([A, B], C) :- !,
-	concat(A, B, C).
+	atom_concat(A, B, C).
 concat_atom(L, Atom) :-
 	$concat_atom(L, Atom).
 
 term_to_atom(Term, Atom) :-
-	$term_to_atom(Term, Atom, 0, 0).
-
-atom_to_term(Atom, Term, Bindings) :-
-	$term_to_atom(Term, Atom, B0, 0),
-	Bindings = B0.
+	atom_to_term(Atom, Term, 0).
 
 int_to_atom(Int, Atom) :-
 	int_to_atom(Int, 10, Atom).
 
-gensym(Base, Atom) :-
-	concat($gs_, Base, Key),
-	flag(Key, Old, Old),
-	succ(Old, New),
-	flag(Key, _, New),
-	concat(Base, New, Atom).
-
 dwim_match(A1, A2) :-
 	dwim_match(A1, A2, _).
 
+
 		/********************************
 		*             SOURCE            *
 		*********************************/
@@ -312,7 +324,8 @@ prolog_load_context(module, Module) :-
 prolog_load_context(file, F) :-
 	source_location(F, _).
 prolog_load_context(stream, S) :-
-	current_input(S).
+	flag($load_input, S, S),
+	S \== -1.
 prolog_load_context(directory, D) :-
 	source_location(F, _),
 	file_directory_name(F, D).
@@ -378,7 +391,7 @@ current_predicate(Name, Term) :-
 	$c_current_predicate(Name, DefModule:Head),
 	$defined_predicate(DefModule:Head), !.
 current_predicate(Name, Term) :-
-	flag($enable_autoload, on, on),
+	current_prolog_flag(autoload, true),
 	$strip_module(Term, Module, Head),
 	functor(Head, Name, Arity),
 	$find_library(Module, Name, Arity, _LoadModule, _Library), !.
@@ -436,6 +449,8 @@ $predicate_property(line_count(LineNumber), Pred) :-
 	$get_predicate_attribute(Pred, line_count, LineNumber).
 $predicate_property(notrace, Pred) :-
 	$get_predicate_attribute(Pred, trace, 0).
+$predicate_property(spying, Pred) :-
+	$get_predicate_attribute(Pred, spy, 1).
 $predicate_property(show_childs, Pred) :-
 	$get_predicate_attribute(Pred, system, 1),
 	$get_predicate_attribute(Pred, hide_childs, 0).
@@ -444,10 +459,10 @@ $predicate_property(hashed(N), Pred) :-
 	N > 0.
 $predicate_property(references(N), Pred) :-
 	$get_predicate_attribute(Pred, references, N),
-	N > 0.
+	N \== 0.			% show negative for debugging!
 $predicate_property(number_of_clauses(N), Pred) :-
 	$get_predicate_attribute(Pred, number_of_clauses, N),
-	N > 0.
+	N \== 0.
 
 :- index(clause_property(0, 1)).
 
@@ -461,30 +476,6 @@ clause_property(Clause, erased) :-
 	$get_clause_attribute(Clause, erased, true).
 
 
-:- module_transparent
-	clause/2,
-	clause/3.
-
-clause(Head, Body, Ref) :-
-	nonvar(Ref), !,
-	$clause(Head, Clause, Ref),
-	$strip_module(Head, _, H),
-	$clause2(H, Body, Clause).
-clause(Head, Body, Ref) :-
-	current_predicate(_, Head),
-	$clause(Head, Clause, Ref),
-	$strip_module(Head, _, H),
-	$clause2(H, Body, Clause).
-
-clause(Head, Body) :-
-	current_predicate(_, Head),
-	$clause(Head, Clause, _),
-	$strip_module(Head, _, H),
-	$clause2(H, Body, Clause).
-
-$clause2(Head, Body, (Head :- Body)) :- !.
-$clause2(Head, true, Head).
-
 recorda(Key, Value) :-
 	recorda(Key, Value, _).
 recordz(Key, Value) :-
@@ -504,9 +495,9 @@ require([N/A|T]) :- !,
 	functor(Head, N, A),
 	$require(Head),
 	require(T).
-require([H|T]) :-
-	$warning('require/1: Illegal predicate specifier: ~w', [H]),
-	require(T).
+require([H|_T]) :-
+	throw(error(type_error(predicate_indicator, H), _)).
+
 
 		/********************************
 		*            MODULES            *
@@ -524,8 +515,8 @@ module(Module) :-
 	current_module(Module), !,
 	$module(_, Module).
 module(Module) :-
-	$break($warning('~w is not a current module', [Module])),
-	$module(_, Module).
+	$module(_, Module),
+	print_message(warning, no_current_module(Module)).
 
 		/********************************
 		*          STATISTICS           *
@@ -566,11 +557,13 @@ statistics :-
 	       [LocalLimit, Local, LocalUsed]),
 	format('Global stack :~t~D~28| ~t~D~41| ~t~D~54| Bytes~n',
 	       [GlobalLimit, Global, GlobalUsed]),
-	format('Trail  stack :~t~D~28| ~t~D~41| ~t~D~54| Bytes~n',
+	format('Trail  stack :~t~D~28| ~t~D~41| ~t~D~54| Bytes~n~n',
 	       [TrailLimit, Trail, TrailUsed]),
 
 	gc_statistics,
-	shift_statistics.
+	agc_statistics,
+	shift_statistics,
+	thread_statistics.
 
 gc_statistics :-
 	statistics(collections, Collections),
@@ -578,10 +571,19 @@ gc_statistics :-
 	statistics(collected, Collected),
 	statistics(gctime, GcTime),
 
-	format('~n~D garbage collections gained ~D bytes in ~2f seconds.~n',
+	format('~D garbage collections gained ~D bytes in ~2f seconds.~n',
 	       [Collections, Collected, GcTime]).
 gc_statistics.
 
+agc_statistics :-
+	catch(statistics(agc, Agc), _, fail),
+	Agc > 0, !,
+	statistics(agc_gained, Gained),
+	statistics(agc_time, Time),
+	format('~D atom garbage collections gained ~D atoms in ~2f seconds.~n',
+	       [Agc, Gained, Time]).
+agc_statistics.
+
 shift_statistics :-
 	statistics(local_shifts, LS),
 	statistics(global_shifts, GS),
@@ -590,10 +592,20 @@ shift_statistics :-
 	;   GS > 0
 	;   TS > 0
 	), !,
-	format('~nStack shifts: ~D local, ~D global, ~D trail.~n',
+	format('Stack shifts: ~D local, ~D global, ~D trail.~n',
 	       [LS, GS, TS]).
 shift_statistics.
 
+thread_statistics :-
+	current_prolog_flag(threads, true), !,
+	statistics(threads, Active),
+	statistics(threads_created, Created),
+	statistics(thread_cputime, CpuTime),
+	Finished is Created - Active,
+	format('~D threads, ~D finished threads used ~2f seconds.~n',
+	       [Active, Finished, CpuTime]).
+thread_statistics.
+
 
 		/********************************
 		*      SYSTEM INTERACTION       *
@@ -629,13 +641,17 @@ on_signal(Signal, _Old, _New) :-
 
 current_signal(Name, Id, Handler) :-
 	between(1, 32, Id),
-	$on_signal(Name, Id, Handler, Handler).
+	$on_signal(Id, Name, Handler, Handler).
 
 
 		 /*******************************
 		 *	      DLOPEN		*
 		 *******************************/
 
+:- module_transparent
+	open_shared_object/2,
+	open_shared_object/3.
+
 dlopen_flag(now,	2'01).		% see pl-load.c for these constants
 dlopen_flag(global,	2'10).		% Solaris only
 
@@ -669,45 +685,6 @@ sformat(String, Format) :-
 		 *	      FILES		*
 		 *******************************/
 
-%	absolute_file_name(+Term, +Args, -AbsoluteFile)
-
-absolute_file_name(Spec, Args, Path) :-
-	(   select(Args, extensions(Exts), Conditions)
-	->  true
-	;   select(Args, file_type(Type), Rest)
-	->  file_type_conditions(Type, Exts, C0),
-	    append(C0, Rest, Conditions)
-	;   Conditions = Args,
-	    Exts = ['']
-	),
-	(   select(Conditions, solutions(Sols), C1)
-	->  true
-	;   Sols = first,
-	    C1 = Conditions
-	),
-	(   select(C1, file_errors(FileErrors), C2)
-	->  true
-	;   FileErrors = fail,
-	    C2 = C1
-	),
-	(   $chk_file(Spec, Exts, C2, Path)
-	->  (   Sols == first
-	    ->  !
-	    ;   true
-	    )
-	;   (   FileErrors == fail
-	    ->  fail
-	    ;   $warning('~w: No such file or directory', Spec),
-		fail
-	    )
-	).
-
-file_type_conditions(txt,        [''],		[]).
-file_type_conditions(prolog,     ['.pl', ''],	[]).
-file_type_conditions(executable, ['.so', ''],	[]).
-file_type_conditions(qlf, 	 ['.qlf', ''],	[]).
-file_type_conditions(directory,  [''],		[]).
-
 %	absolute_file_name(+Term, -AbsoluteFile)
 
 absolute_file_name(Name, Abs) :-
