@@ -1,9 +1,32 @@
-/*  $Id: help.pl,v 1.11 1997/10/13 10:07:58 jan Exp $
+/*  $Id: help.pl,v 1.18 2002/02/01 16:49:11 jan Exp $
 
-    Copyright (c) 1990 Jan Wielemaker. All rights reserved.
-    jan@swi.psy.uva.nl
+    Part of SWI-Prolog
 
-    Purpose: Give online help
+    Author:        Jan Wielemaker
+    E-mail:        jan@swi.psy.uva.nl
+    WWW:           http://www.swi-prolog.org
+    Copyright (C): 1985-2002, University of Amsterdam
+
+    This program is free software; you can redistribute it and/or
+    modify it under the terms of the GNU General Public License
+    as published by the Free Software Foundation; either version 2
+    of the License, or (at your option) any later version.
+
+    This program is distributed in the hope that it will be useful,
+    but WITHOUT ANY WARRANTY; without even the implied warranty of
+    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+    GNU General Public License for more details.
+
+    You should have received a copy of the GNU Lesser General Public
+    License along with this library; if not, write to the Free Software
+    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+
+    As a special exception, if you link this library with other files,
+    compiled with a Free Software compiler, to produce an executable, this
+    library does not by itself cause the resulting executable to be covered
+    by the GNU General Public License. This exception does not however
+    invalidate any other reasons why the executable file might be covered by
+    the GNU General Public License.
 */
 
 :- module(online_help,
@@ -14,31 +37,37 @@
 
 :- use_module(library(helpidx)).
 
+:- multifile
+	prolog:help_hook/1.		% Generic help hook.
+
 /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 This module  defines the  online  help  facility of   SWI-Prolog.   It
 assumes  (Prolog) index file  at library(help_index)   and  the actual
 manual  at library(online_manual).   Output  is piped through  a  user
 defined pager, which defaults to `more'.
-
-BUGS:
-If the pager  is quit prematurely Prolog will  abort  on  noticing the
-broken pipe.
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
 %	help/0
 
+help :-
+	prolog:help_hook(help), !.
 help :-
 	help(help/1).
 
 %	help(+Subject)
+%
 %	Display online help on specified subject.
 
+help(What) :-
+	prolog:help_hook(help(What)), !.
 help(What) :-
 	give_help(What).
 
 %	apropos(Pattern)
 %	Give a list of subjects that might be appropriate.
 
+apropos(What) :-
+	prolog:help_hook(apropos(What)), !.
 apropos(What) :-
 	give_apropos(What).
 
@@ -90,27 +119,15 @@ show_help(_, Ranges) :-
 	write_ranges_to_file(Ranges, Outfile),
 	call_emacs('(view-file-other-window "~w")', [Outfile]).
 show_help(_, Ranges) :-
-	\+ feature(pipe, true), !,
-	online_manual_stream(Manual),
-	show_ranges(Ranges, Manual, user_output).
-show_help(_, [Start-End]) :-
-	End - Start > 4000, !,
-	find_manual(Manual),
-	find_pager(Pager),
-	plbite_flags(Flags),
-	sformat(Cmd, 'pl-bite ~w ~d:~d ~a | ~a',
-		[Flags, Start, End, Manual, Pager]),
-	shell(Cmd).	
-show_help(_, Ranges) :-
+	current_prolog_flag(pipe, true), !,
 	online_manual_stream(Manual),
 	pager_stream(Pager),
-	show_ranges(Ranges, Manual, Pager),
+	catch(show_ranges(Ranges, Manual, Pager), _, true),
 	close(Manual),
-	close(Pager).
-
-plbite_flags('-e') :-
-	feature(write_help_with_overstrike, true), !.
-plbite_flags('').
+	catch(close(Pager), _, true).
+show_help(_, Ranges) :-
+	online_manual_stream(Manual),
+	show_ranges(Ranges, Manual, user_output).
 
 show_ranges([], _, _) :- !.
 show_ranges([From-To|Rest], Manual, Pager) :-
@@ -124,7 +141,8 @@ copy_chars(N, From, To) :-
 	get0(From, C0),
 	copy_chars(N, From, To, C0).
 
-copy_chars(0, _, _, _) :- !.
+copy_chars(N, _, _, _) :-
+	N =< 0, !.
 copy_chars(N, _, To, _) :-
 	0 =:= N mod 4096,
 	flush_output(To),
@@ -132,9 +150,9 @@ copy_chars(N, _, To, _) :-
 copy_chars(N, From, To, C) :-
 	get0(From, C1),
 	(   C1 == 8,			% backspace
-	    \+ feature(write_help_with_overstrike, true)
+	    \+ current_prolog_flag(write_help_with_overstrike, true)
 	->  get0(From, C2),
-	    NN is N - 3,
+	    NN is N - 2,
 	    copy_chars(NN, From, To, C2)
 	;   put_printable(To, C),
 	    NN is N - 1,
@@ -166,12 +184,12 @@ Set the write_help_with_overstrike feature.
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
 set_overstrike_feature :-
-	feature(write_help_with_overstrike, _), !.
+	current_prolog_flag(write_help_with_overstrike, _), !.
 set_overstrike_feature :-
 	getenv('TERM', xterm), !,
-	set_feature(write_help_with_overstrike, true).
+	set_prolog_flag(write_help_with_overstrike, true).
 set_overstrike_feature :-
-	set_feature(write_help_with_overstrike, false).
+	set_prolog_flag(write_help_with_overstrike, false).
 	
 :- initialization set_overstrike_feature.
 
