@@ -1,7 +1,32 @@
-/*  File:    shell.pl
-    Purpose: Limited Unix Shell Emulation
-    Author:  Jan Wielemaker
-    Date:    Sep 16 1989
+/*  $Id: shell.pl,v 1.16 2002/02/01 16:49:11 jan Exp $
+
+    Part of SWI-Prolog
+
+    Author:        Jan Wielemaker
+    E-mail:        jan@swi.psy.uva.nl
+    WWW:           http://www.swi-prolog.org
+    Copyright (C): 1985-2002, University of Amsterdam
+
+    This program is free software; you can redistribute it and/or
+    modify it under the terms of the GNU General Public License
+    as published by the Free Software Foundation; either version 2
+    of the License, or (at your option) any later version.
+
+    This program is distributed in the hope that it will be useful,
+    but WITHOUT ANY WARRANTY; without even the implied warranty of
+    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+    GNU General Public License for more details.
+
+    You should have received a copy of the GNU Lesser General Public
+    License along with this library; if not, write to the Free Software
+    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+
+    As a special exception, if you link this library with other files,
+    compiled with a Free Software compiler, to produce an executable, this
+    library does not by itself cause the resulting executable to be covered
+    by the GNU General Public License. This exception does not however
+    invalidate any other reasons why the executable file might be covered by
+    the GNU General Public License.
 */
 
 :- module(shell,
@@ -40,7 +65,7 @@
 
 cd(Dir) :-
 	name_to_atom(Dir, Name),
-	chdir(Name).
+	working_directory(_, Name).
 
 %	dirs	-- Print Directory Stack
 %	pushd	-- Push Directory Stack
@@ -58,20 +83,19 @@ pushd(N) :-
 	(   nth1(N, Ds, Go),
 	    retract(stack(Go))
 	->  pushd(Go)
-	;   warning('Directory stack not that deep'),
+	;   warning('Directory stack not that deep', []),
 	    fail
 	).
 pushd(Dir) :-
 	name_to_atom(Dir, Name),
-	absolute_file_name('', Old),
-	chdir(Name),
+	working_directory(Old, Name),
 	asserta(stack(Old)).
 
 popd :-
 	retract(stack(Dir)), !,
-	chdir(Dir).
+	working_directory(_, Dir).
 popd :-
-	warning('Directory stack empty'),
+	warning('Directory stack empty', []),
 	fail.
 
 dirs :-
@@ -91,16 +115,16 @@ pwd :-
 
 dir_name('/', '/') :- !.
 dir_name(Path, Name) :-
-	concat(P, /, Path), !,
+	atom_concat(P, /, Path), !,
 	dir_name(P, Name).
 dir_name(Path, Name) :-
-	feature(unix, true),
+	current_prolog_flag(unix, true),
 	absolute_file_name('~', Home0),
-	(   concat(Home, /, Home0)
+	(   atom_concat(Home, /, Home0)
 	->  true
 	;   Home = Home0
 	),
-	concat(Home, FromHome, Path), !,
+	atom_concat(Home, FromHome, Path), !,
 	sformat(Name, '~~~w', [FromHome]).
 dir_name(Path, Path).
 
@@ -118,16 +142,14 @@ ls(Spec) :-
 
 ls_([Dir]) :-
 	exists_directory(Dir), !,
-	absolute_file_name('', Here),
-	chdir(Dir),
+	working_directory(Here, Dir),
 	expand_file_name('*', Files),
-	chdir(Here),
-	ls__(Files).
+	call_cleanup(ls__(Files), working_directory(_, Here)).
 ls_(Files) :-
 	ls__(Files).
 
 ls__([]) :- !,
-	warning('No Match'),
+	warning('No Match', []),
 	fail.
 ls__(Files) :-
 	maplist(tag_file, Files, Tagged),
@@ -135,7 +157,7 @@ ls__(Files) :-
 
 tag_file(File, Dir) :-
 	exists_directory(File),	
-	concat(File, /, Dir).
+	atom_concat(File, /, Dir).
 tag_file(File, File).
 
 %	mv(+From, +To)	--- Move (Rename) a file
@@ -150,21 +172,25 @@ rm(File) :-
 	name_to_atom(File, A),
 	delete_file(A).
 
+
 %	name_to_atom(Typed, Atom)
+%
 %	Convert a typed name into an atom
 
-name_to_atom(Atom, Atom) :-
-	atomic(Atom), !.
-name_to_atom(Term, Atom) :-
-	term_to_atom(Term, Raw),
-	name(Raw, S0),
-	sublist(non_blank, S0, S1),
-	name(Atom, S1).
-
-non_blank(C) :-
-	between(0, 32, C), !,
-	fail.
-non_blank(_).
+name_to_atom(Spec, File) :-
+	(   atomic(Spec)
+	->  S1 = Spec
+	;   sformat(S1, '~w', [Spec])
+	),
+	expand_file_name(Spec, Expanded),
+	(   Expanded = [File]
+	->  true
+	;   Expanded == []
+	->  print_message(warning, format('No match: ~w', [Spec])),
+	    fail
+	;   print_message(warning, format('Ambiguous: ~w', [Spec])),
+	    fail
+	).
 
 
 %	list_atoms(+List, +Width)
@@ -211,8 +237,5 @@ longest([_|T], S, M) :-
 
 %	warning(Fmt, [Args]).
 
-warning(Fmt) :-
-	warning(Fmt, []).
-
 warning(Fmt, Args) :-
-	'$break'('$warning'(Fmt, Args)).
+	print_message(warning, format(Fmt, Args)).
