@@ -1,10 +1,25 @@
-/*  $Id: pl-arith.c,v 1.34 1997/08/07 07:57:34 jan Exp $
+/*  $Id: pl-arith.c,v 1.52 2002/02/01 16:49:12 jan Exp $
 
-    Copyright (c) 1990 Jan Wielemaker. All rights reserved.
-    See ../LICENCE to find out about your rights.
-    jan@swi.psy.uva.nl
+    Part of SWI-Prolog
 
-    Purpose: arithmetic built in functions
+    Author:        Jan Wielemaker
+    E-mail:        jan@swi.psy.uva.nl
+    WWW:           http://www.swi-prolog.org
+    Copyright (C): 1985-2002, University of Amsterdam
+
+    This library is free software; you can redistribute it and/or
+    modify it under the terms of the GNU Lesser General Public
+    License as published by the Free Software Foundation; either
+    version 2.1 of the License, or (at your option) any later version.
+
+    This library is distributed in the hope that it will be useful,
+    but WITHOUT ANY WARRANTY; without even the implied warranty of
+    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
+    Lesser General Public License for more details.
+
+    You should have received a copy of the GNU Lesser General Public
+    License along with this library; if not, write to the Free Software
+    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
 
 /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
@@ -28,6 +43,10 @@ day.
 
 #include <math.h>			/* avoid abs() problem with MSVC++ */
 #include "pl-incl.h"
+
+#undef LD
+#define LD LOCAL_LD
+
 #ifndef M_PI
 #define M_PI (3.14159265358979323846)
 #endif
@@ -40,10 +59,35 @@ day.
 #define HAVE_ISNAN
 #endif
 
-#ifdef WIN32
+#ifdef HAVE___TRY
 #include <excpt.h>
 #endif
 
+/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
+On some machines, notably  FreeBSD  upto   version  3.x,  floating point
+operations raise signals rather then leaving an error condition and this
+behaviour can be changed to be   IEEE754  using fpsetmask() and friends.
+Here  we  test  whether  this  interface  is   present  and  set  it  up
+accordingly.
+
+With many thanks to NIDE  Naoyuki  for   the  clear  explanation  of the
+problem.
+- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
+
+#if defined(HAVE_FLOATINGPOINT_H) && defined(HAVE_FPSETMASK) && defined(HAVE_FPRESETSTICKY)
+#define O_INHIBIT_FP_SIGNALS
+#include <floatingpoint.h>
+#ifndef FP_X_DZ
+#define FP_X_DZ 0
+#endif
+#ifndef FP_X_INV
+#define FP_X_INV 0
+#endif
+#ifndef FP_X_OFL
+#define FP_X_OFL 0
+#endif
+#endif
+
 typedef int (*ArithF)();
 
 struct arithFunction
@@ -215,12 +259,14 @@ ar_compare(Number n1, Number n2, int what)
 
 static word
 compareNumbers(term_t n1, term_t n2, int what)
-{ number left, right;
+{ GET_LD
+  number left, right;
 
-  TRY(valueExpression(n1, &left) &&
-      valueExpression(n2, &right));
+  if ( valueExpression(n1, &left PASS_LD) &&
+       valueExpression(n2, &right PASS_LD) )
+    return ar_compare(&left, &right, what);
 
-  return ar_compare(&left, &right, what);
+  fail;
 }
 
 
@@ -274,6 +320,7 @@ isCurrentArithFunction(functor_t f, Module m)
       { if ( m2 == a->module && l < level )
 	{ r = a;
 	  level = l;
+	  break;
 	}
       }
     }
@@ -282,42 +329,6 @@ isCurrentArithFunction(functor_t f, Module m)
   return r;
 }
 
-#if HAVE_SIGNAL
-typedef void (*OsSigHandler)(int);
-
-static void
-realExceptionHandler(int sig, int type, SignalContext scp, char *addr)
-{
-#ifndef BSD_SIGNALS
-  signal(sig, (OsSigHandler)realExceptionHandler);
-#endif
-  if ( LD->in_arithmetic > 0 )
-  { warning("Floating point exception");
-#ifndef O_RUNTIME
-    Sfprintf(Serror, "[PROLOG STACK:\n");
-    backTrace(NULL, 10);
-    Sfprintf(Serror, "]\n");
-#endif
-    pl_abort();
-  } else
-  { deliverSignal(sig, type, scp, addr);
-  }
-}
-#endif
-
-#if __TURBOC__
-static int
-realExceptionHandler(e)
-struct exception *e;
-{ warning("Floating point exception");
-
-  pl_abort();
-  /*NOTREACHED*/
-  fail;				/* make tc happy */
-}
-#endif
-
-
 #if O_PROLOG_FUNCTIONS
 
 static int prologFunction(ArithFunction, term_t, Number);
@@ -342,7 +353,8 @@ prologFunction(ArithFunction f, term_t av, Number r)
   qid = PL_open_query(NULL, PL_Q_CATCH_EXCEPTION, f->proc, av);
 
   if ( PL_next_solution(qid) )
-  { rval = valueExpression(av+arity-1, r);
+  { GET_LD
+    rval = valueExpression(av+arity-1, r PASS_LD);
     PL_close_query(qid);
     PL_discard_foreign_frame(fid);
   } else
@@ -366,7 +378,7 @@ prologFunction(ArithFunction f, term_t av, Number r)
 #endif /* O_PROLOG_FUNCTIONS */
 
 int
-valueExpression(term_t t, Number r)
+valueExpression(term_t t, Number r ARG_LD)
 { ArithFunction f;
   functor_t functor;
   Word p = valTermRef(t);
@@ -433,8 +445,8 @@ valueExpression(term_t t, Number r)
     for(n=0; n<arity; n++)
     { number n1;
 
-      PL_get_arg(n+1, t, h0+n);
-      if ( valueExpression(h0+n, &n1) )
+      _PL_get_arg(n+1, t, h0+n);
+      if ( valueExpression(h0+n, &n1 PASS_LD) )
       { _PL_put_number(h0+n, &n1);
       } else
 	fail;
@@ -446,9 +458,14 @@ valueExpression(term_t t, Number r)
   }
 #endif
 
+  DEBUG(3, Sdprintf("Starting __try ...\n"));
+
   { int rval;
 
-#ifdef WIN32
+#ifdef HAVE___TRY
+#ifndef EXCEPTION_EXECUTE_HANDLER	/* lcc */
+#define EXCEPTION_EXECUTE_HANDLER 1
+#endif
     __try
     {
 #else
@@ -462,8 +479,8 @@ valueExpression(term_t t, Number r)
       { term_t a = PL_new_term_ref();
 	number n1;
 
-	PL_get_arg(1, t, a);
-	if ( valueExpression(a, &n1) )
+	_PL_get_arg_ld(1, t, a PASS_LD);
+	if ( valueExpression(a, &n1 PASS_LD) )
 	  rval = (*f->function)(&n1, r);
 	else
 	  rval = FALSE;
@@ -475,10 +492,10 @@ valueExpression(term_t t, Number r)
       { term_t a = PL_new_term_ref();
 	number n1, n2;
 
-	PL_get_arg(1, t, a);
-	if ( valueExpression(a, &n1) )
-	{ PL_get_arg(2, t, a);
-	  if ( valueExpression(a, &n2) )
+	_PL_get_arg_ld(1, t, a PASS_LD);
+	if ( valueExpression(a, &n1 PASS_LD) )
+	{ _PL_get_arg_ld(2, t, a PASS_LD);
+	  if ( valueExpression(a, &n2 PASS_LD) )
 	    rval = (*f->function)(&n1, &n2, r);
 	  else
 	    rval = FALSE;
@@ -492,7 +509,7 @@ valueExpression(term_t t, Number r)
 	sysError("Illegal arity for arithmic function");
         rval = FALSE;
     }
-#if defined(WIN32)
+#if defined(HAVE___TRY)
     } __except(EXCEPTION_EXECUTE_HANDLER)
     { warning("Floating point exception");
 #ifndef O_RUNTIME
@@ -500,17 +517,22 @@ valueExpression(term_t t, Number r)
       backTrace(NULL, 10);
       Sfprintf(Serror, "]\n");
 #endif
-      pl_abort();
+      pl_abort(ABORT_NORMAL);
     }
-#else
+#else /*HAVE___TRY*/
     LD->in_arithmetic--;
-#endif
+#endif /*HAVE___TRY*/
 
     if ( r->type == V_REAL )
     {
-#ifdef HUGE_VAL
-      if ( r->value.f == HUGE_VAL )
+#ifdef DBL_MAX
+      if ( r->value.f > DBL_MAX || r->value.f < -DBL_MAX ) 
 	return PL_error(NULL, 0, NULL, ERR_AR_OVERFLOW);
+#else
+#ifdef HUGE_VAL
+      if ( r->value.f == HUGE_VAL || r->value.f == -HUGE_VAL )
+	return PL_error(NULL, 0, NULL, ERR_AR_OVERFLOW);
+#endif
 #endif
 #ifdef HAVE_ISNAN
       if ( isnan(r->value.f) )
@@ -561,18 +583,22 @@ toIntegerNumber(Number n)
 
 void
 canoniseNumber(Number n)
-{ if ( n->type == V_REAL )		/* only if not explicit! */
-  { long l;
+{					/* only if not explicit! */
+  if ( n->type == V_REAL )
+  { GET_LD
+    if ( !trueFeature(ISO_FEATURE) )
+    { long l;
 
 #ifdef DOUBLE_TO_LONG_CAST_RAISES_SIGFPE
-    if ( !((n->value.f >= PLMININT) && (n->value.f <= PLMAXINT)) )
-      return;
+      if ( !((n->value.f >= PLMININT) && (n->value.f <= PLMAXINT)) )
+        return;
 #endif
 
-    l = (long)n->value.f;
-    if ( n->value.f == (real) l )
-    { n->value.i = l;
-      n->type = V_INTEGER;
+      l = (long)n->value.f;
+      if ( n->value.f == (real) l )
+      { n->value.i = l;
+	n->type = V_INTEGER;
+      }
     }
   }
 }
@@ -584,7 +610,9 @@ canoniseNumber(Number n)
 
 static int
 ar_add(Number n1, Number n2, Number r)
-{ if ( intNumber(n1) && intNumber(n2) ) 
+{ GET_LD
+
+  if ( intNumber(n1) && intNumber(n2) ) 
   { r->value.i = n1->value.i + n2->value.i; 
     
     if ( n1->value.i > 0 && n2->value.i > 0 && r->value.i <= 0 )
@@ -594,9 +622,13 @@ ar_add(Number n1, Number n2, Number r)
 
     r->type = V_INTEGER;
     succeed;
-  } 
 
 overflow:
+    if ( trueFeature(ISO_FEATURE) )
+      return PL_error("+", 2, NULL, ERR_EVALUATION, ATOM_int_overflow);
+  }
+
+  
   promoteToRealNumber(n1);
   promoteToRealNumber(n2);
   r->value.f = n1->value.f + n2->value.f; 
@@ -608,7 +640,9 @@ overflow:
 
 static int
 ar_minus(Number n1, Number n2, Number r)
-{ if ( intNumber(n1) && intNumber(n2) ) 
+{ GET_LD
+
+  if ( intNumber(n1) && intNumber(n2) ) 
   { r->value.i = n1->value.i - n2->value.i; 
     
     if ( n1->value.i > 0 && n2->value.i < 0 && r->value.i <= 0 )
@@ -618,9 +652,12 @@ ar_minus(Number n1, Number n2, Number r)
 
     r->type = V_INTEGER;
     succeed;
-  } 
 
 overflow:
+    if ( trueFeature(ISO_FEATURE) )
+      return PL_error("-", 2, NULL, ERR_EVALUATION, ATOM_int_overflow);
+  } 
+
   promoteToRealNumber(n1);
   promoteToRealNumber(n2);
   r->value.f = n1->value.f - n2->value.f; 
@@ -629,6 +666,22 @@ overflow:
   succeed;
 }
 
+static int
+ar_mod(Number n1, Number n2, Number r)
+{ if ( !toIntegerNumber(n1) )
+    return PL_error("mod", 2, NULL, ERR_AR_TYPE, ATOM_integer, n1);
+
+  if ( !toIntegerNumber(n2) )
+    return PL_error("mod", 2, NULL, ERR_AR_TYPE, ATOM_integer, n2);
+
+  if ( n2->value.i == 0 )
+    return PL_error("mod", 2, NULL, ERR_DIV_BY_ZERO);
+
+  r->value.i = n1->value.i % n2->value.i;
+  r->type = V_INTEGER;
+
+  succeed;
+}
 
 /* Unary functions requiring double argument */
 
@@ -674,7 +727,6 @@ UNAIRY_FLOAT_FUNCTION(ar_exp, exp)
 BINAIRY_FLOAT_FUNCTION(ar_atan2, atan2)
 BINAIRY_FLOAT_FUNCTION(ar_pow, pow)
 
-BINAIRY_INT_FUNCTION(ar_mod, "mod", %)
 BINAIRY_INT_FUNCTION(ar_disjunct, "\\/", |)
 BINAIRY_INT_FUNCTION(ar_conjunct, "/\\", &)
 BINAIRY_INT_FUNCTION(ar_shift_right, ">>", >>)
@@ -781,7 +833,9 @@ ar_rem(Number n1, Number n2, Number r)
 
 static int
 ar_divide(Number n1, Number n2, Number r)
-{ if ( intNumber(n1) && intNumber(n2) )
+{ GET_LD
+
+  if ( !trueFeature(ISO_FEATURE) && (intNumber(n1) && intNumber(n2)) )
   { if ( n2->value.i == 0 )
       return PL_error("/", 2, NULL, ERR_DIV_BY_ZERO);
 
@@ -838,6 +892,9 @@ ar_max(Number n1, Number n2, Number r)
 
   r->value.f = (n1->value.f > n2->value.f ? n1->value.f : n2->value.f);
   r->type = V_REAL;
+
+  (void)toIntegerNumber(r);
+
   succeed;
 }
 
@@ -855,6 +912,9 @@ ar_min(Number n1, Number n2, Number r)
 
   r->value.f = (n1->value.f < n2->value.f ? n1->value.f : n2->value.f);
   r->type = V_REAL;
+
+  (void)toIntegerNumber(r);
+
   succeed;
 }
 
@@ -943,6 +1003,8 @@ ar_floor(Number n1, Number r)
 #ifdef HAVE_FLOOR
     r->value.f = floor(n1->value.f);
     r->type = V_REAL;
+    if ( !toIntegerNumber(r) )
+      return PL_error("floor", 1, NULL, ERR_EVALUATION, ATOM_int_overflow);
 #else
     r->value.i = (long)n1->value.f;
     if ( n1->value.f < 0 && (real)r->value.i != n1->value.f )
@@ -963,10 +1025,12 @@ ar_ceil(Number n1, Number r)
 #ifdef HAVE_CEIL
     r->value.f = ceil(n1->value.f);
     r->type = V_REAL;
+    if ( !toIntegerNumber(r) )
+      return PL_error("ceil", 1, NULL, ERR_EVALUATION, ATOM_int_overflow);
 #else
     r->value.i = (long)n1->value.f;
     if ( (real)r->value.i < n1->value.f )
-       r->value.i)++;
+       r->value.i++;
     r->type = V_INTEGER;
 #endif
   }
@@ -984,8 +1048,7 @@ ar_float_fractional_part(Number n1, Number r)
   { if ( n1->value.f > 0 )
     { r->value.f = n1->value.f - floor(n1->value.f);
     } else
-    { TRY(ar_ceil(n1, r));
-      r->value.f = n1->value.f - ceil(n1->value.f);
+    { r->value.f = n1->value.f - ceil(n1->value.f);
     }
     r->type = V_REAL;
   }
@@ -1020,6 +1083,13 @@ ar_random(Number n1, Number r)
 { if ( !toIntegerNumber(n1) )
     return PL_error("random", 1, NULL, ERR_AR_TYPE, ATOM_integer, n1);
 
+  if ( n1->value.i < 1 )
+  { term_t i = PL_new_term_ref();
+    PL_put_integer(i, n1->value.i);
+
+    return PL_error("random", 1, NULL, ERR_DOMAIN, ATOM_not_less_than_zero, i);
+  }
+
   r->value.i = Random() % n1->value.i;
   r->type = V_INTEGER;
 
@@ -1047,7 +1117,7 @@ ar_e(Number r)
 
 static int
 ar_cputime(Number r)
-{ r->value.f = CpuTime();
+{ r->value.f = CpuTime(CPU_USER);
 
   r->type = V_REAL;
   succeed;
@@ -1060,11 +1130,16 @@ ar_cputime(Number r)
 
 word
 pl_is(term_t v, term_t e)
-{ number arg;
+{ GET_LD
+  number arg;
 
-  if ( valueExpression(e, &arg) )
+  if ( valueExpression(e, &arg PASS_LD) )
   { canoniseNumber(&arg);
-    return _PL_unify_number(v, &arg);
+
+    if ( intNumber(&arg) )
+      return PL_unify_integer(v, arg.value.i);
+    else
+      return PL_unify_float(v, arg.value.f);
   }
 
   fail;
@@ -1074,8 +1149,9 @@ pl_is(term_t v, term_t e)
 #if O_PROLOG_FUNCTIONS
 word
 pl_arithmetic_function(term_t descr)
-{ Procedure proc;
-  Definition def;
+{ GET_LD
+
+  Procedure proc;
   functor_t fd;
   FunctorDef fdef;
   ArithFunction f;
@@ -1091,7 +1167,6 @@ pl_arithmetic_function(term_t descr)
     return warning("arithmetic_function/1: Illegal arity");
 
   proc = lookupProcedure(fd, m);
-  def = proc->definition;
   fd = lookupFunctorDef(fdef->name, fdef->arity - 1);
   if ( (f = isCurrentArithFunction(fd, m)) && f->module == m )
     succeed;				/* already registered */
@@ -1162,6 +1237,47 @@ pl_current_arithmetic_function(term_t f, word h)
   fail;
 }
 
+word
+pl_prolog_arithmetic_function(term_t f, word h)
+{ GET_LD
+  ArithFunction a;
+  term_t tmp = PL_new_term_ref();
+
+  switch( ForeignControl(h) )
+  { case FRG_FIRST_CALL:
+      a = arithFunctionTable[0];
+      break;
+    case FRG_REDO:
+      a = ForeignContextPtr(h);
+      break;
+    case FRG_CUTTED:
+    default:
+      succeed;
+  }
+
+  for( ; a; a = a->next )
+  { mark m;
+
+    while( isTableRef(a) )
+    { a = unTableRef(ArithFunction, a);
+      if ( !a )
+        fail;
+    }
+
+    Mark(m);
+    PL_put_functor(tmp, a->functor);
+    if ( a->proc &&
+	 PL_unify_term(f,
+		       PL_FUNCTOR, FUNCTOR_module2,
+		         PL_ATOM, a->module->name,
+		         PL_TERM, tmp) )
+      return_next_table(ArithFunction, a, ;);
+    Undo(m);
+  }
+
+  fail;
+}
+
 #endif /* O_PROLOG_FUNCTIONS */
 
 typedef struct
@@ -1263,15 +1379,7 @@ registerBuiltinFunctions()
 
 void
 initArith(void)
-{ 
-#ifdef SIGFPE
-  pl_signal(SIGFPE, (handler_t) realExceptionHandler);
-#endif
-#if __TURBOC__
-  setmatherr(realExceptionHandler);
-#endif
-
-  initBuffer(function_array);
+{ initBuffer(function_array);
 					/* link the table to enumerate */
   { ArithFunction *f;
     int n;
@@ -1281,8 +1389,24 @@ initArith(void)
   }
 
   registerBuiltinFunctions();
+
+#ifdef O_INHIBIT_FP_SIGNALS
+  fpsetmask(fpgetmask() & ~(FP_X_DZ|FP_X_INV|FP_X_OFL));
+#endif
 }
 
+
+void
+cleanupArith(void)
+{ discardBuffer(function_array);
+
+#ifdef O_INHIBIT_FP_SIGNALS
+  fpresetsticky(FP_X_DZ|FP_X_INV|FP_X_OFL);
+  fpsetmask(FP_X_DZ|FP_X_INV|FP_X_OFL);
+#endif
+}
+
+
 #if O_COMPILE_ARITH
 
 		/********************************
@@ -1290,7 +1414,7 @@ initArith(void)
 		*********************************/
 
 int
-indexArithFunction(functor_t fdef, register Module m)
+indexArithFunction(functor_t fdef, Module m)
 { ArithFunction f;
 
   if ( !(f = isCurrentArithFunction(fdef, m)) )
@@ -1315,7 +1439,8 @@ ar_func_n(code n, int argc, Number *stack)
 
   sp -= argc;
   if ( f->proc )
-  { LocalFrame lSave = lTop;		/* TBD (check with stack!) */
+  { GET_LD
+    LocalFrame lSave = lTop;		/* TBD (check with stack!) */
     term_t h0;
     int n;
 
@@ -1347,9 +1472,14 @@ ar_func_n(code n, int argc, Number *stack)
   if ( rval )
   { if ( result.type == V_REAL )
     {
-#ifdef HUGE_VAL
-      if ( result.value.f == HUGE_VAL )
+#ifdef DBL_MAX
+      if ( result.value.f > DBL_MAX || result.value.f < -DBL_MAX ) 
 	return PL_error(NULL, 0, NULL, ERR_AR_OVERFLOW);
+#else
+#ifdef HUGE_VAL
+      if ( result.value.f == HUGE_VAL || result.value.f == -HUGE_VAL )
+	return PL_error(NULL, 0, NULL, ERR_AR_OVERFLOW);
+#endif
 #endif
 #ifdef HAVE_ISNAN
       if ( isnan(result.value.f) )
