@@ -1,10 +1,25 @@
-/*  $Id: pl-global.h,v 1.5 1998/08/06 13:27:07 jan Exp $
+/*  $Id: pl-global.h,v 1.35 2002/02/01 16:49:13 jan Exp $
 
     Part of SWI-Prolog
-    Designed and implemented by Jan Wielemaker
-    E-mail: jan@swi.psy.uva.nl
 
-    Copyright (C) 1997 University of Amsterdam. All rights reserved.
+    Author:        Jan Wielemaker
+    E-mail:        jan@swi.psy.uva.nl
+    WWW:           http://www.swi-prolog.org
+    Copyright (C): 1985-2002, University of Amsterdam
+
+    This library is free software; you can redistribute it and/or
+    modify it under the terms of the GNU Lesser General Public
+    License as published by the Free Software Foundation; either
+    version 2.1 of the License, or (at your option) any later version.
+
+    This library is distributed in the hope that it will be useful,
+    but WITHOUT ANY WARRANTY; without even the implied warranty of
+    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
+    Lesser General Public License for more details.
+
+    You should have received a copy of the GNU Lesser General Public
+    License along with this library; if not, write to the Free Software
+    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
 
 #ifndef PL_GLOBAL_H_INCLUDED
@@ -41,6 +56,11 @@ typedef struct
 #endif
 } PL_code_data_t;
 
+typedef struct
+{ atom_t	file;			/* current source file */
+  int	  	line;			/* current line */
+  long		character;		/* current character location */
+} source_location;
 
 		 /*******************************
 		 *	    GLOBAL DATA		*
@@ -51,19 +71,22 @@ typedef struct
   char *base_of_heap;			/* lowest allocated heap address */
   ulong rounded_heap_base;		/* heap-base rounded downwards */
   int	critical;			/* heap is being modified */
-  char *cannot_save_program;		/* why program cannot be saved */
-  pl_loaderstatus_t _loaderstatus;	/* status of foreign code loader */
   pl_defaults_t	    defaults;		/* system default settings */
   pl_options_t	    options;		/* command-line options */
   State		stateList;		/* list of loaded states */
-  ulong		bases[STG_MASK+1];	/* area base addresses */
   int		initialised;		/* Heap is initialised */
   int		io_initialised;		/* I/O system has been initialised */
   int		bootsession;		/* -b boot compilation */
   int		debug_level;		/* Maintenance debugging: 0..9 */
-  Feature	_feature_list;		/* global features */
   void *	resourceDB;		/* program resource database */
 
+#ifdef HAVE_SIGNAL
+  sig_handler sig_handlers[MAXSIGNAL];	/* How Prolog preceives signals */
+#endif  
+#ifdef O_LOGICAL_UPDATE
+  unsigned long generation;		/* generation of the database */
+#endif
+
   struct
   { int		argc;			/* main(int argc, char **argv) */
     char **	argv;
@@ -80,10 +103,19 @@ typedef struct
   struct
   { long	heap;			/* heap in use */
     int		atoms;			/* No. of atoms defined */
+    long	atomspace;		/* # bytes used to store atoms */
+#ifdef O_ATOMGC
+    long	atomspacefreed;		/* Freed atom-space */
+#endif
     int		functors;		/* No. of functors defined */
     int		predicates;		/* No. of predicates defined */
     int		modules;		/* No. of modules in the system */
     long	codes;			/* No. of byte codes generated */
+#ifdef O_PLMT
+    int		threads_created;	/* # threads created */
+    int		threads_finished;	/* # finished threads */
+    real	thread_cputime;		/* Total CPU time of threads */
+#endif
   } statistics;
 
   struct
@@ -93,6 +125,7 @@ typedef struct
 
   struct
   { Table	modules;		/* atom --> module */
+    Table	record_lists;		/* Available record lists */
   } tables;
 
   struct
@@ -103,10 +136,21 @@ typedef struct
   struct
   { buffer	array;			/* index --> atom */
     int		buckets;		/* # buckets in char * --> atom */
-    int		locked;			/* table is locked */
     Atom *	table;			/* hash-table */
     int		lookups;		/* # atom lookups */
     int		cmps;			/* # string compares for lookup */
+#ifdef O_ATOMGC
+    int		gc_active;		/* Atom-GC is in progress */
+    int		builtin;		/* Locked atoms (atom-gc) */
+    int		no_hole_before;		/* You won't find a hole before here */
+    int		margin;			/* # atoms to grow before collect */
+    int		non_garbage;		/* # atoms for after last AGC */
+    int		gc;			/* # atom garbage collections */
+    long	collected;		/* # collected atoms */
+    real	gc_time;		/* Time spent on atom-gc */
+    PL_agc_hook_t gc_hook;		/* Current hook */
+#endif
+    atom_t	for_code[256];		/* code --> one-char-atom */
   } atoms;
 
   struct
@@ -116,6 +160,10 @@ typedef struct
   struct
   { ExtensionCell _ext_head;		/* head of registered extensions */
     ExtensionCell _ext_tail;		/* tail of this chain */
+
+    InitialiseHandle initialise_head;	/* PL_initialise_hook() */
+    InitialiseHandle initialise_tail;
+
     int		  _loaded;		/* system extensions are loaded */
   } foreign;
 
@@ -127,17 +175,16 @@ typedef struct
   { Table	table;			/* flag key --> flag */
   } flags;
 
+  struct 
+  { Table	table;			/* global (read-only) features */
+  } feature;
+
   struct
   { buffer	array;			/* index --> functor */
     int		buckets;		/* # buckets in atom --> functor */
-    int		locked;			/* table is locked */
     FunctorDef* table;			/* hash-table */
   } functors;
 
-  struct
-  { Operator	table[OPERATORHASHSIZE]; /* global operator table */
-  } op;
-
   struct
   { TempFile		_tmpfile_head;
     TempFile		_tmpfile_tail;
@@ -147,15 +194,14 @@ typedef struct
     char *		fredshome;	/* home of fred */
     OnHalt		on_halt_list;	/* list of onhalt hooks */
     int			halting;	/* process is shutting down */
-    int			prompt_next;	/* prompt on next read operation */
+    int			gui_app;	/* Win32: Application is a gui app */
     IOFUNCTIONS		iofunctions;	/* initial IO functions */
     IOFUNCTIONS 	org_terminal;	/* IO+Prolog terminal functions */
     IOFUNCTIONS		rl_functions;	/* IO+Terminal+Readline functions */
   } os;
 
   struct
-  { Procedure	alt0;			/* $alt/0, see C_OR */
-    Procedure	garbage_collect0;
+  { Procedure	garbage_collect0;
     Procedure 	block3;
     Procedure	catch3;
     Procedure	true0;
@@ -165,11 +211,24 @@ typedef struct
     Procedure	foreign_registered2;	/* $foreign_registered/2 */
     Procedure	prolog_trace_interception4;
     Procedure	portray;		/* portray/1 */
+    Procedure   dcall1;			/* $call/1 */
+    Procedure	call_cleanup3;		/* call_cleanup/2 */
   } procedures;
 
   struct
-  { buffer	source_files;
+  { buffer	source_files;		/* index --> file */
+    int		_source_index;		/* current index */
+    Table	_source_table;		/* file --> index */
   } files;
+
+#ifdef HAVE_TGETENT
+  struct
+  { int    initialised;			/* initialisation status */
+    char  *_string_area;		/* static area for tgetent */
+    char  *_buf_area;			/* another one */
+    Table  _capabilities;		/* User-level capability table */
+  } terminal;
+#endif
 } PL_global_data_t;
 
 
@@ -178,34 +237,43 @@ typedef struct
 		 *******************************/
 
 typedef struct
-{ LocalFrame environment;		/* Current local frame */
-  FliFrame   foreign_environment;	/* Current foreign context */
-  pl_stacks_t stacks;			/* Prolog runtime stacks */
-#ifdef HAVE_SIGNAL
-  sig_handler sig_handlers[MAXSIGNAL];	/* How Prolog preceives signals */
-#endif  
+{ LocalFrame    environment;		/* Current local frame */
+  Choice	choicepoints;		/* Choice-point chain */
+  FliFrame      foreign_environment;	/* Current foreign context */
+  Word		mark_bar;		/* Mark globals > this one */
+  pl_stacks_t   stacks;			/* Prolog runtime stacks */
+  ulong		bases[STG_MASK+1];	/* area base addresses */
   ulong		pending_signals;	/* PL_raise() pending signals */
   int		current_signal;		/* Currently handled signal */
   int		aborted;		/* thread asked for abort */
   Stack		outofstack;		/* thread is out of stack */
   int		trim_stack_requested;	/* perform a trim-stack */
-  int		autoload;		/* do autoloading */
   int		in_arithmetic;		/* doing arithmetic */
+  int		autoload_nesting;	/* Nesting level in autoloader */
+  void *	glob_info;		/* pl-glob.c */
 
   struct
-  { atom_t	  file;			/* current source file */
-    int	  	  line;			/* current line */
-    long	  character;		/* current character location */
-  } read_source;
+  { int		active;			/* doing pipe I/O */
+    jmp_buf	context;		/* context of longjmp() */
+  } pipe;
 
+  struct
+  { atom_t	current;		/* current global prompt */
+    char *	first;			/* how to prompt first line */
+    int		first_used;		/* did we do the first line? */
+    int		next;			/* prompt on next read operation */
+  } prompt;
+
+  source_location read_source;		/* file, line, char of last term */
   int	 _fileerrors;			/* current file-error status */
-  atom_t _float_format;			/* floating point format */
+  const char   *float_format;		/* floating point format */
 
   struct
   { term_t	term;			/* exception term */
     term_t	bin;			/* temporary handle for exception */
     term_t	printed;		/* already printed exception */
     term_t	tmp;			/* tmp for errors */
+    term_t	pending;		/* used by the debugger */
   } exception;
 
   struct
@@ -214,6 +282,9 @@ typedef struct
     int		profiling;		/* profiler is on */
     ulong	profile_ticks;		/* profile ticks total */
 #endif /* O_PROFILE */
+    ulong	last_cputime;		/* milliseconds last CPU time */
+    ulong	last_systime;		/* milliseconds last SYSTEM time */
+    ulong	last_walltime;		/* milliseconds last Wall time */
   } statistics;
 
   struct
@@ -222,11 +293,11 @@ typedef struct
   } modules;
 
   struct
-  { Atom 	generator;		/* See PL_atom_generator() */
+  { long 	generator;		/* See PL_atom_generator() */
   } atoms;
 
   struct
-  { Assoc	bags;			/* findall/setof bags storage */
+  { struct assoc* bags;			/* findall/setof bags storage */
   } bags;
 
   struct
@@ -240,6 +311,11 @@ typedef struct
     int		_CWDlen;
   } os;
 
+  struct
+  { Table	  table;		/* Feature table */
+    pl_features_t mask;			/* Masked access to booleans */
+  } feature;
+
   struct
   { FindData	find;			/* /<ports> <goal> in tracer */
   } trace;
@@ -248,9 +324,6 @@ typedef struct
   { AbortHandle	_abort_head;		/* PL_abort_hook() */
     AbortHandle _abort_tail;
     
-    InitialiseHandle _initialise_head;	/* PL_initialise_hook() */
-    InitialiseHandle _initialise_tail;
-
     PL_dispatch_hook_t _dispatch_events; /* PL_dispatch_hook() */
 
     buffer	_discardable_buffer;	/* PL_*() character buffers */
@@ -258,6 +331,23 @@ typedef struct
     int		_current_buffer_id;
   } fli;
 
+  struct				/* Local IO stuff */
+  { IOSTREAM *streams[6];		/* handles for standard streams */
+#if 0
+    IOSTREAM *user_input;		/* current user input */
+    IOSTREAM *user_output;		/* current user output */
+    IOSTREAM *user_error;		/* current user error */
+    IOSTREAM *curin;			/* current input stream */
+    IOSTREAM *curout;			/* current output stream */
+    IOSTREAM *din;			/* debugger input */
+    IOSTREAM *dout;			/* debuffer output */
+    IOSTREAM *log;			/* stream used for protocolling */
+    IOSTREAM *term;			/* terminal stream */
+#endif
+    struct input_context *input_stack;	/* maintain input stream info */
+    struct output_context *output_stack; /* maintain output stream info */
+  } IO;
+
 #ifdef O_LIMIT_DEPTH
   struct
   { ulong limit;
@@ -265,50 +355,94 @@ typedef struct
   } depth_info;
 #endif
 
-  pl_gc_status_t	_gc_status;	/* Garbage collection status */
+  struct
+  { long _total_marked;			/* # marked global cells */
+    long _trailcells_deleted;		/* # garbage trailcells */
+    long _relocation_chains;		/* # relocation chains (debugging) */
+    long _relocation_cells;		/* # relocation cells */
+    long _relocated_cells;		/* # relocated cells */
+    long _needs_relocation;		/* # cells that need relocation */
+    long _local_marked;			/* # marked local -> global ptrs */
+    long _marks_swept;			/* # marks swept */
+    long _marks_unswept;		/* # marks swept */
+    long _alien_relocations;		/* # alien_into_relocation_chain() */
+    long _local_frames;			/* frame count for debugging */
+    long _choice_count;			/* choice-point count for debugging */
+
+    pl_gc_status_t	status;		/* Garbage collection status */
+  } gc;
+
 #ifdef O_SHIFT_STACKS
   pl_shift_status_t	_shift_status;	/* Stack shifter status */
 #endif
   
   pl_debugstatus_t _debugstatus;	/* status of the debugger */
-  pl_features_t	   _features;		/* thread-local features */
+
+#ifdef O_PLMT
+  struct
+  { long   magic;			/* PL_THREAD_MAGIC (checking) */
+    atom_t name;			/* name of the thread (if any) */
+    struct _PL_thread_info_t *info;	/* info structure */
+					/* Message queues */
+    pthread_mutex_t      queue_mutex;	/* Message queue mutex */
+    pthread_cond_t       cond_var;	/* condition variable of queue */
+    struct _thread_msg   *msg_head;	/* Head of message queue */
+    struct _thread_msg   *msg_tail;	/* Tail of message queue */
+    struct _thread_sig   *sig_head;	/* Head of signal queue */
+    struct _thread_sig   *sig_tail;	/* Tail of signal queue */
+    struct _at_exit_goal *exit_goals;	/* thread_at_exit/1 goals */
+  } thread;
+#endif
 } PL_local_data_t;
 
-GLOBAL PL_local_data_t  PL_local_data;
 GLOBAL PL_global_data_t PL_global_data;
 GLOBAL PL_code_data_t	PL_code_data;
 
-#define LD (&PL_local_data)
+#ifndef O_PLMT
+GLOBAL PL_local_data_t  PL_local_data;
+#define GET_LD
+#define ARG_LD
+#define ARG1_LD void
+#define PASS_LD
+#define PASS_LD1
+#define LOCAL_LD  (&PL_local_data)
+#define GLOBAL_LD (&PL_local_data)
+#define LD	  GLOBAL_LD
+#endif
 #define GD (&PL_global_data)
 #define CD (&PL_code_data)
 
 #define hTop			(GD->top_of_heap)
 #define hBase			(GD->base_of_heap)
 #define heap_base		(GD->rounded_heap_base)
-#define loaderstatus		(GD->_loaderstatus)
-#define base_addresses		(GD->bases)
 #define functor_array		(GD->functors.array)
 #define atom_array		(GD->atoms.array)
 #define systemDefaults		(GD->defaults)
+#define features		(LD->feature.mask)
 
 #define environment_frame 	(LD->environment)
 #define fli_context	  	(LD->foreign_environment)
 #define source_file_name	(LD->read_source.file)
 #define source_line_no		(LD->read_source.line)
 #define source_char_no		(LD->read_source.character)
-#define LD_sig_handler(n)	(LD->sig_handlers[(n)])
-#define signalled		(LD->pending_signals)
 #define exception_term		(LD->exception.term)
 #define exception_bin		(LD->exception.bin)
 #define exception_printed	(LD->exception.printed)
 #define fileerrors		(LD->_fileerrors)
-#define float_format		(LD->_float_format)
-#define gc_status		(LD->_gc_status)
+#define gc_status		(LD->gc.status)
 #define shift_status		(LD->_shift_status)
 #define debugstatus		(LD->_debugstatus)
-#define features		(LD->_features)
 #define depth_limit		(LD->depth_info.limit)
 #define depth_reached		(LD->depth_info.reached)
+#define base_addresses		(LD->bases)
+#define Suser_input		(LD->IO.streams[0])
+#define Suser_output		(LD->IO.streams[1])
+#define Suser_error		(LD->IO.streams[2])
+#define Scurin			(LD->IO.streams[3])
+#define Scurout			(LD->IO.streams[4])
+#define Sprotocol		(LD->IO.streams[5])
+#define Sdin			Suser_input 		/* not used for now */
+#define Sdout			Suser_output
 
 #ifdef VMCODE_IS_ADDRESS
 #define dewam_table		(CD->_dewam_table)
