@@ -1,10 +1,25 @@
-/*  $Id: pl-itf.h,v 1.93 1997/10/28 13:40:56 jan Exp $
+/*  $Id: pl-itf.h,v 1.174 2002/02/01 16:49:13 jan Exp $
 
-    Copyright (c) 1990 Jan Wielemaker. All rights reserved.
-    See ../LICENCE to find out about your rights.
-    jan@swi.psy.uva.nl
+    Part of SWI-Prolog
 
-    Purpose: Public declarations for defining C-predicates and embedding.
+    Author:        Jan Wielemaker
+    E-mail:        jan@swi.psy.uva.nl
+    WWW:           http://www.swi-prolog.org
+    Copyright (C): 1985-2002, University of Amsterdam
+
+    This library is free software; you can redistribute it and/or
+    modify it under the terms of the GNU Lesser General Public
+    License as published by the Free Software Foundation; either
+    version 2.1 of the License, or (at your option) any later version.
+
+    This library is distributed in the hope that it will be useful,
+    but WITHOUT ANY WARRANTY; without even the implied warranty of
+    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
+    Lesser General Public License for more details.
+
+    You should have received a copy of the GNU Lesser General Public
+    License along with this library; if not, write to the Free Software
+    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
 
 #ifndef _FLI_H_INCLUDED
@@ -28,49 +43,70 @@ before loading this file.  See end of this file.
 /* PLVERSION: 10000 * <Major> + 100 * <Minor> + <Patch> */
 
 #ifndef PLVERSION
-#define PLVERSION 20906
+#define PLVERSION 50000
 #endif
 
-/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
-_declspec(dllexport) is used by MSVC++ 2.0 to declare exports from DLL's.
-- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
+		 /*******************************
+		 *	       EXPORT		*
+		 *******************************/
 
-/* Get export declarations right.  Also in SWI-Stream.h, hence the
-   check to avoid doing it twice.
-*/
+/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
+Traditional and ELF-based Unix systems  don't   need  all this, but COFF
+based systems need  to  import  and   export  symbols  explicitely  from
+executables and shared objects (DLL). On some systems (e.g. AIX) this is
+achieved using import/export files, on Windows   this  is achieved using
+special  declarations  on  exported  symbols.  So,  a  symbol  is  local
+(static), shared between the objects building   an executable or DLL (no
+special declaration) or exported from the executable or DLL.
+
+Both using native Microsoft MSVC as well   as recent CygWin (tested 1.1)
+compilers support __declspec(...) for exporting symbols.
+
+As SWI-Prolog.h can be included seperately or together with this file we
+duplicated this stuff.
+- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
 #ifndef _PL_EXPORT_DONE
 #define _PL_EXPORT_DONE
-#ifdef WIN32
-#ifndef __WIN32__
+#if defined(WIN32) && !defined(__WIN32__)
 #define __WIN32__
 #endif
+
+#if (defined(__WIN32__) || defined(__CYGWIN32__)) && !defined(__LCC__)
+#define HAVE_DECLSPEC
 #endif
 
-#ifdef __WIN32__
+#ifdef HAVE_DECLSPEC
 #ifdef PL_KERNEL
-#define __pl_export	 _declspec(dllexport)
-#define __pl_export_data _declspec(dllexport)
+#define __pl_export	 __declspec(dllexport)
+#define __pl_export_data __declspec(dllexport)
 #define install_t	 void
 #else
 #define __pl_export	 extern
-#define __pl_export_data _declspec(dllimport)
-#define install_t	 _declspec(dllexport) void
+#define __pl_export_data __declspec(dllimport)
+#define install_t	 __declspec(dllexport) void
 #endif
-#else /*__WIN32__*/
+#else /*HAVE_DECLSPEC*/
 #define __pl_export	 extern
 #define __pl_export_data extern
 #define install_t	 void
-#endif /*__WIN32__*/
+#endif /*HAVE_DECLSPEC*/
 #endif /*_PL_EXPORT_DONE*/
 
+
+		 /*******************************
+		 *	       TYPES		*
+		 *******************************/
+
 #ifdef _PL_INCLUDE_H
 typedef Module		module_t;	/* a module */
 typedef Procedure	predicate_t;	/* a predicate handle */
+typedef Record		record_t;	/* handle to a recorded term */
 #else
 typedef	unsigned long	atom_t;		/* Prolog atom */
 typedef void *		module_t;	/* Prolog module */
 typedef void *		predicate_t;	/* Prolog procedure */
+typedef void *		record_t;	/* Prolog recorded term */
 typedef unsigned long	term_t;		/* opaque term handle */
 typedef unsigned long	qid_t;		/* opaque query handle */
 typedef unsigned long	PL_fid_t;	/* opaque foreign context handle */
@@ -87,6 +123,18 @@ typedef foreign_t	(*pl_function_t)(); /* foreign language functions */
 
 #define fid_t PL_fid_t			/* avoid AIX name-clash */
 
+					/* values for PL_get_term_value() */
+typedef union
+{ long   i;				/* PL_INTEGER */
+  double f;				/* PL_FLOAT */
+  char * s;				/* PL_STRING */
+  atom_t a;				/* PL_ATOM */
+  struct				/* PL_TERM */
+  { atom_t name;
+    int    arity;
+  } t;
+} term_value_t;
+
 #ifndef TRUE
 #define TRUE	(1)
 #define FALSE	(0)
@@ -97,17 +145,28 @@ typedef foreign_t	(*pl_function_t)(); /* foreign language functions */
 		 *******************************/
 
 					/* PL_unify_term() arguments */
-#define	PL_VARIABLE	(1)		/* nothing */
-#define PL_ATOM		(2)		/* const char * */
-#define PL_INTEGER	(3)		/* int */
-#define PL_FLOAT	(4)		/* double */
-#define PL_STRING	(5)		/* const char * */
-#define PL_TERM		(6)
+#define	PL_VARIABLE	 (1)		/* nothing */
+#define PL_ATOM		 (2)		/* const char * */
+#define PL_INTEGER	 (3)		/* int */
+#define PL_FLOAT	 (4)		/* double */
+#define PL_STRING	 (5)		/* const char * */
+#define PL_TERM		 (6)
 
-#define PL_FUNCTOR	(10)		/* functor_t, arg ... */
-#define PL_LIST		(11)		/* length, arg ... */
-#define PL_CHARS	(12)		/* const char * */
-#define PL_POINTER	(13)		/* void * */
+					/* PL_unify_term() */
+#define PL_FUNCTOR	 (10)		/* functor_t, arg ... */
+#define PL_LIST		 (11)		/* length, arg ... */
+#define PL_CHARS	 (12)		/* const char * */
+#define PL_POINTER	 (13)		/* void * */
+					/* PlArg::PlArg(text, type) */
+#define PL_CODE_LIST	 (14)		/* [ascii...] */
+#define PL_CHAR_LIST	 (15)		/* [h,e,l,l,o] */
+#define PL_BOOL		 (16)		/* PL_set_feature() */
+#define PL_FUNCTOR_CHARS (17)		/* PL_unify_term() */
+#define _PL_PREDICATE_INDICATOR (18)	/* predicate_t (Procedure) */
+#define PL_SHORT	 (19)		/* short */
+#define PL_INT		 (20)		/* int */
+#define PL_LONG		 (21)		/* long */
+#define PL_DOUBLE	 (22)		/* double */
 
 
 		/********************************
@@ -117,6 +176,7 @@ typedef foreign_t	(*pl_function_t)(); /* foreign language functions */
 #define	PL_succeed	return TRUE	/* succeed deterministically */
 #define PL_fail		return FALSE	/* fail */
 
+
 		/********************************
 		* NON-DETERMINISTIC CALL/RETURN *
 		*********************************/
@@ -155,13 +215,21 @@ typedef struct _PL_extension
 #define PL_FA_NOTRACE		(0x01)	/* foreign cannot be traced */
 #define PL_FA_TRANSPARENT	(0x02)	/* foreign is module transparent */
 #define PL_FA_NONDETERMINISTIC	(0x04)	/* foreign is non-deterministic */
+#define PL_FA_VARARGS		(0x08)	/* call using t0, ac, ctx */
 
 extern			PL_extension PL_extensions[]; /* not Win32! */
-__pl_export void	PL_register_extensions(PL_extension *e);
+__pl_export void	PL_register_extensions(const PL_extension *e);
+__pl_export void	PL_load_extensions(const PL_extension *e);
 __pl_export int		PL_register_foreign(const char *name, int arity,
 					    pl_function_t func, int flags);
 
 
+		 /*******************************
+		 *	      LICENSE		*
+		 *******************************/
+
+void			PL_license(const char *license, const char *module);
+
 		/********************************
 		*            MODULES            *
 		*********************************/
@@ -189,6 +257,7 @@ __pl_export int		PL_strip_module(term_t in, module_t *m, term_t out);
 
 			/* Foreign context frames */
 __pl_export fid_t	PL_open_foreign_frame(void);
+__pl_export void	PL_rewind_foreign_frame(fid_t cid);
 __pl_export void	PL_close_foreign_frame(fid_t cid);
 __pl_export void	PL_discard_foreign_frame(fid_t cid);
 
@@ -211,9 +280,12 @@ __pl_export void	PL_cut_query(qid_t qid);
 __pl_export int		PL_call(term_t t, module_t m);
 __pl_export int		PL_call_predicate(module_t m, int debug,
 					  predicate_t pred, term_t t0);
+			/* Handling exceptions */
 __pl_export term_t	PL_exception(qid_t qid);
+__pl_export int		PL_raise_exception(term_t exception);
 __pl_export int		PL_throw(term_t exception);
 
+
 		 /*******************************
 		 *        TERM-REFERENCES	*
 		 *******************************/
@@ -226,17 +298,35 @@ __pl_export void	PL_reset_term_refs(term_t r);
 
 			/* Constants */
 __pl_export atom_t	PL_new_atom(const char *s);
+__pl_export atom_t	PL_new_atom_nchars(unsigned int len, const char *s);
 __pl_export const char *PL_atom_chars(atom_t a);
+__pl_export const char *PL_atom_nchars(atom_t a, unsigned int *len);
+#ifndef O_DEBUG_ATOMGC
+__pl_export void	PL_register_atom(atom_t a);
+__pl_export void	PL_unregister_atom(atom_t a);
+#endif
 __pl_export functor_t	PL_new_functor(atom_t f, int a);
 __pl_export atom_t	PL_functor_name(functor_t f);
 __pl_export int		PL_functor_arity(functor_t f);
 
 			/* Get C-values from Prolog terms */
 __pl_export int		PL_get_atom(term_t t, atom_t *a);
+__pl_export int		PL_get_bool(term_t t, int *value);
 __pl_export int		PL_get_atom_chars(term_t t, char **a);
-__pl_export int		PL_get_string(term_t t, char **s, int *len);
-__pl_export int		PL_get_list_chars(term_t l, char **s, unsigned flags);
-__pl_export int		PL_get_chars(term_t t, char **s, unsigned flags);
+#define PL_get_string_chars(t, s, l) PL_get_string(t,s,l)
+					/* PL_get_string() is depreciated */
+__pl_export int		PL_get_string(term_t t, char **s, unsigned int *len);
+__pl_export int		PL_get_chars(term_t t, char **s, unsigned int flags);
+__pl_export int		PL_get_list_chars(term_t l, char **s,
+					  unsigned int flags);
+__pl_export int		PL_get_atom_nchars(term_t t,
+					   unsigned int *length, char **a);
+__pl_export int		PL_get_list_nchars(term_t l,
+					   unsigned int *length, char **s,
+					   unsigned int flags);
+__pl_export int		PL_get_nchars(term_t t,
+				      unsigned int *length, char **s,
+				      unsigned int flags);
 __pl_export int		PL_get_integer(term_t t, int *i);
 __pl_export int		PL_get_long(term_t t, long *i);
 __pl_export int		PL_get_pointer(term_t t, void **ptr);
@@ -249,6 +339,8 @@ __pl_export int		PL_get_list(term_t l, term_t h, term_t t);
 __pl_export int		PL_get_head(term_t l, term_t h);
 __pl_export int		PL_get_tail(term_t l, term_t t);
 __pl_export int		PL_get_nil(term_t l);
+__pl_export int		PL_get_term_value(term_t t, term_value_t *v);
+__pl_export char *	PL_quote(int chr, const char *data);
 
 			/* Verify types */
 __pl_export int		PL_term_type(term_t t);
@@ -269,6 +361,16 @@ __pl_export void	PL_put_atom(term_t t, atom_t a);
 __pl_export void	PL_put_atom_chars(term_t t, const char *chars);
 __pl_export void	PL_put_string_chars(term_t t, const char *chars);
 __pl_export void	PL_put_list_chars(term_t t, const char *chars);
+__pl_export void	PL_put_list_codes(term_t t, const char *chars);
+__pl_export void	PL_put_atom_nchars(term_t t,
+					   unsigned int l, const char *chars);
+__pl_export void	PL_put_string_nchars(term_t t,
+					     unsigned int len,
+					     const char *chars);
+__pl_export void	PL_put_list_nchars(term_t t,
+					   unsigned int l, const char *chars);
+__pl_export void	PL_put_list_ncodes(term_t t,
+					   unsigned int l, const char *chars);
 __pl_export void	PL_put_integer(term_t t, long i);
 __pl_export void	PL_put_pointer(term_t t, void *ptr);
 __pl_export void	PL_put_float(term_t t, double f);
@@ -279,6 +381,7 @@ __pl_export void	PL_put_term(term_t t1, term_t t2);
 
 			/* construct a functor or list-cell */
 __pl_export void	PL_cons_functor(term_t h, functor_t f, ...);
+__pl_export void	PL_cons_functor_v(term_t h, functor_t fd, term_t a0);
 __pl_export void	PL_cons_list(term_t l, term_t h, term_t t);
 
 			/* Unify term-references */
@@ -286,9 +389,17 @@ __pl_export int		PL_unify(term_t t1, term_t t2);
 __pl_export int		PL_unify_atom(term_t t, atom_t a);
 __pl_export int		PL_unify_atom_chars(term_t t, const char *chars);
 __pl_export int		PL_unify_list_chars(term_t t, const char *chars);
+__pl_export int		PL_unify_list_codes(term_t t, const char *chars);
 __pl_export int		PL_unify_string_chars(term_t t, const char *chars);
+__pl_export int		PL_unify_atom_nchars(term_t t,
+					     unsigned int l, const char *s);
+__pl_export int		PL_unify_list_ncodes(term_t t,
+					     unsigned int l, const char *s);
+__pl_export int		PL_unify_list_nchars(term_t t,
+					     unsigned int l, const char *s);
 __pl_export int		PL_unify_string_nchars(term_t t,
-					       int len, const char *chars);
+					       unsigned int len,
+					       const char *chars);
 __pl_export int		PL_unify_integer(term_t t, long n);
 __pl_export int		PL_unify_float(term_t t, double f);
 __pl_export int		PL_unify_pointer(term_t t, void *ptr);
@@ -303,6 +414,8 @@ __pl_export int		PL_unify_term(term_t t, ...);
 		 *******************************/
 
 __pl_export char *	PL_get_filename(term_t n, char *buf, unsigned int sz);
+__pl_export void	PL_changed_cwd(void); /* foreign code changed CWD */
+__pl_export const char *PL_cwd();
 
 
 		 /*******************************
@@ -326,6 +439,7 @@ __pl_export int		PL_cvt_o_atom(atom_t c, term_t p);
 		 *******************************/
 
 __pl_export int		PL_compare(term_t t1, term_t t2);
+__pl_export int		PL_same_compound(term_t t1, term_t t2);
 
 		 /*******************************
 		 *	     MESSAGES		*
@@ -334,6 +448,18 @@ __pl_export int		PL_compare(term_t t1, term_t t2);
 __pl_export int		PL_warning(const char *fmt, ...);
 __pl_export void	PL_fatal_error(const char *fmt, ...);
 
+		 /*******************************
+		 *      RECORDED DATABASE	*
+		 *******************************/
+
+__pl_export record_t	PL_record(term_t term);
+__pl_export void	PL_recorded(record_t record, term_t term);
+__pl_export void	PL_erase(record_t record);
+
+__pl_export char *	PL_record_external(term_t t, unsigned int *size);
+__pl_export int		PL_recorded_external(const char *rec, term_t term);
+__pl_export int		PL_erase_external(char *rec);
+
 		 /*******************************
 		 *	      FEATURES		*
 		 *******************************/
@@ -349,7 +475,7 @@ __pl_export void	_PL_put_atomic(term_t t, atomic_t a);
 __pl_export int		_PL_unify_atomic(term_t t, atomic_t a);
 __pl_export void	_PL_copy_atomic(term_t t, atomic_t a);
 	    int		_PL_get_name_arity(term_t t, atom_t *name, int *arity);
-	    void	_PL_get_arg(int index, term_t t, term_t a);
+__pl_export void	_PL_get_arg(int index, term_t t, term_t a);
 
 
 		 /*******************************
@@ -364,7 +490,9 @@ __pl_export void	_PL_copy_atomic(term_t t, atomic_t a);
 #define CVT_VARIABLE	0x0020
 #define CVT_NUMBER	(CVT_INTEGER|CVT_FLOAT)
 #define CVT_ATOMIC	(CVT_NUMBER|CVT_ATOM|CVT_STRING)
-#define CVT_ALL		0x00ff
+#define CVT_WRITE	0x0040		/* as of version 3.2.10 */
+#define CVT_ALL		(CVT_ATOMIC|CVT_LIST)
+#define CVT_MASK	0x00ff
 
 #define BUF_DISCARDABLE	0x0000
 #define BUF_RING	0x0100
@@ -376,17 +504,40 @@ __pl_export void	_PL_copy_atomic(term_t t, atomic_t a);
 		 *******************************/
 
 					/* Make IOSTREAM known to Prolog */
-__pl_export int PL_open_stream(term_t t, IOSTREAM *s);
+__pl_export int  PL_open_stream(term_t t, IOSTREAM *s); /* compatibility */
+__pl_export int  PL_unify_stream(term_t t, IOSTREAM *s);
+__pl_export int  PL_get_stream_handle(term_t t, IOSTREAM **s);
+__pl_export void PL_release_stream(IOSTREAM *s);
+__pl_export IOSTREAM *PL_open_resource(module_t m,
+				       const char *name,
+				       const char *rc_class,
+				       const char *mode);
+
+#define PL_WRT_QUOTED		0x01	/* quote atoms */
+#define PL_WRT_IGNOREOPS	0x02	/* ignore list/operators */
+#define PL_WRT_NUMBERVARS	0x04	/* print $VAR(N) as a variable */
+#define PL_WRT_PORTRAY		0x08	/* call portray */
+#define PL_WRT_CHARESCAPES	0x10	/* Output ISO escape sequences */
+
+__pl_export int PL_write_term(IOSTREAM *s,
+			      term_t term,
+			      int precedence,
+			      int flags);
 #endif
 
+__pl_export int PL_chars_to_term(const char *chars,
+				 term_t term);
+
 
 		 /*******************************
 		 *	    EMBEDDING		*
 		 *******************************/
 
 __pl_export int		PL_initialise(int argc, char **argv);
+__pl_export int		PL_is_initialised(int *argc, char ***argv);
 install_t		PL_install_readline(void);
 __pl_export int		PL_toplevel(void);
+__pl_export int		PL_cleanup(int status);
 __pl_export void	PL_halt(int status);
 
 		 /*******************************
@@ -414,9 +565,9 @@ __pl_export int		PL_dispatch(int fd, int wait);
 __pl_export int		PL_ttymode(int fd);
 __pl_export void	PL_add_to_protocol(const char *buf, int count);
 __pl_export char *	PL_prompt_string(int fd);
-__pl_export void	PL_write_prompt(int fd, int dowrite);
+__pl_export void	PL_write_prompt(int dowrite);
 __pl_export void	PL_prompt_next(int fd);
-__pl_export char *	PL_atom_generator(char *prefix, int state);
+__pl_export char *	PL_atom_generator(const char *prefix, int state);
 __pl_export void	PL_clock_wait_ticks(long waited);
 
 
@@ -431,14 +582,14 @@ typedef int  (*PL_dispatch_hook_t)(int fd);
 typedef void (*PL_abort_hook_t)(void);
 typedef void (*PL_initialise_hook_t)(int argc, char **argv);
 typedef void (*PL_async_hook_t)(void);	/* Win32 only (O_ASYNC_HOOK) */
-typedef int  (*PL_getkey_hook_t)(int fd);
+typedef int  (*PL_agc_hook_t)(atom_t a);
 
 __pl_export PL_dispatch_hook_t PL_dispatch_hook(PL_dispatch_hook_t);
 __pl_export void	       PL_abort_hook(PL_abort_hook_t);
 __pl_export void	       PL_initialise_hook(PL_initialise_hook_t);
 __pl_export int		       PL_abort_unhook(PL_abort_hook_t);
 __pl_export PL_async_hook_t    PL_async_hook(unsigned int, PL_async_hook_t);
-__pl_export PL_getkey_hook_t   PL_getkey_hook(PL_getkey_hook_t);
+__pl_export PL_agc_hook_t      PL_agc_hook(PL_agc_hook_t);
 
 
 		/********************************
@@ -460,9 +611,10 @@ __pl_export void PL_handle_signals(void);
 #define PL_ACTION_BREAK		4	/* create a break environment */
 #define PL_ACTION_HALT		5	/* halt Prolog execution */
 #define PL_ACTION_ABORT		6	/* generate a Prolog abort */
-#define PL_ACTION_SYMBOLFILE	7	/* make arg. the symbol file */
+					/* 7: Obsolete PL_ACTION_SYMBOLFILE */
 #define PL_ACTION_WRITE		8	/* write via Prolog i/o buffer */
 #define PL_ACTION_FLUSH		9	/* Flush Prolog i/o buffer */
+#define PL_ACTION_GUIAPP	10	/* Win32: set when this is a gui */
 
 __pl_export int	 PL_action(int, ...);	/* perform some action */
 __pl_export void PL_on_halt(void (*)(int, void *), void *);
@@ -473,17 +625,36 @@ __pl_export void PL_on_halt(void (*)(int, void *), void *);
 
 #define PL_QUERY_ARGC		1	/* return main() argc */
 #define PL_QUERY_ARGV		2	/* return main() argv */
-#define PL_QUERY_SYMBOLFILE	3	/* return current symbol file */
-#define PL_QUERY_ORGSYMBOLFILE	4	/* symbol file before first load */
+					/* 3: Obsolete PL_QUERY_SYMBOLFILE */
+					/* 4: Obsolete PL_QUERY_ORGSYMBOLFILE*/
 #define PL_QUERY_GETC		5	/* Read character from terminal */
 #define PL_QUERY_MAX_INTEGER	6	/* largest integer */
 #define PL_QUERY_MIN_INTEGER	7	/* smallest integer */
 #define PL_QUERY_MAX_TAGGED_INT	8	/* largest tagged integer */
 #define PL_QUERY_MIN_TAGGED_INT	9	/* smallest tagged integer */
-#define PL_QUERY_VERSION       10	/* 207006 = 2.7.6 */
+#define PL_QUERY_VERSION        10	/* 207006 = 2.7.6 */
 
 __pl_export long	PL_query(int);	/* get information from Prolog */
 
+
+		 /*******************************
+		 *	  PROLOG THREADS	*
+		 *******************************/
+
+typedef struct
+{ unsigned long	    local_size;		/* Stack sizes */
+  unsigned long	    global_size;
+  unsigned long	    trail_size;
+  unsigned long	    argument_size;
+  char *	    alias;		/* alias name */
+} PL_thread_attr_t;
+
+
+__pl_export int	PL_thread_self(void);	/* Prolog thread id (-1 if none) */
+__pl_export int PL_thread_attach_engine(PL_thread_attr_t *attr);
+__pl_export int PL_thread_destroy_engine(void);
+
+
 		 /*******************************
 		 *       FAST XPCE SUPPORT	*
 		 *******************************/
@@ -496,10 +667,10 @@ typedef struct
   } value;
 } xpceref_t;
 
-__pl_export int		_PL_get_xpce_reference(term_t t, xpceref_t *ref);
-__pl_export int		_PL_unify_xpce_reference(term_t t, xpceref_t *ref);
-__pl_export void	_PL_put_xpce_reference_i(term_t t, unsigned long r);
-__pl_export void	_PL_put_xpce_reference_a(term_t t, atom_t name);
+__pl_export int	 _PL_get_xpce_reference(term_t t, xpceref_t *ref);
+__pl_export int  _PL_unify_xpce_reference(term_t t, xpceref_t *ref);
+__pl_export void _PL_put_xpce_reference_i(term_t t, unsigned long r);
+__pl_export void _PL_put_xpce_reference_a(term_t t, atom_t name);
 
 		 /*******************************
 		 *        COMPATIBILITY		*
