@@ -1,10 +1,25 @@
-/*  $Id: pl-main.c,v 1.66 1998/07/06 10:30:04 jan Exp $
+/*  $Id: pl-main.c,v 1.117 2002/02/01 16:49:13 jan Exp $
 
-    Copyright (c) 1990 Jan Wielemaker. All rights reserved.
-    See ../LICENCE to find out about your rights.
-    jan@swi.psy.uva.nl
+    Part of SWI-Prolog
 
-    Purpose: Prologs main module
+    Author:        Jan Wielemaker
+    E-mail:        jan@swi.psy.uva.nl
+    WWW:           http://www.swi-prolog.org
+    Copyright (C): 1985-2002, University of Amsterdam
+
+    This library is free software; you can redistribute it and/or
+    modify it under the terms of the GNU Lesser General Public
+    License as published by the Free Software Foundation; either
+    version 2.1 of the License, or (at your option) any later version.
+
+    This library is distributed in the hope that it will be useful,
+    but WITHOUT ANY WARRANTY; without even the implied warranty of
+    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
+    Lesser General Public License for more details.
+
+    You should have received a copy of the GNU Lesser General Public
+    License along with this library; if not, write to the Free Software
+    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
 
 /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
@@ -21,19 +36,27 @@ option  parsing,  initialisation  and  handling  of errors and warnings.
 #include <sys/param.h>
 #endif
 
+#define LOCK()   PL_LOCK(L_INIT)
+#define UNLOCK() PL_UNLOCK(L_INIT)
+
 #ifdef FORCED_MALLOC_BASE
 #include "morecore.c"
 #endif
 
+#if defined(_DEBUG) && defined(WIN32)
+#include <crtdbg.h>
+#endif
+
 static int	usage(void);
 static int	giveVersionInfo(const char *a);
 static bool	vsysError(const char *fm, va_list args);
 
 #define	optionString(s) { if (argc > 1) \
-			  { s = argv[1]; argc--; argv++; \
+			  { if ( s ) remove_string(s); \
+			    s = store_string(argv[1]); \
+			    argc--; argv++; \
 			  } else \
-			  { usage(); \
-			    exit(1); \
+			  { return -1; \
 			  } \
 			}
 #define K * 1024L
@@ -123,24 +146,6 @@ findHome(char *symbols)
        ExistsDirectory(home) )
     return store_string(home);
 
-#if tos || __DOS__ || __WINDOWS__
-#if tos
-#define HasDrive(c) (Drvmap() & (1 << (c - 'a')))
-#else
-#define HasDrive(c) 1
-#endif
-  { char *drv;
-    static char drvs[] = "cdefghijklmnopab";
-    char home[MAXPATHLEN];
-
-    for(drv = drvs; *drv; drv++)
-    { Ssprintf(home, "/%c:/pl", *drv);
-      if ( HasDrive(*drv) && ExistsDirectory(home) )
-        return store_string(home);
-    }
-  }
-#endif
-
   return NULL;
 }
 
@@ -172,7 +177,7 @@ defaultSystemInitFile(char *a0)
   if ( strlen(buf) > 0 )
     return store_string(buf);
 
-  return "pl";
+  return store_string("pl");
 }
 
 
@@ -214,7 +219,7 @@ setupGNUEmacsInferiorMode()
 
   if ( ((s = getenv3("EMACS", envbuf, sizeof(envbuf))) && streq(s, "t")) ||
        ((s = getenv3("INFERIOR", envbuf, sizeof(envbuf))) && streq(s, "yes")) )
-    GD->cmdline.notty = TRUE;
+    clearFeatureMask(TTY_CONTROL_FEATURE);
 }
 
 
@@ -223,12 +228,19 @@ initPaths()
 { char plp[MAXPATHLEN];
   char *symbols = NULL;			/* The executable */
 
-  if ( !(symbols = Symbols(plp)) ||
+  if ( !(symbols = findExecutable(GD->cmdline.argv[0], plp)) ||
        !(symbols = DeRefLink(symbols, plp)) )
     symbols = GD->cmdline.argv[0];
 
-  GD->paths.executable	     = store_string(symbols);
+  DEBUG(2, Sdprintf("rc-module: %s\n", symbols));
+
   systemDefaults.home	     = findHome(symbols);
+
+#ifdef __WIN32__			/* we want no module but the .EXE */
+  symbols = findExecutable(NULL, plp);
+  DEBUG(2, Sdprintf("Executable: %s\n", symbols));
+#endif
+  GD->paths.executable	     = store_string(symbols);
   systemDefaults.startup     = store_string(PrologPath(DEFSTARTUP, plp));
   GD->options.systemInitFile = defaultSystemInitFile(GD->cmdline.argv[0]);
 
@@ -254,11 +266,17 @@ initDefaults()
   systemDefaults.toplevel    = "prolog";
   systemDefaults.notty       = NOTTYCONTROL;
 
+#ifdef __WIN32__
+  getDefaultsFromRegistry();
+#endif
+
   GD->io_initialised	     = FALSE;
   GD->initialised	     = FALSE;
-  GD->cmdline.notty	     = systemDefaults.notty;
   GD->bootsession	     = FALSE;
-  LD->autoload		     = TRUE;
+  if ( systemDefaults.notty )
+    clearFeatureMask(TTY_CONTROL_FEATURE);
+  else
+    setFeatureMask(TTY_CONTROL_FEATURE);
 }
 
 
@@ -316,16 +334,17 @@ do_value:
 
 static void
 initDefaultOptions()
-{ GD->options.compileOut    = "a.out";
+{ GD->options.compileOut    = store_string("a.out");
   GD->options.localSize     = systemDefaults.local    K;
   GD->options.globalSize    = systemDefaults.global   K;
   GD->options.trailSize     = systemDefaults.trail    K;
   GD->options.argumentSize  = systemDefaults.argument K;
   GD->options.heapSize      = systemDefaults.heap     K;
-  GD->options.goal	    = systemDefaults.goal;
-  GD->options.topLevel      = systemDefaults.toplevel;
-  GD->options.initFile      = systemDefaults.startup;
-  GD->options.saveclass	    = "none";
+  GD->options.goal	    = store_string(systemDefaults.goal);
+  GD->options.topLevel      = store_string(systemDefaults.toplevel);
+  GD->options.initFile      = store_string(systemDefaults.startup);
+  GD->options.scriptFile    = store_string("");
+  GD->options.saveclass	    = store_string("none");
 
   if ( !GD->bootsession && GD->resourceDB )
   { IOSTREAM *op = SopenRC(GD->resourceDB, "$options", "$prolog", RC_RDONLY);
@@ -343,6 +362,13 @@ initDefaultOptions()
 }
 
 
+/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
+Does the commandline option parsing.  Actually   we  should  use the GNU
+getopt package and deal nicely with long   arguments  as well as shorts,
+but these options are  too  widely  used   as  they  are  to change them
+overnight. Returns -1 on error.
+- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
+
 static int
 parseCommandLineOptions(int argc0, char **argv, int *compile)
 { int argc = argc0;
@@ -352,9 +378,14 @@ parseCommandLineOptions(int argc0, char **argv, int *compile)
 
     if ( streq(s, "-" ) )		/* pl <plargs> -- <app-args> */
       break;
+    if ( *s == '-' )
+      continue;				/* don't handle --long=value */
 
     if ( streq(s, "tty") )	/* +/-tty */
-    { GD->cmdline.notty = (s[-1] == '-');
+    { if ( s[-1] == '+' )
+	setFeatureMask(TTY_CONTROL_FEATURE);
+      else
+	clearFeatureMask(TTY_CONTROL_FEATURE);
       continue;
     }
 
@@ -364,17 +395,13 @@ parseCommandLineOptions(int argc0, char **argv, int *compile)
 			{ GD->debug_level = atoi(argv[1]);
 			  argc--, argv++;
 			} else
-			{ usage();
-			  exit(1);
-			}
+			  return -1;
 			break;
 	case 'p':	if (!argc)	/* handled in Prolog */
-			{ usage();
-			  exit(1);
-			}
+			  return -1;
 			argc--, argv++;
 			break;
-	case 'O':	GD->cmdline.optimise = TRUE;
+	case 'O':	GD->cmdline.optimise = TRUE; /* see initFeatures() */
 			break;
 	case 'x':
   	case 'o':	optionString(GD->options.compileOut);
@@ -383,6 +410,8 @@ parseCommandLineOptions(int argc0, char **argv, int *compile)
 			break;
 	case 'F':	optionString(GD->options.systemInitFile);
 			break;
+	case 's':	optionString(GD->options.scriptFile);
+			break;
 	case 'g':	optionString(GD->options.goal);
 			break;
 	case 't':	optionString(GD->options.topLevel);
@@ -391,6 +420,8 @@ parseCommandLineOptions(int argc0, char **argv, int *compile)
 			break;
 	case 'b':	GD->bootsession = TRUE;
 			break;
+	case 'q':	GD->options.silent = TRUE;
+			break;
 	case 'L':
 	case 'G':
 	case 'T':
@@ -399,9 +430,7 @@ parseCommandLineOptions(int argc0, char **argv, int *compile)
         { unsigned long size = memarea_limit(&s[1]);
 	  
 	  if ( size == MEMAREA_INVALID_SIZE )
-	  { usage();
-	    exit(1);
-	  }
+	    return -1;
 
 	  switch(*s)
 	  { case 'L':	GD->options.localSize    = size; goto next;
@@ -421,7 +450,7 @@ parseCommandLineOptions(int argc0, char **argv, int *compile)
 }
 
 
-void
+static void
 replace_extension(char *path, const char *ext)
 { char *s = path + strlen(path);
 
@@ -444,6 +473,18 @@ replace_extension(char *path, const char *ext)
 Find the resource database.
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
+#ifndef BOOTFILE		/* normally delivered through config.h */
+#if SIZEOF_LONG == 4
+#define BOOTFILE "boot32.prc"
+#else
+#if SIZEOF_LONG == 8
+#define BOOTFILE "boot64.prc"
+#else
+#define BOOTFILE "boot.prc"
+#endif
+#endif
+#endif
+
 RcArchive
 openResourceDB(int argc, char **argv)
 { RcArchive rc;
@@ -488,7 +529,8 @@ openResourceDB(int argc, char **argv)
 
   if ( systemDefaults.home )
   { strcpy(tmp, systemDefaults.home);
-    strcat(tmp, "/boot.prc");
+    strcat(tmp, "/");
+    strcat(tmp, BOOTFILE);
 
     return rc_open_archive(tmp, flags);
   }
@@ -497,14 +539,182 @@ openResourceDB(int argc, char **argv)
 }
 
 
+int
+PL_is_initialised(int *argc, char ***argv)
+{ if ( GD->initialised )
+  { if ( argc )
+      *argc = GD->cmdline.argc;
+    if ( argv )
+      *argv = GD->cmdline.argv;
+
+    return TRUE;
+  }
+
+  return FALSE;
+}
+
+
+/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
+Argument handling. This routine also takes care of a script-file launced
+with the first line
+
+#!/path/to/pl -L0 -f
+
+On Unix this is passed as below.  We   need  to break the first argument
+ourselves and we need to restore the argv[0]  path as pl might not be on
+$PATH!
+
+	{pl, '-L0 -f', <file>}
+
+On Windows this is simply passed as below.   We have to analyse the file
+ourselves. Unfortunately this needs to be done  on C as it might contain
+stack-parameters.
+
+	{plwin.exe <file>}
+- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
+
+#ifndef MAXLINE
+#define MAXLINE 1024
+#endif
+#ifndef MAXARGV
+#define MAXARGV 1024
+#endif
+
+static void
+script_argv(int argc, char **argv)
+{ FILE *fd;
+  int i;
+
+  DEBUG(1,
+	{ for(i=0; i< argc; i++)
+	    Sdprintf("argv[%d] = '%s'\n", i, argv[i]);
+	});
+
+#ifdef __unix__
+  if ( argc >= 3 &&
+       (strpostfix(argv[1], "-f") || strpostfix(argv[1], "-s")) &&
+       (fd = fopen(argv[2], "r")) )	/* ok, this is a script invocation */
+#else
+  if ( argc >= 2 &&
+       (fd = fopen(argv[1], "r")) )
+#endif
+  { char buf[MAXLINE];
+    char *s;
+    char *av[MAXARGV];
+    int  an = 0;
+
+#ifdef WIN32
+    if ( argc == 2 )
+    { char tmp[MAXPATHLEN];
+      char dir[MAXPATHLEN];
+
+      _xos_canonical_filename(argv[1], tmp);
+      if ( IsAbsolutePath(tmp) )
+	chdir(DirName(tmp, dir));
+    }
+#endif
+
+    fgets(buf, sizeof(buf), fd);
+    if ( !strprefix(buf, "#!") )
+    { fclose(fd);
+      goto noscript;
+    }
+
+    for(s = &buf[2]; *s; )
+    { while( *s && isBlank(*s) )
+	s++;
+
+      if ( *s )
+      { char *start = s;
+	char *o = s;
+
+	while( *s && !isBlank(*s) )
+	{ if ( *s == '\'' || *s == '"' )
+	  { int c0 = *s++;
+
+	    while(*s && *s != c0)
+	      *o++ = *s++;
+	    if ( *s )
+	      s++;
+	  } else
+	    *o++ = *s++;
+	}
+
+#ifndef __unix__
+	if ( an == 0 )
+	{ av[an++] = argv[0];		/* the original interpreter */
+	  if ( *o != '-' )		/* The interpreter */
+	    continue;
+	}
+#endif
+	av[an] = allocHeap(o-start+1);
+	strncpy(av[an], start, o-start);
+	av[an][o-start] = EOS;
+	if ( ++an >= MAXARGV )
+	  fatalError("Too many script arguments");
+      }
+    }
+    if ( an+argc-2+2 > MAXARGV )	/* skip 2, add -- and NULL */
+      fatalError("Too many script arguments");
+
+#ifdef __unix__
+    i = 2;
+#else
+    i = 1;
+#endif
+    av[an++] = argv[i++];		/* the script file */
+    av[an++] = "--";			/* separate arguments */
+    for(; i<argc; i++)
+      av[an++] = argv[i];
+    GD->cmdline.argc = an;
+    av[an++] = NULL;
+    GD->cmdline.argv = allocHeap(sizeof(char *) * an);
+    memcpy(GD->cmdline.argv, av, sizeof(char *) * an);
+
+    fclose(fd);
+  } else
+  { noscript:
+    GD->cmdline.argc = argc;
+    GD->cmdline.argv = argv;
+  }
+}
+
+
 int
 PL_initialise(int argc, char **argv)
 { int n;
   bool compile = FALSE;
   const char *rcpath = "<none>";
 
-  GD->cmdline.argc = argc;
-  GD->cmdline.argv = argv;
+#if defined(_DEBUG) && defined(WIN32) /*&& 0*/
+  _CrtSetDbgFlag(_CRTDBG_ALLOC_MEM_DF|
+		 _CRTDBG_CHECK_CRT_DF|
+		 _CRTDBG_CHECK_ALWAYS_DF/*|
+		 _CRTDBG_DELAY_FREE_MEM_DF|
+		 _CRTDBG_LEAK_CHECK_DF*/);
+#endif
+
+  LOCK();
+  if ( GD->initialised )
+  { UNLOCK();
+    succeed;
+  }
+
+  SinitStreams();			/* before anything else */
+
+  script_argv(argc, argv);		/* hande #! arguments */
+  argc = GD->cmdline.argc;
+  argv = GD->cmdline.argv;
+  DEBUG(1,
+  { int i;
+    
+    Sdprintf("argv =");
+    for(i=0; i<argc; i++)
+      Sdprintf(" %s", argv[i]);
+    Sdprintf("\n");
+  });
+
+      
   GD->debug_level  = 0;			/* 1-9: debug, also -d <level> */
 
 /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
@@ -521,12 +731,15 @@ properly on Linux. Don't bother with it.
   malloc_debug(O_MALLOC_DEBUG);
 #endif
 
+#ifdef O_PLMT
+  initPrologThreads();			/* initialise thread system */
+#endif
+
   initOs();				/* Initialise OS bindings */
   initDefaults();			/* Initialise global defaults */
   initPaths();				/* fetch some useful paths */
   setupGNUEmacsInferiorMode();		/* Detect running under EMACS */
 
-  
   if ( (GD->resourceDB = rc_open_archive(GD->paths.executable, RC_RDONLY)) )
   { rcpath = ((RcArchive)GD->resourceDB)->path;
     initDefaultOptions();
@@ -534,10 +747,13 @@ properly on Linux. Don't bother with it.
 
   if ( !GD->resourceDB ||
        !streq(GD->options.saveclass, "runtime") )
-  { argc--; argv++;
+  { int done;
+    argc--; argv++;
 
     if ( argc == 1 && giveVersionInfo(argv[0]) ) /* -help, -v, etc */
+    { UNLOCK();
       exit(0);
+    }
 
     for(n=0; n<argc; n++)		/* need to check this first */
     { if ( streq(argv[n], "--" ) )	/* --: terminates argument list */
@@ -551,21 +767,29 @@ properly on Linux. Don't bother with it.
     DEBUG(1, if (GD->bootsession) Sdprintf("Boot session\n"););
 
     if ( !GD->resourceDB )
-    { int done;
-
-      if ( !(GD->resourceDB = openResourceDB(argc, argv)) )
+    { if ( !(GD->resourceDB = openResourceDB(argc, argv)) )
+      { UNLOCK();
 	fatalError("Could not find system resources");
+      }
       rcpath = ((RcArchive)GD->resourceDB)->path;
 
       initDefaultOptions();
-      done = parseCommandLineOptions(argc, argv, &compile);
-      argc -= done;
-      argv += done;
     }
+
+    if ( (done = parseCommandLineOptions(argc, argv, &compile)) < 0 )
+    { usage();
+      UNLOCK();
+      fail;
+    }
+    argc -= done;
+    argv += done;
   }
 
   setupProlog();
-  CSetFeature("resource_database", rcpath);
+#ifdef O_PLMT
+  aliasThread(PL_thread_self(), PL_new_atom("main"));
+#endif
+  defFeature("resource_database", FT_ATOM|FF_READONLY, rcpath);
   initialiseForeign(GD->cmdline.argc, /* PL_initialise_hook() functions */
 		    GD->cmdline.argv);
   systemMode(TRUE);
@@ -575,9 +799,10 @@ properly on Linux. Don't bother with it.
     char *rcpathcopy = store_string(rcpath); /* rcpath is destroyed on close */
 
     if ( !compileFileList(s, argc, argv) )
-      Halt(1);
-    if ( Sclose(s) != 0 ||
-	 !rc_close_archive(GD->resourceDB) )
+    { UNLOCK();
+      PL_halt(1);
+    }
+    if ( Sclose(s) != 0 || !rc_save_archive(GD->resourceDB, NULL) )
     { 
 #if defined(__WINDOWS__) || defined(__WIN32__)
       PlMessage("Failed to save system resources: %s", rc_strerror(rc_errno));
@@ -586,7 +811,8 @@ properly on Linux. Don't bother with it.
 	       "[ERROR: Failed to save system resources %s]\n",
 	       rc_strerror(rc_errno));
 #endif
-      Halt(1);
+      UNLOCK();
+      PL_halt(1);
     }
 #if defined(__WINDOWS__) || defined(__WIN32__)
     PlMessage("Boot compilation has created %s", rcpathcopy);
@@ -594,20 +820,25 @@ properly on Linux. Don't bother with it.
     Sfprintf(Serror,
 	     "Boot compilation has created %s\n", rcpathcopy);
 #endif
-    Halt(0);
+    UNLOCK();
+    PL_halt(0);
   } else
   { IOSTREAM *statefd = SopenRC(GD->resourceDB, "$state", "$prolog", RC_RDONLY);
 
     if ( statefd )
     { GD->bootsession = TRUE;
       if ( !loadWicFromStream(statefd) )
-	Halt(1);
+      { UNLOCK();
+	fail;
+      }
       GD->bootsession = FALSE;
 
       Sclose(statefd);
     } else
+    { UNLOCK();
       fatalError("Resource database \"%s\" does not contain a saved state",
 		 rcpath);
+    }
   }
 
   debugstatus.styleCheck = (LONGATOM_CHECK|
@@ -615,17 +846,29 @@ properly on Linux. Don't bother with it.
 			    DISCONTIGUOUS_STYLE);
   systemMode(FALSE);
   GD->initialised = TRUE;
+  registerForeignLicenses();
 
   DEBUG(1, Sdprintf("Starting Prolog Part of initialisation\n"));
 
   if ( compile )
-  { Halt(prolog(lookupAtom("$compile")) ? 0 : 1);
+  { int status = prologToplevel(PL_new_atom("$compile")) ? 0 : 1;
+
+    UNLOCK();
+    PL_halt(status);
+    fail;				/* make compile happy */
+  } else
+  { int status = prologToplevel(PL_new_atom("$initialise"));
+    UNLOCK();
+    return status;
   }
-    
-  return prolog(lookupAtom("$initialise"));
 }
 
+
+#ifdef __GNUC__
 typedef const char *cline;
+#else
+typedef char *cline;
+#endif
 
 static int
 usage()
@@ -641,13 +884,15 @@ usage()
     "    7) %s [options] [-o output] -b bootfile -c file ...\n",
     "Options:\n",
     "    -x state         Start from state (must be first)\n",
-    "    -[LGTAH]size[KM] Specify {Local,Global,Trail,Argument,Heap} limits\n",
+    "    -[LGTA]size[KM]  Specify {Local,Global,Trail,Argument} limits\n",
     "    -t toplevel      Toplevel goal\n",
     "    -g goal          Initialisation goal\n",
-    "    -f file          Initialisation file\n",
-    "    -F file          System Initialisation file\n",
+    "    -f file          User initialisation file\n",
+    "    -F file          System initialisation file\n",
+    "    -s file          Script source file\n",
     "    [+/-]tty         Allow tty control\n",
     "    -O               Optimised compilation\n",
+    "    -q               Quiet operation\n",
     NULL
   };
   const cline *lp = lines;
@@ -677,7 +922,6 @@ arch()
   return TRUE;
 }
 
-
 static int
 runtime_vars()
 { Sprintf("CC=\"%s\";\n"
@@ -685,17 +929,29 @@ runtime_vars()
 	  "PLARCH=\"%s\";\n"
 	  "PLLIBS=\"%s\";\n"
 	  "PLLDFLAGS=\"%s\";\n"
+#ifdef SO_EXT
+	  "PLSOEXT=\"%s\";\n"
+#endif
 	  "PLVERSION=\"%d\";\n"
 #if defined(HAVE_DLOPEN) || defined(HAVE_SHL_LOAD)
-	  "PLSHARED=\"yes\";\n",
+	  "PLSHARED=\"yes\";\n"
 #else
-	  "PLSHARED=\"no\";\n",
+	  "PLSHARED=\"no\";\n"
 #endif
+#ifdef O_PLMT
+	  "PLTHREADS=\"yes\";\n"
+#else
+	  "PLTHREADS=\"no\";\n"
+#endif
+	  ,
 	  C_CC,
 	  systemDefaults.home ? systemDefaults.home : "<no home>",
 	  ARCH,
 	  C_LIBS,
 	  C_LDFLAGS,
+#ifdef SO_EXT
+	  SO_EXT,
+#endif
 	  PLVERSION);
 
   return TRUE;
@@ -717,6 +973,122 @@ giveVersionInfo(const char *a)
 }
 
 
+		 /*******************************
+		 *	     CLEANUP		*
+		 *******************************/
+
+typedef void (*halt_function)(int, Void);
+
+struct on_halt
+{ halt_function	function;
+  Void		argument;
+  OnHalt	next;
+};
+
+
+void
+PL_on_halt(halt_function f, Void arg)
+{ if ( !GD->os.halting )
+  { OnHalt h = allocHeap(sizeof(struct on_halt));
+
+    h->function = f;
+    h->argument = arg;
+    startCritical;
+    h->next = GD->os.on_halt_list;
+    GD->os.on_halt_list = h;
+    endCritical;
+  }
+}
+
+
+int
+PL_cleanup(int rval)
+{ static int cleaning = FALSE;
+  OnHalt h;
+
+  LOCK();
+  if ( cleaning )
+  { UNLOCK();
+    return FALSE;
+  }
+  cleaning = TRUE;
+
+  pl_notrace();				/* avoid recursive tracing */
+#ifdef O_PROFILE
+  stopProfiler();			/* don't do profiling anymore */
+#endif
+#ifdef O_PLMT
+  exitPrologThreads();
+#endif
+
+  Scurout = Soutput;			/* reset output stream to user */
+
+  if ( !GD->os.halting++ )
+  { for(h = GD->os.on_halt_list; h; h = h->next)
+      (*h->function)(rval, h->argument);
+
+    if ( GD->initialised )
+    { fid_t cid = PL_open_foreign_frame();
+      predicate_t proc = PL_predicate("$run_at_halt", 0, "system");
+      PL_call_predicate(MODULE_system, FALSE, proc, 0);
+      PL_discard_foreign_frame(cid);
+    }
+
+#if defined(__WINDOWS__) || defined(__WIN32__)
+    if ( rval != 0 && !hasConsole() )
+      PlMessage("Exit status is %d", rval);
+#endif
+
+    qlfCleanup();			/* remove errornous .qlf files */
+    dieIO();				/* streams may refer to foreign code */
+					/* Standard I/O is only flushed! */
+
+    if ( GD->initialised )
+    { fid_t cid = PL_open_foreign_frame();
+      predicate_t proc = PL_predicate("unload_all_foreign_libraries", 0,
+				      "shlib");
+      if ( isDefinedProcedure(proc) )
+	PL_call_predicate(MODULE_system, FALSE, proc, 0);
+      PL_discard_foreign_frame(cid);
+    }
+
+    RemoveTemporaryFiles();
+  }
+
+  if ( GD->resourceDB )
+  { rc_close_archive(GD->resourceDB);
+    GD->resourceDB = NULL;
+  }
+
+  cleanupSignals();
+  freeStacks(LD);
+#ifdef HAVE_DMALLOC_H
+  dmalloc_verify(0);
+#endif
+  freeLocalData(LD);
+  cleanupSourceFiles();
+  cleanupAtoms();
+  cleanupFunctors();
+  cleanupArith();
+  cleanupMemAlloc();
+  cleanupInitialiseHooks();
+  cleanupExtensions();
+  cleanupOs();
+  Scleanup();
+
+  memset(GD, 0, sizeof(*GD));
+  memset(LD, 0, sizeof(*LD));
+
+  cleaning = FALSE;			/* prepare for another */
+  UNLOCK();
+
+  return TRUE;
+}
+
+		 /*******************************
+		 *	ERRORS AND WARNINGS	*
+		 *******************************/
+
 #include <stdarg.h>
 
 bool
@@ -757,50 +1129,58 @@ warning(const char *fm, ...)
 
 static bool
 vsysError(const char *fm, va_list args)
-{ Sfprintf(Serror, "[PROLOG INTERNAL ERROR:\n\t");
+{ static int active = 0;
+
+  if ( active++ )
+    PL_halt(3);
+
+#ifdef O_PLMT
+  Sfprintf(Serror, "[PROLOG SYSTEM ERROR:  Thread %d\n\t",
+	   PL_thread_self());
+#else
+  Sfprintf(Serror, "[PROLOG SYSTEM ERROR:\n\t");
+#endif
   Svfprintf(Serror, fm, args);
   if ( gc_status.active )
   { Sfprintf(Serror,
 	    "\n[While in %ld-th garbage collection; skipping stacktrace]\n",
 	    gc_status.collections);
   }
-  if ( GD->bootsession || !GD->initialised )
-  { Sfprintf(Serror,
-	     "\n[While initialising; quitting]\n");
-    Halt(1);
-  }
 
 #if defined(O_DEBUGGER)
   if ( !gc_status.active )
-  { Sfprintf(Serror, "\n[Switched to system mode: style_check(+dollar)]\n");
-    debugstatus.styleCheck |= DOLLAR_STYLE;
-    Sfprintf(Serror, "PROLOG STACK:\n");
+  { systemMode(TRUE);
+    Sfprintf(Serror, "\n\nPROLOG STACK:\n");
     backTrace(NULL, 10);
     Sfprintf(Serror, "]\n");
   }
 #endif /*O_DEBUGGER*/
 
+  if ( GD->bootsession )
+    PL_halt(1);
+
 action:
-  Sprintf("\nAction? "); Sflush(Soutput);
+  Sfprintf(Serror, "\nAction? "); Sflush(Soutput);
   ResetTty();
-  switch(getSingleChar())
+  switch(getSingleChar(Sinput))
   { case 'a':
-      pl_abort();
+      pl_abort(ABORT_FATAL);
       break;
     case EOF:
-      Sprintf("EOF: exit\n");
+      Sfprintf(Serror, "EOF: exit\n");
     case 'e':
-      Halt(3);
+      PL_halt(3);
       break;
     default:
-      Sprintf("Unknown action.  Valid actions are:\n"
+      Sfprintf(Serror,
+	       "Unknown action.  Valid actions are:\n"
 	      "\ta\tabort to toplevel\n"
 	      "\te\texit Prolog\n");
       goto action;
   }
 
-  pl_abort();
-  Halt(3);
+  pl_abort(ABORT_FATAL);
+  PL_halt(3);
   PL_fail;
 }
 
@@ -821,54 +1201,67 @@ vfatalError(const char *fm, va_list args)
   Sfprintf(Serror, "]\n");
 #endif
 
-  Halt(2);
+  PL_halt(2);
   PL_fail;
 }
 
 
+/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
+One day, warning() should be replaced by   PL_error()  or direct call to
+print_message/2. For now we make warning call print_message/2, so we can
+move the rest of the warnings gradually. For this reason we make a term
+
+	message_lines(ListOfLines)
+
+Where ListOfLines is a list of string objects.
+- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
+
 bool
 vwarning(const char *fm, va_list args)
-{ toldString();
+{ toldString();				/* play safe */
 
   if ( trueFeature(REPORT_ERROR_FEATURE) )
-  { if ( ReadingSource &&
-	 !GD->bootsession && GD->initialised &&
+  { if ( !GD->bootsession && GD->initialised &&
 	 !LD->outofstack )		/* cannot call Prolog */
-    { fid_t cid = PL_open_foreign_frame();
-      term_t argv = PL_new_term_refs(3);
-      predicate_t pred = PL_pred(FUNCTOR_exception3, MODULE_user);
-      term_t a = PL_new_term_ref();
-      char message[LINESIZ];
-      qid_t qid;
-      int rval;
-  
+    { char message[LINESIZ];
+      char *s = message;
+      fid_t cid   = PL_open_foreign_frame();
+      term_t av   = PL_new_term_refs(2);
+      term_t tail = PL_copy_term_ref(av+1);
+      term_t head = PL_new_term_ref();
+
       Svsprintf(message, fm, args);
-  
-      PL_put_atom(   argv+0, ATOM_warning);
-      PL_put_functor(argv+1, FUNCTOR_warning3);
-      PL_get_arg(1, argv+1, a); PL_unify_atom(a, source_file_name);
-      PL_get_arg(2, argv+1, a); PL_unify_integer(a, source_line_no);
-      PL_get_arg(3, argv+1, a); PL_unify_string_chars(a, message);
       
-      qid = PL_open_query(MODULE_user, PL_Q_NODEBUG, pred, argv);
-      rval = PL_next_solution(qid);
-      PL_close_query(qid);
-      PL_discard_foreign_frame(cid);
-  
-      if ( !rval )
-      { Sfprintf(Serror, "[WARNING: (%s:%d)\n\t%s]\n",
-		 stringAtom(source_file_name), source_line_no, message);
+      for(;;)
+      { char *eol = strchr(s, '\n');
+
+	if ( eol )
+	{ PL_unify_list(tail, head, tail);
+	  PL_unify_string_nchars(head, eol-s, s);
+	  s = eol+1;
+	} else
+	{ if ( *s )
+	  { PL_unify_list(tail, head, tail);
+	    PL_unify_string_chars(head, s);
+	  }
+	  PL_unify_nil(tail);
+	  break;
+	}
       }
-  
-      PL_fail;				/* handled */
+      PL_cons_functor(av+1, FUNCTOR_message_lines1, av+1);
+      PL_put_atom(av, ATOM_error);	/* error? */
+
+      PL_call_predicate(NULL, PL_Q_NODEBUG, PROCEDURE_print_message2, av);
+      PL_discard_foreign_frame(cid);
+    } else
+    { Sfprintf(Suser_error, "ERROR: ");
+      Svfprintf(Suser_error, fm, args);
+      Sfprintf(Suser_error, "\n");
+      Pause(0.5);
     }
-  
-    Sfprintf(Serror, "[WARNING: ");
-    Svfprintf(Serror, fm, args);
-    Sfprintf(Serror, "]\n");
   }
 
-  if ( trueFeature(DEBUG_ON_ERROR_FEATURE) )
+  if ( !ReadingSource && trueFeature(DEBUG_ON_ERROR_FEATURE) )
     pl_trace();
 
   PL_fail;
