@@ -1,10 +1,25 @@
-/*  pl-os.c,v 1.45 1994/04/11 08:37:37 jan Exp
+/*  $Id: pl-os.c,v 1.135 2002/02/01 16:49:13 jan Exp $
 
-    Copyright (c) 1990 Jan Wielemaker. All rights reserved.
-    See ../LICENCE to find out about your rights.
-    jan@swi.psy.uva.nl
+    Part of SWI-Prolog
 
-    Purpose: Operating System Dependencies
+    Author:        Jan Wielemaker
+    E-mail:        jan@swi.psy.uva.nl
+    WWW:           http://www.swi-prolog.org
+    Copyright (C): 1985-2002, University of Amsterdam
+
+    This library is free software; you can redistribute it and/or
+    modify it under the terms of the GNU Lesser General Public
+    License as published by the Free Software Foundation; either
+    version 2.1 of the License, or (at your option) any later version.
+
+    This library is distributed in the hope that it will be useful,
+    but WITHOUT ANY WARRANTY; without even the implied warranty of
+    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
+    Lesser General Public License for more details.
+
+    You should have received a copy of the GNU Lesser General Public
+    License along with this library; if not, write to the Free Software
+    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
 
 /*  Modified (M) 1993 Dave Sherratt  */
@@ -18,7 +33,8 @@
 #include <os2.h>                /* this has to appear before pl-incl.h */
 #endif
 
-#include <math.h>			/* avoid abs() problem with msvc++ */
+#include <math.h>		/* avoid abs() problem with msvc++ */
+#include <stdio.h>		/* rename() and remove() prototypes */
 #include "pl-incl.h"
 #include "pl-ctype.h"
 
@@ -37,12 +53,15 @@
 #ifdef HAVE_UNISTD_H
 #include <unistd.h>
 #endif
-#ifdef HAVE_SYS_PARAM_H
-#include <sys/param.h>
-#endif
 #ifdef HAVE_SYS_FILE_H
 #include <sys/file.h>
 #endif
+#if defined(HAVE_SYS_RESOURCE_H)
+#include <sys/resource.h>
+#endif
+#ifdef HAVE_FTIME
+#include <sys/timeb.h>
+#endif
 
 #include <fcntl.h>
 #ifndef __WATCOMC__			/* appears a conflict */
@@ -58,11 +77,12 @@
 static real initial_time;
 #endif /* OS2 */
 
-forwards void	initExpand(void);
-forwards void	initRandom(void);
-forwards void	initEnviron(void);
-forwards long	Time(void);
-static void	RemoveTemporaryFiles(void);
+static void	initExpand(void);
+static void	cleanupExpand(void);
+static void	initRandom(void);
+static void	initEnviron(void);
+static long	Time(void);
+static char *	Which(const char *program, char *fullname);
 
 #ifndef DEFAULT_PATH
 #define DEFAULT_PATH "/bin:/usr/bin"
@@ -135,75 +155,13 @@ initOs(void)
   succeed;
 }
 
-typedef void (*halt_function)(int, Void);
-
-struct on_halt
-{ halt_function	function;
-  Void		argument;
-  OnHalt	next;
-};
-
 
 void
-PL_on_halt(halt_function f, Void arg)
-{ if ( !GD->os.halting )
-  { OnHalt h = allocHeap(sizeof(struct on_halt));
-
-    h->function = f;
-    h->argument = arg;
-    startCritical;
-    h->next = GD->os.on_halt_list;
-    GD->os.on_halt_list = h;
-    endCritical;
-  }
+cleanupOs(void)
+{ cleanupExpand();
 }
 
 
-volatile void
-Halt(int rval)
-{ OnHalt h;
-  extern int Output;
-
-  pl_notrace();				/* avoid recursive tracing */
-  Output = 1;				/* reset output stream to user */
-
-  if ( !GD->os.halting )
-  { GD->os.halting++;			/* avoid recursion */
-
-    for(h = GD->os.on_halt_list; h; h = h->next)
-      (*h->function)(rval, h->argument);
-
-    if ( GD->initialised )
-    { fid_t cid = PL_open_foreign_frame();
-      predicate_t proc = PL_predicate("$run_at_halt", 0, "system");
-      PL_call_predicate(MODULE_system, FALSE, proc, 0);
-      PL_discard_foreign_frame(cid);
-    }
-
-#if defined(__WINDOWS__) || defined(__WIN32__)
-    if ( rval != 0 )
-      PlMessage("Exit status is %d", rval);
-#endif
-
-    qlfCleanup();			/* remove errornous .qlf files */
-    dieIO();
-
-    if ( GD->initialised )
-    { fid_t cid = PL_open_foreign_frame();
-      predicate_t proc = PL_predicate("unload_all_foreign_libraries", 0,
-				      "shlib");
-      if ( isDefinedProcedure(proc) )
-	PL_call_predicate(MODULE_system, FALSE, proc, 0);
-      PL_discard_foreign_frame(cid);
-    }
-
-    RemoveTemporaryFiles();
-  }
-
-  exit(rval);
-  /*NOTREACHED*/
-}
-
 		/********************************
 		*            OS ERRORS          *
 		*********************************/
@@ -225,7 +183,7 @@ OsError(void)
 #else /*HAVE_STRERROR*/
 static char errmsg[64];
 
-#if unix
+#ifdef __unix__
   extern int sys_nerr;
 #if !EMX
   extern char *sys_errlist[];
@@ -246,7 +204,7 @@ static char errmsg[64];
 		*********************************/
 
 /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
-    real CpuTime()
+    double CpuTime(cputime_kind)
 
     Returns a floating point number, representing the amount  of  (user)
     CPU-seconds  used  by the process Prolog is in.  For systems that do
@@ -273,21 +231,26 @@ static char errmsg[64];
 #endif /*HAVE_TIMES*/
 
 
-real
-CpuTime(void)
+double
+CpuTime(cputime_kind which)
 {
 #ifdef HAVE_TIMES
   struct tms t;
   static int MTOK_got_hz = FALSE;
-  static real MTOK_hz;
+  static double MTOK_hz;
 
   if ( !MTOK_got_hz )
-  { MTOK_hz = (real) Hz;
+  { MTOK_hz = (double) Hz;
     MTOK_got_hz++;
   }
   times(&t);
 
-  return (real) t.tms_utime / MTOK_hz;
+  switch( which )
+  { case CPU_USER:
+      return (double) t.tms_utime / MTOK_hz;
+    case CPU_SYSTEM:
+      return (double) t.tms_stime / MTOK_hz;
+  }
 #endif
 
 #if OS2 && EMX
@@ -303,65 +266,81 @@ CpuTime(void)
 #ifdef HAVE_CLOCK
   return (real) (clock() - clock_wait_ticks) / (real) CLOCKS_PER_SEC;
 #endif
+
+  return 0.0;
 }
 
 #endif /*__WIN32__*/
 
-#ifdef HAVE_CLOCK
 void
 PL_clock_wait_ticks(long waited)
-{ clock_wait_ticks += waited;
+{
+#ifdef HAVE_CLOCK
+  clock_wait_ticks += waited;
+#endif
 }
+
+
+double
+WallTime(void)
+{ double stime;
+
+#ifdef HAVE_GETTIMEOFDAY
+  struct timeval tp;
+
+  gettimeofday(&tp, NULL);
+  stime = (double)tp.tv_sec + (double)tp.tv_usec/1000000.0;
+#else
+#ifdef HAVE_FTIME
+  struct timeb tb;
+
+  ftime(&tb);
+  stime = (double)tb.time + (double)tb.millitm/1000.0;
+#else
+  stime = (double)time((time_t *)NULL);
+#endif
 #endif
 
-		/********************************
-		*       MEMORY MANAGEMENT       *
-		*********************************/
-
-/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
-    long *Allocate(n)
-	  long n;
-
-    Allocate a memory area of `n' bytes from the operating system.   `n'
-    is a long as we need to allocate one uniform array of longs for both
-    the  local  stack  and  global  stack,  which  implies  it should be
-    possible to allocate at least a few hundred Kbytes.  If  you  cannot
-    implement  this  function  you  are in deep trouble.  You either can
-    decide to redesign large part of the data representation, or  forget
-    about  SWI-Prolog.   Memory  is never returned to the system.  As it
-    would only concern small areas,  all  over  SWI-Prolog's  memory  no
-    currently  available operating system (I'm aware of) will be able to
-    handle it anyway.  THE RETURN VALUE SHOULD BE ROUNDED TO BE A  VALID
-    POINTER FOR LONGS AND STRUCTURES AND AT LEAST A MULTIPLE OF 4.
-- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
-
-Void
-Allocate(long n)
-{ Void mem = malloc(n);
-
-  return (Void) mem;
+  return stime;
 }
 
+		 /*******************************
+		 *	       MEMORY		*
+		 *******************************/
 
-		/********************************
-		*     STRING MANIPULATION	*
-		********************************/
+ulong
+UsedMemory(void)
+{
+#ifdef HAVE_GETRUSAGE
+  struct rusage usage;
 
-/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
-The builtin strcmp() of SunOs is broken on some machines ...
-- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
-
-#if sun
-int
-strcmp(const char *s1, const char *s2)
-{ while(*s1 && *s1 == *s2)
-    s1++, s2++;
-
-  return *(const unsigned char *)s1 -
-	 *(const unsigned char *)s2;
-}
+  if ( getrusage(RUSAGE_SELF, &usage) == 0 &&
+       usage.ru_idrss )
+  { return usage.ru_idrss;		/* total unshared data */
+  }
 #endif
 
+  return (GD->statistics.heap +
+	  usedStack(global) +
+	  usedStack(local) +
+	  usedStack(trail));
+}
+
+
+ulong
+FreeMemory(void)
+{ ulong used = UsedMemory();
+
+#if defined(HAVE_GETRLIMIT) && defined(RLIMIT_DATA)
+  struct rlimit limit;
+
+  if ( getrlimit(RLIMIT_DATA, &limit) == 0 )
+    return limit.rlim_cur - used;
+#endif
+
+  return 0L;
+}
+
 
 		/********************************
 		*           ARITHMETIC          *
@@ -447,12 +426,14 @@ atom_t
 TemporaryFile(const char *id)
 { char temp[MAXPATHLEN];
   TempFile tf = allocHeap(sizeof(struct tempfile));
+  char envbuf[MAXPATHLEN];
   char *tmpdir;
 
-  if ( !((tmpdir = getenv("TEMP")) || (tmpdir = getenv("TMP"))) )
+  if ( !((tmpdir = getenv3("TEMP", envbuf, sizeof(envbuf))) ||
+	 (tmpdir = getenv3("TMP",  envbuf, sizeof(envbuf)))) )
     tmpdir = DEFTMPDIR;
 
-#if unix
+#ifdef __unix__
 { static int MTOK_temp_counter = 0;
 
   Ssprintf(temp, "%s/pl_%s_%d_%d",
@@ -464,7 +445,11 @@ TemporaryFile(const char *id)
 { char *tmp;
   static int temp_counter = 0;
 
+#ifdef __LCC__
+  if ( (tmp = tmpnam(NULL)) )
+#else
   if ( (tmp = _tempnam(tmpdir, id)) )
+#endif
   { PrologPath(tmp, temp);
   } else
     Ssprintf(temp, "%s/pl_%s_%d", tmpdir, id, temp_counter++);
@@ -486,7 +471,7 @@ TemporaryFile(const char *id)
   tmpnam(temp);
 #endif
 
-  tf->name = lookupAtom(temp);
+  tf->name = PL_new_atom(temp);		/* locked: ok! */
   tf->next = NULL;
   
   startCritical;
@@ -501,8 +486,8 @@ TemporaryFile(const char *id)
   return tf->name;
 }
 
-static void
-RemoveTemporaryFiles()
+void
+RemoveTemporaryFiles(void)
 { TempFile tf, tf2;  
 
   startCritical;
@@ -523,15 +508,6 @@ macros to deal with 16-bit machines, but are not  defined  as  functions
 here.   Some  more  specific things SWI-Prolog wants to know about files
 are defined here:
 
-    int  getdtablesize()
-
-    SWI-Prolog assumes it can refer to open i/o streams via  read()  and
-    write() by small integers, returned by open(). These integers should
-    be  in  the range [0, ..., getdtablesize()). If your system does not
-    do this you better redefine the Open(), Read() and Write() macros so
-    they  do  meet  this  requirement.   Prolog  allocates  a  table  of
-    structures with getdtablesize() entries.
-
     long LastModifiedFile(path)
 	 char *path;
 
@@ -580,32 +556,6 @@ are defined here:
     compiler and the creation of stand-alone executables.
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
-#ifndef HAVE_GETDTABLESIZE
-int
-getdtablesize()
-{
-#ifdef OPEN_MAX
-  return OPEN_MAX;
-#else
-#ifdef _SC_OPEN_MAX			/* POSIX, USG */
-  return sysconf(_SC_OPEN_MAX);
-#else
-#ifdef HAVE_GETRLIMIT
-#ifdef HAVE_SYS_RESOURCE_H
-#include <sys/resource.h>
-#endif
-#ifdef RLIMIT_NOFILE
-  { struct rlimit rlp;
-    getrlimit(RLIMIT_NOFILE,&rlp);
-    return (rlp.rlim_cur);
-  }
-#endif /*RLIMIT_NOFILE*/
-#endif /*HAVE_GETRLIMIT*/
-#endif /*_SC_OPEN_MAX*/
-#endif /*OPEN_MAX*/
-}
-#endif /*HAVE_GETDTABLESIZE*/
-
 /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 Size of a VM page of memory.  Most BSD machines have this function.  If not,
 here are several alternatives ...
@@ -685,7 +635,7 @@ OsPath(const char *plpath, char *path)
 } 
 #endif /* O_HPFS */
 
-#if unix
+#ifdef __unix__
 char *
 PrologPath(const char *p, char *buf)
 { strcpy(buf, p);
@@ -699,7 +649,7 @@ OsPath(const char *p, char *buf)
 
   return buf;
 }
-#endif /*unix*/
+#endif /*__unix__*/
 
 #if O_XOS
 char *
@@ -897,6 +847,7 @@ RenameFile(const char *old, const char *new)
   osnew = OsPath(new, newbuf);
 
 #ifdef HAVE_RENAME
+  remove(osnew);			/* assume we have this too */
   return rename(osold, osnew) == 0 ? TRUE : FALSE;
 #else
 { int rval;
@@ -924,7 +875,7 @@ SameFile(const char *f1, const char *f2)
       succeed;
   }
 
-#ifdef unix				/* doesn't work on most not Unix's */
+#ifdef __unix__				/* doesn't work on most not Unix's */
   { struct stat buf1;
     struct stat buf2;
     char tmp[MAXPATHLEN];
@@ -972,8 +923,8 @@ MarkExecutable(const char *name)
   { term_t file = PL_new_term_ref();
 
     PL_put_atom_chars(file, name);
-    PL_error(NULL, 0, OsError(), ERR_FILE_OPERATION,
-	     ATOM_stat, ATOM_file, file);
+    return PL_error(NULL, 0, OsError(), ERR_FILE_OPERATION,
+		    ATOM_stat, ATOM_file, file);
   }
 
   if ( (buf.st_mode & 0111) == (~um & 0111) )
@@ -984,8 +935,8 @@ MarkExecutable(const char *name)
   { term_t file = PL_new_term_ref();
 
     PL_put_atom_chars(file, name);
-    PL_error(NULL, 0, OsError(), ERR_FILE_OPERATION,
-	     ATOM_chmod, ATOM_file, file);
+    return PL_error(NULL, 0, OsError(), ERR_FILE_OPERATION,
+		    ATOM_chmod, ATOM_file, file);
   }
 #endif /* defined(HAVE_STAT) && defined(HAVE_CHMOD) */
 
@@ -1045,7 +996,9 @@ initExpand(void)
   CWDlen = 0;
 
 #ifdef O_CANONISE_DIRS
-  if ( (cpaths = getenv("CANONICAL_PATHS")) )
+{ char envbuf[MAXPATHLEN];
+
+  if ( (cpaths = getenv3("CANONICAL_PATHS", envbuf, sizeof(envbuf))) )
   { char buf[MAXPATHLEN];
 
     while(*cpaths)
@@ -1065,14 +1018,65 @@ initExpand(void)
     }
   }
 
-  if ( (dir = getenv("HOME")) ) canoniseDir(dir);
-  if ( (dir = getenv("PWD"))  ) canoniseDir(dir);
-  if ( (dir = getenv("CWD"))  ) canoniseDir(dir);
+  if ( (dir = getenv3("HOME", envbuf, sizeof(envbuf))) ) canoniseDir(dir);
+  if ( (dir = getenv3("PWD",  envbuf, sizeof(envbuf))) ) canoniseDir(dir);
+  if ( (dir = getenv3("CWD",  envbuf, sizeof(envbuf))) ) canoniseDir(dir);
+}
 #endif
 }
 
 #ifdef O_CANONISE_DIRS
 
+static void
+cleanupExpand(void)
+{ CanonicalDir dn = canonical_dirlist, next;
+
+  canonical_dirlist = NULL;
+  for( ; dn; dn = next )
+  { next = dn->next;
+    free(dn);
+  }
+}
+
+
+static void
+registerParentDirs(const char *path)
+{ const char *e = path + strlen(path);
+
+  while(e>path)
+  { char dirname[MAXPATHLEN];
+    char tmp[MAXPATHLEN];
+    CanonicalDir d;
+    struct stat buf;
+
+    for(e--; *e != '/' && e > path + 1; e-- )
+      ;
+
+    strncpy(dirname, path, e-path);
+    dirname[e-path] = EOS;
+
+    for(d = canonical_dirlist; d; d = d->next)
+    { if ( streq(d->name, dirname) )
+	return;
+    }
+	
+    if ( statfunc(OsPath(dirname, tmp), &buf) == 0 )
+    { CanonicalDir dn   = malloc(sizeof(*dn));
+
+      dn->next		= canonical_dirlist;
+      dn->name		= store_string(dirname);
+      dn->inode		= buf.st_ino;
+      dn->device	= buf.st_dev;
+      dn->canonical	= dn->name;
+      canonical_dirlist	= dn;
+
+      DEBUG(1, Sdprintf("Registered canonical dir %s\n", dirname));
+    } else
+      return;
+  }
+}
+
+
 static char *
 canoniseDir(char *path)
 { CanonicalDir d;
@@ -1091,8 +1095,13 @@ canoniseDir(char *path)
     }
   }
 
+					/* we need to use malloc() here */
+					/* because allocHeap() only ensures */
+					/* alignment for `word', and inode_t */
+					/* is sometimes bigger! */
+
   if ( statfunc(OsPath(path, tmp), &buf) == 0 )
-  { CanonicalDir dn = allocHeap(sizeof(struct canonical_dir));
+  { CanonicalDir dn = malloc(sizeof(*dn));
     char dirname[MAXPATHLEN];
     char *e = path + strlen(path);
 
@@ -1121,20 +1130,22 @@ canoniseDir(char *path)
 	  strcat(dirname, e);
 	  strcpy(path, dirname);
 	  dn->canonical = store_string(path);
+	  canonical_dirlist = dn;
 	  DEBUG(1, Sdprintf("(replace) %s\n", path));
+	  registerParentDirs(path);
 	  return path;
 	}
       }
 
       for(e--; *e != '/' && e > path + 1; e-- )
 	;
-
     } while( e > path );
 
     dn->canonical = dn->name;
     canonical_dirlist = dn;
 
     DEBUG(1, Sdprintf("(new, existing) %s\n", path));
+    registerParentDirs(path);
     return path;
   }
 
@@ -1146,6 +1157,11 @@ canoniseDir(char *path)
 
 #define canoniseDir(d)
 
+static void
+cleanupExpand(void)
+{
+}
+
 #endif /*O_CANONISE_DIRS*/
 
 
@@ -1157,6 +1173,8 @@ canoniseFileName(char *path)
 
   while( in[0] == '/' && in[1] == '.' && in[2] == '.' && in[3] == '/' )
     in += 3;
+  while( in[0] == '.' && in[1] == '/' )
+    in += 2;
   if ( in[0] == '/' )
     *out++ = '/';
   osave[osavep++] = out;
@@ -1166,17 +1184,27 @@ canoniseFileName(char *path)
     {
     again:
       if ( *in )
-      { while( in[1] == '/' )
+      { while( in[1] == '/' )		/* delete multiple / */
 	  in++;
-	if ( in[1] == '.' && (in[2] == '/' || in[2] == EOS) )
-	{ in += 2;
-	  goto again;
-	}
-	if ( in[1] == '.' && in[2] == '.' &&
-	     (in[3] == '/' || in[3] == EOS) && osavep > 0 )
-	{ out = osave[--osavep];
-	  in += 3;
-	  goto again;
+	if ( in[1] == '.' )
+	{ if ( in[2] == '/' )		/* delete /./ */
+	  { in += 2;
+	    goto again;
+	  }
+	  if ( in[2] == EOS )		/* delete trailing /. */
+	  { *out = EOS;
+	    return path;
+	  }
+	  if ( in[2] == '.' &&		/* delete /foo/../ */
+	       (in[3] == '/' || in[3] == EOS) && osavep > 0 )
+	  { out = osave[--osavep];
+	    in += 3;
+	    if ( in[0] == EOS && out > path )
+	    { out[-1] = EOS;		/* delete trailing / */
+	      return path;
+	    }
+	    goto again;
+	  }
 	}
       }
       if ( *in )
@@ -1259,7 +1287,9 @@ expandVars(const char *pattern, char *expanded)
       value = _xos_home();
 #else /*O_XOS*/
       if ( !(value = GD->os.myhome) )
-      { if ( (value = getenv("HOME")) )
+      { char envbuf[MAXPATHLEN];
+
+	if ( (value = getenv3("HOME", envbuf, sizeof(envbuf))) )
 	{ value = GD->os.myhome = store_string(PrologPath(value, word));
 	} else
 	{ value = GD->os.myhome = "/";
@@ -1310,8 +1340,9 @@ expandVars(const char *pattern, char *expanded)
     { case EOS:
 	break;
       case '$':
-	{ char *var = takeWord(&pattern, word);
-	  char *value = getenv(var);
+	{ char envbuf[MAXPATHLEN];
+	  char *var = takeWord(&pattern, word);
+	  char *value = getenv3(var, envbuf, sizeof(envbuf));
 	  int l;
 
 	  if ( value == (char *) NULL )
@@ -1461,11 +1492,12 @@ char *
 AbsoluteFile(const char *spec, char *path)
 { char tmp[MAXPATHLEN];
   char buf[MAXPATHLEN];
-  char *file;  
-
-  PrologPath(spec, buf);
-  if ( !(file = ExpandOneFile(buf, tmp)) )
-    return (char *) NULL;
+  char *file = PrologPath(spec, buf);
+  
+  if ( trueFeature(FILEVARS_FEATURE) )
+  { if ( !(file = ExpandOneFile(buf, tmp)) )
+      return (char *) NULL;
+  }
 
   if ( IsAbsolutePath(file) )
   { strcpy(path, file);
@@ -1486,7 +1518,36 @@ AbsoluteFile(const char *spec, char *path)
   }
 #endif /*O_HASDRIVES*/
 
-  if ( CWDlen == 0 )
+  if ( !PL_cwd() )
+    return NULL;
+
+  if ( (CWDlen + strlen(file) + 1) >= MAXPATHLEN )
+  { PL_error(NULL, 0, NULL, ERR_REPRESENTATION, ATOM_max_path_length);
+    return (char *) NULL;
+  }
+  
+  strcpy(path, CWDdir);
+  if ( file[0] != EOS )
+    strcpy(&path[CWDlen], file);
+  if ( strchr(file, '.') || strchr(file, '/') )
+    return canonisePath(path);
+  else
+    return path;
+}
+
+
+void
+PL_changed_cwd(void)
+{ if ( CWDdir )
+    remove_string(CWDdir);
+  CWDdir = NULL;
+  CWDlen = 0;
+}
+
+
+const char *
+PL_cwd(void)
+{ if ( CWDlen == 0 )
   { char buf[MAXPATHLEN];
     char *rval;
 
@@ -1507,6 +1568,8 @@ to be implemented directly.  What about other Unixes?
       PL_put_atom(tmp, ATOM_dot);
       PL_error(NULL, 0, OsError(), ERR_FILE_OPERATION,
 	       ATOM_getcwd, ATOM_directory, tmp);
+
+      return NULL;
     }
 
     canonisePath(buf);
@@ -1519,31 +1582,20 @@ to be implemented directly.  What about other Unixes?
     CWDdir = store_string(buf);
   }
 
-  if ( (CWDlen + strlen(file) + 1) >= MAXPATHLEN )
-  { PL_error(NULL, 0, NULL, ERR_REPRESENTATION, ATOM_max_path_length);
-    return (char *) NULL;
-  }
-  
-  strcpy(path, CWDdir);
-  if ( file[0] != EOS )
-    strcpy(&path[CWDlen], file);
-  if ( strchr(file, '.') || strchr(file, '/') )
-    return canonisePath(path);
-  else
-    return path;
+  return (const char *)CWDdir;
 }
 
 
 char *
-BaseName(register char *f)
-{ register char *base;
+BaseName(const char *f)
+{ const char *base;
 
   for(base = f; *f; f++)
   { if (*f == '/')
       base = f+1;
   }
 
-  return base;
+  return (char *)base;
 }
 
 
@@ -1553,10 +1605,17 @@ DirName(const char *f, char *dir)
 
   for(base = p = f; *p; p++)
   { if (*p == '/' && p[1] != EOS )
-      base = p+1;
+      base = p;
+  }
+  if ( base == f )
+  { if ( *f == '/' )
+      strcpy(dir, "/");
+    else
+      strcpy(dir, ".");
+  } else
+  { strncpy(dir, f, base-f);
+    dir[base-f] = EOS;
   }
-  strncpy(dir, f, base-f);
-  dir[base-f] = EOS;
   
   return dir;
 }
@@ -1591,7 +1650,7 @@ DeRefLink1(const char *f, char *lbuf)
     { char *q;
 
       strcpy(lbuf, f);
-      q = &lbuf[strlen(buf)];
+      q = &lbuf[strlen(lbuf)];
       while(q>lbuf && q[-1] != '/')
 	q--;
       strcpy(q, l);
@@ -1665,7 +1724,6 @@ ChDir(const char *path)
 }
 
 
-
 		/********************************
 		*        TIME CONVERSION        *
 		*********************************/
@@ -1717,10 +1775,11 @@ ResetStdin()
     Clear the Sinput buffer after a saved state.  Only necessary
     if O_SAVE is defined.
 
-PushTty()
-    Push the tty to the specified state.
+PushTty(IOSTREAM *s, ttybuf *buf, int state)
+    Push the tty to the specified state and save the old state in
+    buf.
 
-PopTty()
+PopTty(IOSTREAM *s, ttybuf *buf)
     Restore the tty state.
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
@@ -1751,31 +1810,34 @@ Swrite_protocol(void *handle, char *buf, int size)
 
 int
 Sread_terminal(void *handle, char *buf, int size)
-{ long h     = (long)handle;
-  atom_t sfn = source_file_name;		/* save over call-back */
-  int    sln = source_line_no;
-  int     fd = (int)h;
+{ long h = (long)handle;
+  int fd = (int)h;
+  source_location oldsrc = LD->read_source;
 
-  if ( GD->os.prompt_next && ttymode != TTY_RAW )
-  { Putf("%s", PrologPrompt());
+  if ( LD->prompt.next && ttymode != TTY_RAW )
+  { Sfputs(PrologPrompt(), Suser_output);
     
-    GD->os.prompt_next = FALSE;
+    LD->prompt.next = FALSE;
   }
 
-  pl_ttyflush();
+  Sflush(Suser_output);
   PL_dispatch(fd, PL_DISPATCH_WAIT);
   size = (*GD->os.org_terminal.read)(handle, buf, size);
 
   if ( size == 0 )			/* end-of-file */
   { if ( fd == 0 )
-    { Sclearerr(Sinput);
-      GD->os.prompt_next = TRUE;
+    { Sclearerr(Suser_input);
+      LD->prompt.next = TRUE;
     }
   } else if ( size > 0 && buf[size-1] == '\n' )
-    GD->os.prompt_next = TRUE;
+    LD->prompt.next = TRUE;
 
-  source_line_no   = sln;
-  source_file_name = sfn;
+  LD->read_source = oldsrc;
+
+#ifdef WIN32
+  if ( size > 0 )			/* Why needed only here? */
+    protocol(buf, size);		/* and what is the correct condition */
+#endif
 
   return size;
 }
@@ -1794,7 +1856,7 @@ ResetTty()
     Soutput->functions = 
     Serror->functions  = &GD->os.iofunctions;
   }
-  GD->os.prompt_next = TRUE;
+  LD->prompt.next = TRUE;
   endCritical;
 }
 
@@ -1810,20 +1872,23 @@ ResetTty()
 #endif
 
 bool
-PushTty(ttybuf *buf, int mode)
+PushTty(IOSTREAM *s, ttybuf *buf, int mode)
 { struct termios tio;
+  int fd;
 
   buf->mode = ttymode;
   ttymode = mode;
 
-  if ( GD->cmdline.notty )
+  if ( (fd = Sfileno(s)) < 0 || !isatty(fd) )
+    succeed;				/* not a terminal */
+  if ( !trueFeature(TTY_CONTROL_FEATURE) )
     succeed;
 
 #ifdef HAVE_TCSETATTR 
-  if ( tcgetattr(0, &buf->tab) )	/* save the old one */
+  if ( tcgetattr(fd, &buf->tab) )	/* save the old one */
     fail;
 #else
-  if ( ioctl(0, TIOCGETA, &buf->tab) )	/* save the old one */
+  if ( ioctl(fd, TIOCGETA, &buf->tab) )	/* save the old one */
     fail;
 #endif
 
@@ -1837,8 +1902,11 @@ PushTty(ttybuf *buf, int mode)
 	tio.c_lflag |= ISIG;
 #else
 	tio.c_lflag &= ~(ECHO|ICANON);
-	tio.c_cc[VTIME] = 0, tio.c_cc[VMIN] = 1;
 #endif
+					/* OpenBSD requires this anyhow!? */
+					/* Bug in OpenBSD or must we? */
+					/* Could this do any harm? */
+	tio.c_cc[VTIME] = 0, tio.c_cc[VMIN] = 1;
 	break;
     case TTY_OUTPUT:
 	tio.c_oflag |= (OPOST|ONLCR);
@@ -1851,7 +1919,7 @@ PushTty(ttybuf *buf, int mode)
   }
 
 #ifdef HAVE_TCSETATTR
-  if ( tcsetattr(0, TCSANOW, &tio) != 0 )
+  if ( tcsetattr(fd, TCSANOW, &tio) != 0 )
   { static int MTOK_warned;			/* MT-OK */
 
     if ( !MTOK_warned++ )
@@ -1859,10 +1927,10 @@ PushTty(ttybuf *buf, int mode)
   }
 #else
 #ifdef TIOCSETAW
-  ioctl(0, TIOCSETAW, &tio);
+  ioctl(fd, TIOCSETAW, &tio);
 #else
-  ioctl(0, TCSETAW, &tio);
-  ioctl(0, TCXONC, (void *)1);
+  ioctl(fd, TCSETAW, &tio);
+  ioctl(fd, TCXONC, (void *)1);
 #endif
 #endif
 
@@ -1871,20 +1939,23 @@ PushTty(ttybuf *buf, int mode)
 
 
 bool
-PopTty(ttybuf *buf)
-{ ttymode = buf->mode;
+PopTty(IOSTREAM *s, ttybuf *buf)
+{ int fd;
+  ttymode = buf->mode;
 
-  if ( GD->cmdline.notty )
+  if ( (fd = Sfileno(s)) < 0 || !isatty(fd) )
+    succeed;				/* not a terminal */
+  if ( !trueFeature(TTY_CONTROL_FEATURE) )
     succeed;
 
 #ifdef HAVE_TCSETATTR
-  tcsetattr(0, TCSANOW, &buf->tab);
+  tcsetattr(fd, TCSANOW, &buf->tab);
 #else
 #ifdef TIOCSETA
-  ioctl(0, TIOCSETA, &buf->tab);
+  ioctl(fd, TIOCSETA, &buf->tab);
 #else
-  ioctl(0, TCSETA, &buf->tab);
-  ioctl(0, TCXONC, (void *)1);
+  ioctl(fd, TCSETA, &buf->tab);
+  ioctl(fd, TCXONC, (void *)1);
 #endif
 #endif
 
@@ -1896,16 +1967,19 @@ PopTty(ttybuf *buf)
 #ifdef HAVE_SGTTYB
 
 bool
-PushTty(ttybuf *buf, int mode)
+PushTty(IOSTREAM *s, ttybuf *buf, int mode)
 { struct sgttyb tio;
+  int fd;
 
   buf->mode = ttymode;
   ttymode = mode;
 
-  if ( GD->cmdline.notty )
+  if ( (fd = Sfileno(s)) < 0 || !isatty(fd) )
+    succeed;				/* not a terminal */
+  if ( !trueFeature(TTY_CONTROL_FEATURE) )
     succeed;
 
-  if ( ioctl(0, TIOCGETP, &buf->tab) )  /* save the old one */
+  if ( ioctl(fd, TIOCGETP, &buf->tab) )  /* save the old one */
     fail;
   tio = buf->tab;
 
@@ -1925,22 +1999,25 @@ PushTty(ttybuf *buf, int mode)
       }
   
   
-  ioctl(0, TIOCSETP,  &tio);
-  ioctl(0, TIOCSTART, NULL);
+  ioctl(fd, TIOCSETP,  &tio);
+  ioctl(fd, TIOCSTART, NULL);
 
   succeed;
 }
 
 
 bool
-PopTty(ttybuf *buf)
+PopTty(IOSTREAM *s, ttybuf *buf)
 { ttymode = buf->mode;
+  int fd;
 
-  if ( GD->cmdline.notty )
+  if ( (fd = Sfileno(s)) < 0 || !isatty(fd) )
+    succeed;				/* not a terminal */
+  if ( !trueFeature(TTY_CONTROL_FEATURE) )
     succeed;
 
-  ioctl(0, TIOCSETP,  &buf->tab);
-  ioctl(0, TIOCSTART, NULL);
+  ioctl(fd, TIOCSETP,  &buf->tab);
+  ioctl(fd, TIOCSTART, NULL);
 
   succeed;
 }
@@ -1948,9 +2025,7 @@ PopTty(ttybuf *buf)
 #else /*HAVE_SGTTYB*/
 
 bool
-PushTty(buf, mode)
-ttybuf *buf;
-int mode;
+PushTty(IOSTREAM *s, ttybuf *buf, int mode)
 { buf->mode = ttymode;
   ttymode = mode;
 
@@ -1959,11 +2034,10 @@ int mode;
 
 
 bool
-PopTty(buf)
-ttybuf *buf;
+PopTty(IOSTREAM *s, ttybuf *buf)
 { ttymode = buf->mode;
   if ( ttymode != TTY_RAW )
-    GD->os.prompt_next = TRUE;
+    LD->prompt.next = TRUE;
 
   succeed;
 }
@@ -1996,25 +2070,58 @@ requested via getenv/2 from Prolog.  Functions
     value, or NULL if the variable did not exist.
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
-#if HAVE_PUTENV
-
+#ifndef __WIN32__
 char *
-Setenv(char *name, char *value)
-{ char *rval;
-  char buf[256];
+getenv3(const char *name, char *buf, int len)
+{ char *s = getenv(name);
 
-  if ( (rval = getenv(name)) )
-    rval = store_string(rval);
-  Ssprintf(buf, "%s=%s", name, value);
-  if ( putenv(store_string(buf)) < 0 )
-    PL_error("setenv", 2, NULL, ERR_NOMEM);
+  if ( s && strlen(s) < len )
+  { strcpy(buf, s);
 
-  return rval;
+    return buf;
+  }
+
+  return NULL;
 }
 
-char *
+int
+getenvl(const char *name)
+{ char *s;
+
+  if ( (s = getenv(name)) )
+    return strlen(s);
+
+  return -1;
+}
+#endif
+
+#if HAVE_PUTENV
+
+int
+Setenv(char *name, char *value)
+{ char *buf = alloca(strlen(name) + strlen(value) + 2);
+
+  if ( buf )
+  { Ssprintf(buf, "%s=%s", name, value);
+
+    if ( putenv(store_string(buf)) < 0 )
+      return PL_error("setenv", 2, NULL, ERR_NOMEM);
+  } else
+    return PL_error("setenv", 2, NULL, ERR_NOMEM);
+
+  succeed;
+}
+
+int
 Unsetenv(char *name)
-{ return Setenv(name, "");
+{
+#ifdef HAVE_UNSETENV
+  unsetenv(name);
+
+  succeed;
+#else
+  return Setenv(name, "");
+#endif
 }
 
 static void
@@ -2169,11 +2276,8 @@ Unsetenv(char *name)
     an alternative.
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
-#ifdef unix
+#ifdef __unix__
 #define SPECIFIC_SYSTEM 1
-#if defined(HAVE_SYS_RESOURCE_H)
-#include <sys/resource.h>
-#endif
 #if defined(HAVE_SYS_WAIT_H) || defined(UNION_WAIT)
 #include <sys/wait.h>
 #endif
@@ -2209,6 +2313,8 @@ System(char *cmd)
   int rval;
   void (*old_int)();
   void (*old_stop)();
+  unsigned char fds[256];
+  int nfds = openFileDescriptors(fds, sizeof(fds));
 
   Setenv("PROLOGCHILD", "yes");
 
@@ -2217,8 +2323,12 @@ System(char *cmd)
   } else if ( pid == 0 )		/* The child */
   { int i;
 
-    for(i = 3; i < getdtablesize(); i++)
-      close(i);
+    for(i = 0; i < nfds; i++)
+    { int fd = fds[i];
+
+      if ( fd >= 3 )
+	close(fd);
+    }
     stopItimer();
 
     execl(shell, BaseName(shell), "-c", cmd, (char *)0);
@@ -2267,7 +2377,7 @@ System(char *cmd)
 
   return rval;
 }
-#endif /* unix */
+#endif /* __unix__ */
 
 #ifdef tos
 #define SPECIFIC_SYSTEM 1
@@ -2424,12 +2534,12 @@ char *command;
 #ifndef __WIN32__			/* Win32 version in pl-nt.c */
 
 char *
-Symbols(char *buffer)
+findExecutable(const char *av0, char *buffer)
 { char *file;
   char buf[MAXPATHLEN];
   char tmp[MAXPATHLEN];
 
-  PrologPath(GD->cmdline.argv[0], buf);
+  PrologPath(av0, buf);
   file = Which(buf, tmp);
 
 #if __unix__				/* argv[0] can be an #! script! */
@@ -2437,10 +2547,10 @@ Symbols(char *buffer)
   { int n, fd;
     char buf[MAXPATHLEN];
 
+					/* Fails if mode is x-only, but */
+					/* then it can't be a script! */
     if ( (fd = open(file, O_RDONLY)) < 0 )
-    { warning("Cannot open %s: %s", file, OsError());
-      return file;
-    }
+      return strcpy(buffer, file);
 
     if ( (n=read(fd, buf, sizeof(buf)-1)) > 0 )
     { close(fd);
@@ -2454,9 +2564,7 @@ Symbols(char *buffer)
 	  ;
 	*q = EOS;
 
-	strcpy(buffer, s);
-
-	return buffer;
+	return strcpy(buffer, s);
       }
     }
 
@@ -2464,14 +2572,12 @@ Symbols(char *buffer)
   }
 #endif /*__unix__*/
 
-  strcpy(buffer, file);
-
-  return buffer;
+  return strcpy(buffer, file ? file : buf);
 }
 #endif /*__WIN32__*/
 
 
-#if unix
+#ifdef __unix__
 static char *
 okToExec(const char *s)
 { struct stat stbuff;
@@ -2484,7 +2590,7 @@ okToExec(const char *s)
     return (char *) NULL;
 }
 #define PATHSEP	':'
-#endif /* unix */
+#endif /* __unix__ */
 
 #ifdef tos
 #define EXEC_EXTENSIONS { ".ttp", ".prg", NULL }
@@ -2521,7 +2627,7 @@ okToExec(const char *s)
 }
 #endif /*EXEC_EXTENSIONS*/
 
-char *
+static char *
 Which(const char *program, char *fullname)
 { char *path, *dir;
   char *e;
@@ -2555,11 +2661,9 @@ Which(const char *program, char *fullname)
 
   while(*path)
   { if ( *path == PATHSEP )
-    { if ( (e = okToExec(program)) != NULL)
-      { strcpy(fullname, e);
-
-        return fullname;
-      } else
+    { if ( (e = okToExec(program)) )
+	return strcpy(fullname, e);
+      else
         path++;				/* fix by Ron Hess (hess@sco.com) */
     } else
     { char tmp[MAXPATHLEN];
@@ -2571,10 +2675,9 @@ Which(const char *program, char *fullname)
       if (strlen(fullname) + strlen(program)+2 > MAXPATHLEN)
         continue;
       *dir++ = '/';
-      *dir = EOS;
       strcpy(dir, program);
-      if ( (e = okToExec(OsPath(fullname, tmp))) != NULL )
-	return e;
+      if ( (e = okToExec(OsPath(fullname, tmp))) )
+	return strcpy(fullname, e);
     }
   }
 
@@ -2590,7 +2693,23 @@ Which(const char *program, char *fullname)
     point,  expressing  the  time  to sleep in seconds.
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
-#ifdef HAVE_SELECT
+#ifdef WIN32
+#define PAUSE_DONE 1			/* see pl-nt.c */
+#endif
+
+#if !defined(PAUSE_DONE) && defined(HAVE_USLEEP)
+#define PAUSE_DONE 1
+
+void
+Pause(real t)
+{ usleep((unsigned long)(t * 1000000.0));
+}
+
+#endif /*HAVE_USLEEP*/
+
+
+#if !defined(PAUSE_DONE) && defined(HAVE_SELECT)
+#define PAUSE_DONE 1
 
 void
 Pause(real time)
@@ -2607,8 +2726,10 @@ Pause(real time)
     sleep( (int)(time+0.5) );
 }
 
-#else /*!HAVE_SELECT*/
-#ifdef HAVE_DOSSLEEP
+#endif /*HAVE_SELECT*/
+
+#if !defined(PAUSE_DONE) && defined(HAVE_DOSSLEEP)
+#define PAUSE_DONE 1
 
 void                            /* a millisecond granualrity. */
 Pause(time)                     /* the EMX function sleep uses a seconds */
@@ -2620,8 +2741,10 @@ real time;                      /* granularity only. */
   DosSleep((ULONG)(time * 1000));
 }
 
-#else /*HAVE_DOSSLEEP*/
-#ifdef HAVE_SLEEP
+#endif /*HAVE_DOSSLEEP*/
+
+#if !defined(PAUSE_DONE) && defined(HAVE_SLEEP)
+#define PAUSE_DONE 1
 
 void
 Pause(real t)
@@ -2630,8 +2753,11 @@ Pause(real t)
 
   sleep((int)(t + 0.5));
 }
-#else /*HAVE_SLEEP*/
-#ifdef HAVE_DELAY
+
+#endif /*HAVE_SLEEP*/
+
+#if !defined(PAUSE_DONE) && defined(HAVE_DELAY)
+#define PAUSE_DONE 1
 
 void
 Pause(real t)
@@ -2639,11 +2765,10 @@ Pause(real t)
 }
 
 #endif /*HAVE_DELAY*/
-#endif /*HAVE_SLEEP*/
-#endif /*HAVE_DOSSLEEP*/
-#endif /*HAVE_SELECT*/
 
-#if tos
+#if !defined(PAUSE_DONE) && defined(tos)
+#define PAUSE_DONE 1
+
 void
 Pause(t)
 real t;
@@ -2661,4 +2786,12 @@ real t;
 
   wait_ticks += end_tick - start_tick;
 }
+
+#endif /*tos*/
+
+#ifndef PAUSE_DONE
+void
+Pause(real t)
+{ Sdprintf("Pause(): not implemented\n");
+}
 #endif
