@@ -1,10 +1,25 @@
-/*  $Id: pl-ssymb.h,v 1.5 1998/08/06 13:27:27 jan Exp $
+/*  $Id: pl-ssymb.h,v 1.12 2002/02/01 16:49:14 jan Exp $
 
-    Copyright (c) 1990 Jan Wielemaker. All rights reserved.
-    See ../LICENCE to find out about your rights.
-    jan@swi.psy.uva.nl
+    Part of SWI-Prolog
 
-    Purpose: Redefine functions fr compilers only handling short symbols
+    Author:        Jan Wielemaker
+    E-mail:        jan@swi.psy.uva.nl
+    WWW:           http://www.swi-prolog.org
+    Copyright (C): 1985-2002, University of Amsterdam
+
+    This library is free software; you can redistribute it and/or
+    modify it under the terms of the GNU Lesser General Public
+    License as published by the Free Software Foundation; either
+    version 2.1 of the License, or (at your option) any later version.
+
+    This library is distributed in the hope that it will be useful,
+    but WITHOUT ANY WARRANTY; without even the implied warranty of
+    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
+    Lesser General Public License for more details.
+
+    You should have received a copy of the GNU Lesser General Public
+    License along with this library; if not, write to the Free Software
+    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
 
 #define pl_dwim_match		Sdwimmatch
@@ -30,7 +45,6 @@
 #define pl_collect_parms	Scollparms
 #define pl_current_arithmetic_function Scfa
 #define pl_current_atom		Sca
-#define pl_current_stream	Scs
 #define pl_current_input	Scin
 #define pl_current_output	Scout
 #define pl_current_functor	Scfunctor
@@ -46,8 +60,6 @@
 #define pl_prolog_current_frame	Spcf
 #define pl_prolog_frame_attribute	Spfa
 #define pl_trace		Strace
-#define pl_flush		Sflush
-#define pl_flush_output		Sflushout
 #define pl_format		Sfmt
 #define pl_format3		Sfmt3
 #define pl_format_number	Sfmtnumber
@@ -56,9 +68,6 @@
 #define	pl_writeq		Swrtq
 #define	pl_write2		Swrt2
 #define	pl_writeq2		Swrtq2
-#define	pl_write_on_string	Swrtos
-#define	pl_write_on_atom	Swrtoa
-#define	pl_write_on_list	Swrtol
 #define pl_string		Sstr
 #define pl_string_length	Sstrl
 #define pl_string_to_atom	Sstrta
@@ -108,8 +117,6 @@
 #define pl_recorda		Sreca
 #define pl_recordz		Srecz
 #define pl_recorded		Sreced
-#define pl_read_variables	Srdvars
-#define pl_read_variables3	Srdvars3
 #define pl_read			Sread
 #define pl_read_clause		Srdcl
 #define pl_read_clause2		Srdcl2
@@ -127,4 +134,3 @@
 #define pl_tty_goto		Sttygoto
 
 #define PROCEDURE_garbage_collect0	Sproc_gc
-#define PROCEDURE_alt0			Sproc_alt
