@@ -1,21 +1,52 @@
-/*  $Id: pl-stream.c,v 1.25 1998/08/06 13:27:29 jan Exp $
+/*  $Id: pl-stream.c,v 1.58 2002/02/01 16:49:14 jan Exp $
 
-    Copyright (c) 1990 Jan Wielemaker. All rights reserved.
-    See ../LICENCE to find out about your rights.
-    jan@swi.psy.uva.nl
+    Part of SWI-Prolog
 
-    Purpose: SWI-Prolog IO streams
+    Author:        Jan Wielemaker
+    E-mail:        jan@swi.psy.uva.nl
+    WWW:           http://www.swi-prolog.org
+    Copyright (C): 1985-2002, University of Amsterdam
+
+    This library is free software; you can redistribute it and/or
+    modify it under the terms of the GNU Lesser General Public
+    License as published by the Free Software Foundation; either
+    version 2.1 of the License, or (at your option) any later version.
+
+    This library is distributed in the hope that it will be useful,
+    but WITHOUT ANY WARRANTY; without even the implied warranty of
+    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
+    Lesser General Public License for more details.
+
+    You should have received a copy of the GNU Lesser General Public
+    License along with this library; if not, write to the Free Software
+    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
 
+#if defined(WIN32) && !defined(__WIN32__)
+#define __WIN32__ 1
+#endif
+
 #ifdef __WIN32__
 #include <uxnt.h>
 #define MD "config/win32.h"
 #endif
 
+/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
+This modules defines the  SWI-Prolog  I/O   streams.  These  streams are
+provided to gain common access to  any   type  of character data: files,
+stdio streams, but also resources, strings, XPCE objects, etc.
+
+MT:
+
+Multithreading is supported through  Slock()   and  Sunlock(). These are
+recursive locks. If a stream handle  might   be  known to another thread
+locking is required. 
+- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
+
 #ifdef MD
 #include MD
 #else
-#include "config.h"
+#include <config.h>
 #endif
 
 #define PL_KERNEL 1
@@ -24,6 +55,10 @@
 #include <errno.h>
 #ifdef HAVE_MALLOC_H
 #include <malloc.h>
+#else
+#ifdef HAVE_SYS_MALLOC_H
+#include <sys/malloc.h>
+#endif
 #endif
 #include <memory.h>
 #include <string.h>
@@ -55,9 +90,36 @@
 int Slinesize = SIO_LINESIZE;		/* Sgets() buffer size */
 
 static int	S__flushbuf(IOSTREAM *s);
+static void	run_close_hooks(IOSTREAM *s);
 
 #define S__fupdatefilepos(s, c) S___fupdatefilepos(s, c)
 
+#ifdef O_PLMT
+#define SLOCK(s)    if ( s->mutex ) recursive_mutex_lock(s->mutex)
+#define SUNLOCK(s)  if ( s->mutex ) recursive_mutex_unlock(s->mutex)
+inline int
+STRYLOCK(IOSTREAM *s)
+{ if ( s->mutex &&
+       recursive_mutex_trylock(s->mutex) == EBUSY )
+    return FALSE;
+
+  return TRUE;
+}
+#else
+#define SLOCK(s)
+#define SUNLOCK(s)
+#define STRYLOCK(s) (TRUE)
+#endif
+
+#include "pl-error.h"
+extern int 			fatalError(const char *fm, ...);
+extern int 			PL_error(const char *pred, int arity,
+					 const char *msg, int id, ...);
+#ifdef O_PLMT
+extern recursive_mutex_t *	newRecursiveMutex(void);
+extern int			freeRecursiveMutex(recursive_mutex_t *m);
+#endif
+
 		 /*******************************
 		 *	      BUFFER		*
 		 *******************************/
@@ -87,7 +149,7 @@ S__setbuf(IOSTREAM *s, char *buffer, int size)
   s->limitp   = &s->buffer[s->bufsize];
   s->bufp     = s->buffer;
 
-  return s->bufsize;
+  return size;
 }
 
 
@@ -110,11 +172,25 @@ S__removebuf(IOSTREAM *s)
 
 int
 Slock(IOSTREAM *s)
-{ if ( s->locks )
-    s->locks++;
-  else if ( s->flags & SIO_NBUF )
-  { s->locks = 1;
-    return S__setbuf(s, NULL, TMPBUFSIZE);
+{ SLOCK(s);
+
+  if ( !s->locks++ )
+  { if ( (s->flags & (SIO_NBUF|SIO_OUTPUT)) == (SIO_NBUF|SIO_OUTPUT) )
+      return S__setbuf(s, NULL, TMPBUFSIZE);
+  }
+
+  return 0;
+}
+
+
+int
+StryLock(IOSTREAM *s)
+{ if ( !STRYLOCK(s) )
+    return -1;
+
+  if ( --s->locks == 0 )
+  { if ( (s->flags & (SIO_NBUF|SIO_OUTPUT)) == (SIO_NBUF|SIO_OUTPUT) )
+      return S__setbuf(s, NULL, TMPBUFSIZE);
   }
 
   return 0;
@@ -123,29 +199,43 @@ Slock(IOSTREAM *s)
 
 int
 Sunlock(IOSTREAM *s)
-{ if ( s->locks )
+{ int rval = 0;
+
+  if ( s->locks )
   { if ( --s->locks == 0 )
-      return S__removebuf(s);
+    { if ( (s->flags & (SIO_NBUF|SIO_OUTPUT)) == (SIO_NBUF|SIO_OUTPUT) )
+	rval = S__removebuf(s);
+    }
   }
 
-  return 0;
+  SUNLOCK(s);
+
+  return rval;
 }
 
+
 		 /*******************************
 		 *	     FLUSH/FILL		*
 		 *******************************/
 
 static int
 S__flushbuf(IOSTREAM *s)
-{ int size = s->bufp - s->buffer;
+{ int rval = 0;
+  int size;
 
-  if ( size > 0 )
-  { if ( (*s->functions->write)(s->handle, s->buffer, size) != size )
+  while ( (size = s->bufp - s->buffer) > 0 )
+  { int n = (*s->functions->write)(s->handle, s->buffer, size);
+
+    if ( n >= 0 )
+    { rval += n;
+      s->bufp -= n;
+    } else
+    { s->flags |= SIO_FERR;
       return -1;
-    s->bufp = s->buffer;
+    }
   }
 
-  return size;
+  return rval;
 }
 
 
@@ -153,19 +243,23 @@ static int
 S__flushbufc(int c, IOSTREAM *s)
 { if ( s->buffer )
   { if ( S__flushbuf(s) < 0 )
-      return -1;
-
-    *s->bufp++ = (c & 0xff);
+      c = -1;
+    else
+      *s->bufp++ = (c & 0xff);
   } else
   { if ( s->flags & SIO_NBUF )
     { char chr = (char)c;
     
       if ( (*s->functions->write)(s->handle, &chr, 1) != 1 )
-	return -1;
+      { s->flags |= SIO_FERR;
+	c = -1;
+      }
     } else
     { if ( S__setbuf(s, NULL, 0) < 0 )
-	return -1;
-      *s->bufp++ = (char)c;
+      { s->flags |= SIO_FERR;
+	c = -1;
+      } else
+	*s->bufp++ = (char)c;
     }
   }
 
@@ -177,9 +271,11 @@ S__flushbufc(int c, IOSTREAM *s)
 
 int
 S__fillbuf(IOSTREAM *s)
-{ if ( s->flags & (SIO_FEOF|SIO_FERR) )
+{ int c;
+
+  if ( s->flags & (SIO_FEOF|SIO_FERR) )
   { s->flags |= SIO_FEOF2;		/* reading past eof */
-    return -1;
+    goto error;
   }
 
   if ( s->flags & SIO_NBUF )
@@ -187,45 +283,52 @@ S__fillbuf(IOSTREAM *s)
     int n;
 
     if ( (n=(*ReadF(s))(s->handle, &chr, 1)) == 1 )
-    { return char_to_int(chr);
+    { c = char_to_int(chr);
+      goto ok;
     } else if ( n == 0 )
     { if ( !(s->flags & SIO_NOFEOF) )
 	s->flags |= SIO_FEOF;
-      return EOF;
+      goto error;
     } else
     { s->flags |= SIO_FERR;
-      return -1;			/* error */
+      goto error;			/* error */
     }
   } else
   { int n;
 
     if ( !s->buffer )
     { if ( S__setbuf(s, NULL, 0) < 0 )
-	return -1;
+	goto error;
       s->limitp = s->buffer;
     }
 
     if ( (n=(*ReadF(s))(s->handle, s->buffer, s->bufsize)) > 0 )
     { s->bufp = s->buffer;
       s->limitp = &s->buffer[n];
-      return char_to_int(*s->bufp++);
+      c = char_to_int(*s->bufp++);
+      goto ok;
     } else
-    {
-#if 0
-      s->bufp = s->buffer;		/* empty the buffer */
-      s->limitp = s->buffer;
-#endif
-
-      if ( n == 0 )
+    { if ( n == 0 )
       { if ( !(s->flags & SIO_NOFEOF) )
 	  s->flags |= SIO_FEOF;
-	return EOF;
+	goto error;
+#ifdef EWOULDBLOCK
+      } else if ( errno == EWOULDBLOCK )
+      { s->bufp = s->buffer;
+	s->limitp = s->buffer;
+	goto error;
+#endif
       } else
       { s->flags |= SIO_FERR;
-	return -1;
+	goto error;
       }
     }
   }
+
+error:
+  c = -1;
+ok:
+  return c;
 }
 
 		 /*******************************
@@ -234,16 +337,32 @@ S__fillbuf(IOSTREAM *s)
 
 
 inline int
-S___fupdatefilepos(IOPOS *p, int c)
-{ if ( p )
+S___fupdatefilepos(IOSTREAM *s, int c)
+{ IOPOS *p;
+
+#if 1
+  if ( s->magic != SIO_MAGIC )
+  { if ( s->magic == SIO_CMAGIC )
+      PL_error(NULL, 0, NULL, ERR_CLOSED_STREAM, s);
+					/* PL_error() should not return */
+    fatalError("Did you load a pre-3.1.2 foreign package?"); 
+  }
+#endif
+
+  if ( (p = s->position) )
   { switch(c)
     { case '\n':
 	p->lineno++;
         p->linepos = 0;
-/*      p->flags &= ~SIO_NOLINEPOS; TBD*/
+	s->flags &= ~SIO_NOLINEPOS;
+#ifdef __WIN32__
+	if ( s->flags & O_TEXT )
+	  p->charno++;			/* writes one extra! */
+#endif
         break;
       case '\r':
 	p->linepos = 0;
+        s->flags &= ~SIO_NOLINEPOS;
 	break;
       case '\b':
 	if ( p->linepos > 0 )
@@ -367,9 +486,9 @@ Sfread(void *data, int size, int elms, IOSTREAM *s)
 
 
 int
-Sfwrite(void *data, int size, int elms, IOSTREAM *s)
+Sfwrite(const void *data, int size, int elms, IOSTREAM *s)
 { int chars = size * elms;
-  char *buf = data;
+  const char *buf = data;
 
   for( ; chars > 0; chars-- )
   { if ( Sputc(*buf++, s) < 0 )
@@ -386,7 +505,22 @@ Sfwrite(void *data, int size, int elms, IOSTREAM *s)
 
 int
 Sfeof(IOSTREAM *s)
-{ return (s->flags & SIO_FEOF) != 0;
+{ if ( s->flags & SIO_FEOF )
+    return TRUE;
+
+  if ( s->bufp < s->limitp )
+    return FALSE;
+
+  if ( s->flags & SIO_NBUF )
+  { errno = EINVAL;
+    return -1;
+  }
+
+  if ( S__fillbuf(s) == -1 )
+    return TRUE;
+
+  s->bufp--;
+  return FALSE;
 }
     
 
@@ -415,7 +549,9 @@ Sclearerr(IOSTREAM *s)
 int
 Sflush(IOSTREAM *s)
 { if ( s->buffer && (s->flags & SIO_OUTPUT) )
-    return S__flushbuf(s);
+  { if ( S__flushbuf(s) < 0 )
+      return -1;
+  }
 
   return 0;
 }
@@ -545,6 +681,15 @@ int
 Sclose(IOSTREAM *s)
 { int rval = 0;
 
+  if ( s->magic != SIO_MAGIC )		/* already closed!? */
+  { errno = EINVAL;
+    return -1;
+  }
+
+  SLOCK(s);
+  while(s->locks > 0)			/* remove buffer-locks */
+    rval = Sunlock(s);
+
   if ( s->buffer )
   { if ( (s->flags & SIO_OUTPUT) && S__flushbuf(s) < 0 )
       rval = -1;
@@ -557,8 +702,18 @@ Sclose(IOSTREAM *s)
 
   if ( s->functions->close && (*s->functions->close)(s->handle) < 0 )
     rval = -1;
+  run_close_hooks(s);
 
-  s->magic = 0;
+  SUNLOCK(s);
+
+#ifdef O_PLMT
+  if ( s->mutex )
+  { freeRecursiveMutex(s->mutex);
+    s->mutex = NULL;
+  }
+#endif
+
+  s->magic = SIO_CMAGIC;
   if ( !(s->flags & SIO_STATIC) )
     free(s);
 
@@ -579,20 +734,20 @@ Sfgets(char *buf, int n, IOSTREAM *s)
 
     if ( c == EOF )
     { *q = '\0';
-      if ( q > buf )
-	return buf;
-      else
-	return NULL;
+      if ( q == buf )
+	buf = NULL;
+      goto out;
     } else
     { *q++ = c;
       if ( c == '\n' )
       { if ( n > 0 )
 	  *q = '\0';
-	return buf;
+	goto out;
       }
     }
   }
 
+out:
   return buf;
 }
 
@@ -614,7 +769,8 @@ Sgets(char *buf)
 
 int
 Sfputs(const char *q, IOSTREAM *s)
-{ for( ; *q; q++)
+{ 
+  for( ; *q; q++)
   { if ( Sputc(*q, s) < 0 )
       return EOF;
   }
@@ -666,7 +822,7 @@ Svprintf(const char *fm, va_list args)
 
 
 #define OUT(s, c)	do { printed++; \
-			     if ( Sputc((c), (s)) < 0 ) return -1; \
+			     if ( Sputc((c), (s)) < 0 ) goto error; \
 			   } while(0)
 #define valdigit(c)	((c) - '0')
 #define A_LEFT	0			/* left-aligned field */
@@ -678,6 +834,8 @@ Svfprintf(IOSTREAM *s, const char *fm, va_list args)
   char buf[TMPBUFSIZE];
   int tmpbuf;
 
+  SLOCK(s);
+
   if ( !s->buffer && (s->flags & SIO_NBUF) )
   { S__setbuf(s, buf, sizeof(buf));
     tmpbuf = TRUE;
@@ -715,12 +873,12 @@ Svfprintf(IOSTREAM *s, const char *fm, va_list args)
 	if ( *fm == '*' )
 	{ has_arg1++;
 	  arg1 = va_arg(args, int);
-	} else if ( isdigit(*fm) )
+	} else if ( isdigit(char_to_int(*fm)) )
 	{ if ( *fm == '0' )
 	    pad = '0';
 	  arg1 = valdigit(*fm);
 	  has_arg1++;
-	  for( fm++; isdigit(*fm); fm++)
+	  for( fm++; isdigit(char_to_int(*fm)); fm++)
 	    arg1 = arg1*10 + valdigit(*fm);
 	}
 	if ( *fm == '.' )
@@ -730,7 +888,7 @@ Svfprintf(IOSTREAM *s, const char *fm, va_list args)
 	  { arg2 = va_arg(args, int);
 	  } else
 	  { arg2 = 0;
-	    for( ; isdigit(*fm); fm++)
+	    for( ; isdigit(char_to_int(*fm)); fm++)
 	      arg2 = arg2*10 + valdigit(*fm);
 	  }
 	}
@@ -798,6 +956,8 @@ Svfprintf(IOSTREAM *s, const char *fm, va_list args)
 	  }
 	  case 's':
 	    fs = va_arg(args, char *);
+	    if ( !fs )
+	      fs = "(null)";
 	    break;
 	}
 
@@ -850,10 +1010,15 @@ Svfprintf(IOSTREAM *s, const char *fm, va_list args)
 
   if ( tmpbuf )
   { if ( S__removebuf(s) < 0 )
-      return -1;
+      goto error;
   }
 
+  SUNLOCK(s);
   return printed;
+
+error:
+  SUNLOCK(s);
+  return -1;
 }
 
 
@@ -882,6 +1047,7 @@ Svsprintf(char *buf, const char *fm, va_list args)
   s.position  = NULL;
   s.handle    = NULL;
   s.functions = NULL;
+  s.mutex     = NULL;
   
   if ( (rval = Svfprintf(&s, fm, args)) >= 0 )
     *s.bufp = '\0';
@@ -894,8 +1060,10 @@ int
 Svdprintf(const char *fm, va_list args)
 { int rval;
 
+  SLOCK(Soutput);
   rval = Svfprintf(Soutput, fm, args);
   Sflush(Soutput);
+  SUNLOCK(Soutput);
 
   return rval;
 }
@@ -1255,6 +1423,7 @@ IOFUNCTIONS Sfilefunctions =
 IOSTREAM *
 Snew(void *handle, int flags, IOFUNCTIONS *functions)
 { IOSTREAM *s;
+  int fd;
 
   if ( !(s = malloc(sizeof(IOSTREAM))) )
   { errno = ENOMEM;
@@ -1270,6 +1439,12 @@ Snew(void *handle, int flags, IOFUNCTIONS *functions)
   { s->position = &s->posbuf;
     s->posbuf.lineno = 1;
   }
+#ifdef O_PLMT
+  if ( !(s->mutex = newRecursiveMutex()) )
+    return NULL;
+#endif
+  if ( (fd = Sfileno(s)) >= 0 && isatty(fd) )
+    s->flags |= SIO_ISATTY;
 
   return s;
 }
@@ -1279,12 +1454,20 @@ Snew(void *handle, int flags, IOFUNCTIONS *functions)
 #define O_BINARY 0
 #endif
 
+/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
+Open a file. In addition to the normal  arguments, "lr" means get a read
+(shared-) lock on the file and  "lw"   means  get  an write (exclusive-)
+lock.  How much do we need to test here?
+- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
+
+
 IOSTREAM *
 Sopen_file(const char *path, const char *how)
 { int fd;
   int oflags = 0, flags = SIO_FILE|SIO_TEXT|SIO_RECORDPOS;
   int op = *how++;
   long lfd;
+  enum {lnone=0,lread,lwrite} lock = lnone;
 
   for( ; *how; how++)
   { switch(*how)
@@ -1295,6 +1478,16 @@ Sopen_file(const char *path, const char *how)
       case 'r':				/* no record */
 	flags &= SIO_RECORDPOS;
         break;
+      case 'l':				/* lock r: read, w: write */
+	if ( *++how == 'r' )
+	  lock = lread;
+        else if ( *how == 'w' )
+	  lock = lwrite;
+        else
+	{ errno = EINVAL;
+	  return NULL;
+	}
+        break;
       default:
 	errno = EINVAL;
         return NULL;
@@ -1308,11 +1501,11 @@ Sopen_file(const char *path, const char *how)
       break;
     case 'a':
       fd = open(path, O_WRONLY|O_CREAT|O_APPEND|oflags, 0666);
-      flags |= SIO_OUTPUT;
+      flags |= SIO_OUTPUT|SIO_APPEND;
       break;
     case 'u':
       fd = open(path, O_WRONLY|O_CREAT|oflags, 0666);
-      flags |= SIO_OUTPUT;
+      flags |= SIO_OUTPUT|SIO_UPDATE;
       break;
     case 'r':
       fd = open(path, O_RDONLY|oflags);
@@ -1326,6 +1519,27 @@ Sopen_file(const char *path, const char *how)
   if ( fd < 0 )
     return NULL;
 
+  if ( lock )
+  { 
+#ifdef FCNTL_LOCKS
+    struct flock buf;
+
+    memset(&buf, 0, sizeof(buf));
+    buf.l_type = (lock == lread ? F_RDLCK : F_WRLCK);
+
+    if ( fcntl(fd, F_SETLKW, &buf) < 0 )
+    { int save = errno;
+      close(fd);
+      errno = save;
+      return NULL;
+    }
+#else					/* we don't have locking */
+    close(fd);
+    errno = EINVAL;
+    return NULL;
+#endif
+  }
+
   lfd = (long)fd;
   return Snew((void *)lfd, flags, &Sfilefunctions);
 }
@@ -1349,35 +1563,73 @@ Sfdopen(int fd, const char *type)
   return Snew((void *)lfd, flags, &Sfilefunctions);
 }
 
+/* MT: as long as s is valid, this should be ok
+*/
 
 int
 Sfileno(IOSTREAM *s)
-{ if ( s->flags & SIO_FILE )
-  { long h = (long)s->handle;
-    return (int)h;
-  }
-  if ( s->flags & SIO_PIPE )
-    return fileno((FILE *)s->handle);
+{ int n;
 
-  errno = EINVAL;
-  return -1;				/* no file stream */
+  if ( s->flags & SIO_FILE )
+  { long h = (long)s->handle;
+    n = (int)h;
+  } else if ( s->flags & SIO_PIPE )
+  { n = fileno((FILE *)s->handle);
+  } else if ( s->functions->control &&
+	      (*s->functions->control)(s->handle,
+				       SIO_GETFILENO,
+				       (void *)&n)  == 0 )
+  { ;
+  } else
+  { errno = EINVAL;
+    n = -1;				/* no file stream */
+  }
+
+  return n;
 }
 
 
 #define STDIO(n, f) { NULL, NULL, NULL, NULL, \
-		      SIO_MAGIC, EOF, 0, f, {0, 0, 0}, NULL, \
-		      ((void *)(n)), &Sfilefunctions \
+		      EOF, SIO_MAGIC, 0, f, {0, 0, 0}, NULL, \
+		      ((void *)(n)), &Sfilefunctions, \
+		      0, NULL \
 		    }
 
-IOSTREAM S__iob[] =
-{ 
-#define SIO_STDIO (SIO_FILE|SIO_STATIC|SIO_NOCLOSE)
-  STDIO(0, SIO_STDIO|SIO_LBUF|SIO_INPUT|SIO_NOFEOF),	/* Sinput */
-  STDIO(1, SIO_STDIO|SIO_LBUF|SIO_OUTPUT), 		/* Soutput */
+#define SIO_STDIO (SIO_FILE|SIO_STATIC|SIO_NOCLOSE|SIO_ISATTY)
+#define STDIO_STREAMS \
+  STDIO(0, SIO_STDIO|SIO_LBUF|SIO_INPUT|SIO_NOFEOF),	/* Sinput */ \
+  STDIO(1, SIO_STDIO|SIO_LBUF|SIO_OUTPUT), 		/* Soutput */ \
   STDIO(2, SIO_STDIO|SIO_NBUF|SIO_OUTPUT)		/* Serror */
+
+
+IOSTREAM S__iob[] =
+{ STDIO_STREAMS
 };
 
 
+static const IOSTREAM S__iob0[] =
+{ STDIO_STREAMS
+};
+
+
+void
+SinitStreams()
+{ static int done;
+
+  if ( !done++ )
+  {
+#ifdef O_PLMT
+    S__iob[0].mutex = newRecursiveMutex();
+    S__iob[1].mutex = newRecursiveMutex();
+    S__iob[2].mutex = newRecursiveMutex();
+#endif
+    if ( !isatty(0) ) S__iob[0].flags &= ~SIO_ISATTY;
+    if ( !isatty(2) ) S__iob[1].flags &= ~SIO_ISATTY;
+    if ( !isatty(3) ) S__iob[2].flags &= ~SIO_ISATTY;
+  }
+}
+
+
 IOSTREAM *
 S__getiob()
 { return S__iob;
@@ -1450,6 +1702,8 @@ Sopen_pipe(const char *command, const char *type)
 Memory streams form a replacement for   sprintf(), sscanf() and friends.
 They allow regarding a piece of  (for output) malloc() maintained memory
 to serve as a temporary buffer.
+
+MT: we assume these handles are not passed between threads
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
 typedef struct
@@ -1580,16 +1834,33 @@ IOFUNCTIONS Smemfunctions =
 
 
 /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
-Sopenmem(char **buffer, int *size, char* mode)
+Sopenmem(char **buffer, int *size, const char* mode)
     Open an memory area as a stream.  Output streams will automatically
     resized using realloc() if *size = 0 or the stream is opened with mode
     "wa".
+
+    If the buffer is allocated or enlarged, this is achieved using malloc()
+    or realloc().  In this case the returned buffer should be freed by the
+    caller when done.  Example:
+
+    { char buf[1024];			(don't allocate for small stuff)
+      char *s = buf;
+      IOSTREAM *fd;
+      int size = sizeof(buf);
+
+      fd = Sopenmem(&s, &size, "w");
+      ...
+      Sclose(fd);
+      ...
+      if ( s != buf )			(appearently moved)
+	free(s);
+    }
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
 IOSTREAM *
 Sopenmem(char **buffer, int *sizep, const char *mode)
 { memfile *mf = malloc(sizeof(memfile));
-  int flags = SIO_FBUF;
+  int flags = SIO_FBUF|SIO_RECORDPOS;
   int size;
 
   if ( !mf )
@@ -1617,6 +1888,8 @@ Sopenmem(char **buffer, int *sizep, const char *mode)
 	mf->malloced = TRUE;
       if ( *buffer )
 	*buffer[0] = '\0';
+      if ( sizep )
+	*sizep = mf->size;
       break;
     default:
       free(mf);
@@ -1635,6 +1908,9 @@ Sopenmem(char **buffer, int *sizep, const char *mode)
 		 *	      STRINGS		*
 		 *******************************/
 
+/* MT: we assume these handles are not passed between threads
+*/
+
 static int
 Sread_string(void *handle, char *buf, int size)
 { return 0;				/* signal EOF */
@@ -1682,10 +1958,10 @@ Sopen_string(IOSTREAM *s, char *buf, int size, const char *mode)
   } else
     flags |= SIO_STATIC;
 
+  memset((char *)s, 0, sizeof(IOSTREAM));
   s->buffer    = buf;
   s->bufp      = buf;
   s->unbuffer  = buf;
-  s->position  = NULL;
   s->handle    = s;			/* for Sclose_string() */
   s->functions = &Sstringfunctions;
 
@@ -1705,6 +1981,7 @@ Sopen_string(IOSTREAM *s, char *buf, int size, const char *mode)
 
   s->flags  = flags;
   s->limitp = &buf[size];
+  s->magic  = SIO_MAGIC;
 
   return s;
 }
@@ -1712,6 +1989,70 @@ Sopen_string(IOSTREAM *s, char *buf, int size, const char *mode)
 #undef S__fupdatefilepos
 
 int
-S__fupdatefilepos(IOPOS *p, int c)
-{ return S___fupdatefilepos(p, c);
+S__fupdatefilepos(IOSTREAM *s, int c)
+{ return S___fupdatefilepos(s, c);
+}
+
+		 /*******************************
+		 *	       HOOKS		*
+		 *******************************/
+
+/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
+This allows external packages (Prolog itself) to monitor the destruction
+of streams.
+- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
+
+typedef struct _close_hook
+{ struct _close_hook *next;
+  void (*hook)(IOSTREAM *s);
+} close_hook;
+
+static close_hook *close_hooks;
+
+static void
+run_close_hooks(IOSTREAM *s)
+{ close_hook *p;
+
+  for(p=close_hooks; p; p = p->next)
+    (*p->hook)(s);
+
+  if ( s->close_hook )
+    (*s->close_hook)(s->closure);
+}
+
+int
+Sclosehook(void (*hook)(IOSTREAM *s))
+{ close_hook *h = malloc(sizeof(*h));
+
+  if ( !h )
+    return -1;
+  h->next = close_hooks;
+  h->hook = hook;
+  close_hooks = h;
+
+  return 0;
+}
+
+
+		 /*******************************
+		 *	       CLEANUP		*
+		 *******************************/
+
+void
+Scleanup(void)
+{ close_hook *p, *next;
+  int i;
+
+  for(p=close_hooks; p; p=next)
+  { next = p->next;
+    free(p);
+  }
+
+  close_hooks = NULL;
+
+  for(i=0; i<=2; i++)
+  { if ( S__iob[i].buffer )
+      free(S__iob[i].buffer);
+    S__iob[i] = S__iob0[i];
+  }
 }
