@@ -1,10 +1,25 @@
-/*  $Id: plld.c,v 1.6 1997/11/12 14:21:24 jan Exp $
+/*  $Id: plld.c,v 1.38 2002/02/01 16:49:14 jan Exp $
 
     Part of SWI-Prolog
-    Designed and implemented by Jan Wielemaker
-    E-mail: jan@swi.psy.uva.nl
 
-    Copyright (C) 1997 University of Amsterdam. All rights reserved.
+    Author:        Jan Wielemaker
+    E-mail:        jan@swi.psy.uva.nl
+    WWW:           http://www.swi-prolog.org
+    Copyright (C): 1985-2002, University of Amsterdam
+
+    This library is free software; you can redistribute it and/or
+    modify it under the terms of the GNU Lesser General Public
+    License as published by the Free Software Foundation; either
+    version 2.1 of the License, or (at your option) any later version.
+
+    This library is distributed in the hope that it will be useful,
+    but WITHOUT ANY WARRANTY; without even the implied warranty of
+    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
+    Lesser General Public License for more details.
+
+    You should have received a copy of the GNU Lesser General Public
+    License along with this library; if not, write to the Free Software
+    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
 
 /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
@@ -31,27 +46,43 @@ embedded application.
 
 #define PROG_PL "plcon.exe"
 #define PROG_LD "link.exe"
+#define SO_LD "link.exe"
+#define PROG_CC "cl.exe"
+#define PROG_CXX "cl.exe"
 #define PROG_OUT "plout.exe"
+#define PROG_CPP "cl.exe -P"
 #define LIB_PL	 "libpl.lib"
+#define LIB_PLMT "libplmt.lib"
+#define LIB_PL_DEBUG "libplD.lib"
 #define EXT_OBJ "obj"
-#define OPT_DEBUG "/debug"
+#define OPT_DEBUG "/DEBUG"
 #else /*WIN32*/
 #include "pl-incl.h"
 
 #define PROG_PL "pl"
-#define PROG_LD cc
+#define PROG_CC "cc"
+#define PROG_CXX "c++"
 #define PROG_OUT "a.out"
-#define LIB_PL	"-lpl"
 #define EXT_OBJ "o"
+#define LIB_PL	"-lpl"
+#define LIB_PLMT "-lplmt"
 #define OPT_DEBUG "-g"
-#ifndef O_BINARY
-#define O_BINARY 0
+
+#ifndef SO_LDFLAGS
+#define SO_LDFLAGS "-shared"
 #endif
+
 #endif /*WIN32*/
 
 #include <stdio.h>
 #include <stdlib.h>
+#ifdef HAVE_MALLOC_H
 #include <malloc.h>
+#else
+#ifdef HAVE_SYS_MALLOC_H
+#include <sys/malloc.h>
+#endif
+#endif
 #include <ctype.h>
 #ifdef HAVE_UNISTD_H
 #include <unistd.h>
@@ -72,6 +103,10 @@ embedded application.
 #define FALSE 0
 #endif
 
+#ifndef O_BINARY
+#define O_BINARY 0
+#endif
+
 #ifndef MAXPATHLEN
 #define MAXPATHLEN 1024
 #endif
@@ -91,9 +126,15 @@ embedded application.
 #define oserror xoserror
 #undef strdup
 #define strdup plld_strdup
+#ifdef HAVE_DMALLOC_H
+#undef xmalloc
+#undef xrealloc
+#undef xfree
+#endif
 #define xmalloc plld_xmalloc
 #define xrealloc plld_xrealloc
 #define xfree plld_xfree
+#define CTOI(c) ((c)&0xff)
 
 typedef struct
 { char **list;
@@ -118,6 +159,8 @@ static arglist lastlibs;		/* libs that must be at the end */
 static arglist libdirs;			/* -L library directories */
 static arglist includedirs;		/* -I include directories */
 
+static char *pllib = LIB_PL;    /* libpl option -lpl, libpl.lib or libplD.lib */
+
 static char *pl;			/* Prolog executable */
 static char *cc;			/* CC executable */
 static char *cxx;			/* C++ executable */
@@ -130,6 +173,8 @@ static char *plarch;			/* Prolog architecture id */
 static char *plgoal;			/* -g goal */
 static char *pltoplevel;		/* -t goal */
 static char *plinitfile;		/* -f file */
+static char *plclass;			/* -class <class> */
+static char *plsysinit;			/* -F file */
 
 static char *ctmp;			/* base executable */
 static char *pltmp;			/* base saved state */
@@ -137,11 +182,15 @@ static char *out;			/* final output */
 
 static int nostate = FALSE;		/* do not make a state */
 static int nolink = FALSE;		/* do not link */
+static int shared = FALSE;		/* -shared: make a shared-object/DLL */
+static char *soext;			/* extension of shared object */
 
 static int verbose = TRUE;		/* verbose operation */
 static int fake = FALSE;		/* don't really do anything */
 
 static void	removeTempFiles();
+static void	parseOptions(int argc, char **argv);
+
 
 		 /*******************************
 		 *	       ERROR		*
@@ -241,6 +290,16 @@ strdup(const char *in)
 }
 
 
+static char *
+strndup(const char *in, int len)
+{ char *r = xmalloc(len+1);
+ 
+  r[len] = '\0'; 
+
+  return memcpy(r, in, len);
+}
+
+
 void
 appendArgList(arglist *list, const char *arg)
 { if ( list->size == 0 )
@@ -277,51 +336,57 @@ concatArgList(arglist *to, const char *prefix, arglist *from)
   for(n=0; n<from->size; n++)
   { char buf[1024];
 
-    sprintf(buf, "%s%s", prefix, from->list[n]);
+    if ( strchr(from->list[n], ' ') )
+      sprintf(buf, "%s\"%s\"", prefix, from->list[n]);
+    else
+      sprintf(buf, "%s%s", prefix, from->list[n]);
     appendArgList(to, buf);
   }
 }
 
 
-void
-addArgs(const char *s, arglist *list)
-{ const char *f;
-  char tmp[1024];
+static int
+breakargs(const char *line, char **argv)
+{ int argc = 0;
 
-  while(*s)
-  { while(*s && isspace(*s))
-      s++;
-    f = s;
-    while(*s && !isspace(*s))
-      s++;
-    if ( s > f )
-    { strncpy(tmp, f, s-f);
-      tmp[s-f] = '\0';
-      appendArgList(list, tmp);
+  while(*line)
+  { while(*line && isspace(CTOI(*line)))
+      line++;
+
+    if ( *line == '"' )			/* Windows-95 quoted arguments */
+    { const char *start = line+1;
+      const char *end = start;
+
+      while( *end && *end != '"' )
+	end++;
+      if ( *end == '"' )
+      { argv[argc++] = strndup(start, end-start);
+	line = end+1;
+	continue;
+      }
+    }
+
+    if ( *line )
+    { const char *start = line;
+
+      while(*line && !isspace(CTOI(*line)))
+	line++;
+      argv[argc++] = strndup(start, line-start);
     }
   }
-}
+  argv[argc] = NULL;			/* add trailing NULL pointer to argv */
+
+  return argc;
+}      
+
 
 
 void
-addLibs(const char *s, arglist *list)
-{ const char *f;
-  char tmp[1024];
+addOptionString(const char *s)
+{ char *argv[256];
+  int argc = breakargs(s, argv);
 
-  while(*s)
-  { while(*s && isspace(*s))
-      s++;
-    f = s;
-    while(*s && !isspace(*s))
-      s++;
-    if ( s > f )
-    { if ( s > f+2 && strprefix(f, "-l") )
-	f += 2;
-      strncpy(tmp, f, s-f);
-      tmp[s-f] = '\0';
-      appendArgList(list, tmp);
-    }
-  }
+  parseOptions(argc, argv);
 }
 
 
@@ -437,7 +502,9 @@ static extdef extdefs[] =
   { "cpp",	&cppfiles },
   { "cxx",	&cppfiles },
   { "cc",	&cppfiles },
+#ifndef WIN32
   { "C",	&cppfiles },
+#endif
   { "pl",	&plfiles },
   { "qlf",	&qlfiles },
   { NULL,	NULL }
@@ -450,6 +517,8 @@ file_name_extension(const char *in)
   for( ; *in; in++)
   { if ( *in == '.' )
       ext = in+1;
+    else if ( *in == '/' || *in == '\\' )
+      ext = NULL;
   }
 
   return ext;
@@ -484,6 +553,7 @@ usage()
 { fprintf(stderr,
 	  "usage: %s -help\n"
 	  "       %s [options] inputfile ...\n"
+	  "       %s -E cppargument ...\n"
 	  "\n"
 	  "options:\n"
 	  "       -o out           define output file\n"
@@ -491,29 +561,37 @@ usage()
 	  "       -v               verbose\n"
 	  "       -f               fake (do not run any commands)\n"
 	  "       -g               Compile/link for debugging\n"
-	  "       -c               Only compile C/C++ files, do not link\n"
 	  "\n"
 	  "       -pl prolog       Prolog to use\n"
 	  "       -ld linker       link editor to use\n"
           "       -cc compiler     compiler for C source files\n"
 	  "       -c++ compiler    compiler for C++ source files\n"
 	  "\n"
+	  "       -c               only compile C/C++ files, do not link\n"
 	  "       -nostate         just relink the kernel\n"
+	  "       -shared          link to shared object or DLL\n"
+	  "       -fpic            compile small position-independent code\n"
+	  "       -fPIC            compile large position-independent code\n"
 	  "\n"
 	  "       -pl-options,...  Add options for Prolog\n"
 	  "       -ld-options,...  Add options for linker\n"
 	  "       -cc-options,...  Add options for C/C++-compiler\n"
+	  "       -F base          Load	swi(base.rc)"
 	  "\n"
 	  "       -goal goal       (Prolog) entry point\n"
 	  "       -toplevel goal   (Prolog) abort toplevel goal\n"
 	  "       -initfile file   (Prolog) profile file to load\n"
+	  "       -class class     {runtime,kernel,development}\n"
+	  "\n"
+	  "       -O*              Optimization passed to compiler\n"
+	  "       -W*              Warning options passed to compiler\n"
 	  "\n"
 	  "       -Dmacro          Define macro (C/C++)\n"
 	  "       -Umacro          Undefine macro (C/C++)\n"
 	  "       -Iincludedir     Include directory (C/C++)\n"
 	  "       -Llibdir         Library directory (C/C++ link)\n"
 	  "       -llib            library (C/C++)\n",
-	plld, plld);
+	plld, plld, plld);
 
   exit(1);
 }
@@ -538,9 +616,50 @@ parseOptions(int argc, char **argv)
     } else if ( streq(opt, "-g") )		/* -g */
     { appendArgList(&coptions, OPT_DEBUG);
       appendArgList(&cppoptions, OPT_DEBUG);
+#ifdef WIN32					/* MSVC DEBUG OPTIONS */
+      appendArgList(&coptions, "/ZI");
+      appendArgList(&coptions, "/Od");
+      appendArgList(&cppoptions, "/ZI");
+      appendArgList(&cppoptions, "/Od");
+#endif
       appendArgList(&ldoptions, OPT_DEBUG);
+#ifdef WIN32
+      pllib = LIB_PL_DEBUG;
+#endif
+    } else if ( strprefix(opt, "-O") )		/* -O* */
+    { appendArgList(&coptions, opt);
+      appendArgList(&cppoptions, opt);
+    } else if ( strprefix(opt, "-W") )		/* -W* */
+    { appendArgList(&coptions, opt);
+      appendArgList(&cppoptions, opt);
     } else if ( streq(opt, "-nostate") ) 	/* -nostate */
     { nostate = TRUE;
+    } else if ( streq(opt, "-shared") )		/* -shared */
+    { shared = TRUE;
+      nostate = TRUE;
+#ifdef SO_pic
+      appendArgList(&coptions, SO_pic);
+#endif
+    } else if ( streq(opt, "-SHARED") )		/* -SHARED */
+    { shared = TRUE;
+      nostate = TRUE;
+#ifdef SO_PIC
+      appendArgList(&coptions, SO_PIC);
+#else
+#ifdef SO_pic
+      appendArgList(&coptions, SO_pic);
+#endif
+#endif
+    } else if ( streq(opt, "-fpic") )		/* -fpic */
+    {
+#ifdef SO_pic
+      appendArgList(&coptions, SO_pic);
+#endif
+    } else if ( streq(opt, "-fPIC") )		/* -fPIC */
+    {
+#ifdef SO_PIC
+      appendArgList(&coptions, SO_PIC);
+#endif
     } else if ( streq(opt, "-o") ) 		/* -o out */
     { if ( argc > 1 )
       { out = argv[1];
@@ -565,6 +684,24 @@ parseOptions(int argc, char **argv)
 	argc--, argv++;
       } else
 	usage();
+    } else if ( streq(opt, "-F") ) 		/* -F base */
+    { if ( argc > 1 )
+      { plsysinit = argv[1];
+	argc--, argv++;
+      } else
+	usage();
+    } else if ( streq(opt, "-class") ) 		/* -class runtime,kernel,
+    							  development */
+    { if ( argc > 1 )
+      { plclass = argv[1];
+	if ( !streq(plclass, "runtime") &&
+	     !streq(plclass, "kernel") &&
+	     !streq(plclass, "development")
+	   )
+	  usage();
+	argc--, argv++;
+      } else
+	usage();
     } else if ( streq(opt, "-pl") ) 		/* -pl prolog */
     { if ( argc > 1 )
       { pl = argv[1];
@@ -599,11 +736,11 @@ parseOptions(int argc, char **argv)
     } else if ( strprefix(opt, "-I") )		/* -I<include> */
     { appendArgList(&includedirs, &opt[2]);
     } else if ( strprefix(opt, "-D") )		/* -D<def> */
-    { appendArgList(&coptions, &opt[2]);
-      appendArgList(&cppoptions, &opt[2]);
+    { appendArgList(&coptions, opt);
+      appendArgList(&cppoptions, opt);
     } else if ( strprefix(opt, "-U") )		/* -U<def> */
-    { appendArgList(&coptions, &opt[2]);
-      appendArgList(&cppoptions, &opt[2]);
+    { appendArgList(&coptions, opt);
+      appendArgList(&cppoptions, opt);
     } else if ( strprefix(opt, "-L") )		/* -L<libdir> */
     { appendArgList(&libdirs, &opt[2]);
     } else if ( streq(opt, "-lccmalloc") )	/* -lccmalloc */
@@ -652,16 +789,32 @@ tmpPath(char **store, const char *base)
 static void
 fillDefaultOptions()
 { char tmp[1024];
-  char *defcxx = "c++";
+  char *defcxx = PROG_CXX;
 
-  defaultProgram(&cc,  "cc");
+  defaultProgram(&cc,  PROG_CC);
   if ( streq(cc, "gcc") )
     defcxx = "g++";
   defaultProgram(&cxx, defcxx);
-  defaultProgram(&ld,  PROG_LD);
+
+  if ( !ld )				/* not specified */
+  { if ( shared )
+    {
+#ifndef SO_LD
+       ld = (cppfiles.size > 0 ? cxx : cc);
+#else
+       ld = SO_LD;
+#endif
+    } else
+#ifndef PROG_LD
+       ld = (cppfiles.size > 0 ? cxx : cc);
+#else
+       ld = PROG_LD;
+#endif
+  }
 
 #ifdef WIN32
-  ensureOption(&coptions, "/MD");
+  if (strcmp(LIB_PL_DEBUG,pllib) == 0) ensureOption(&coptions, "/MDd");
+  else ensureOption(&coptions, "/MD");
   ensureOption(&coptions, "/DWIN32");
   ensureOption(&coptions, "/nologo");
   ensureOption(&ldoptions, "/nologo");
@@ -669,14 +822,31 @@ fillDefaultOptions()
 
   tmpPath(&ctmp,   "ctmp-");
   tmpPath(&pltmp,  "pltmp-");
+#if defined(__CYGWIN32__)
+/* Compile generates .exe files on CygWin */
+  replaceExtension(ctmp, "exe", tmp);
+  free(ctmp);
+  ctmp = strdup(tmp);
+#endif
+#if defined(WIN32) || defined(__CYGWIN32__)
+/* Saved states have the .exe extension under Windows */
+  replaceExtension(pltmp, "exe", tmp);
+  free(pltmp);
+  pltmp = strdup(tmp);
+#endif
+  if ( shared && !out && !nolink )
+  { fprintf(stderr, "%s: \"-o out\" required for linking shared object\n", plld);
+    exit(1);
+  }
   defaultPath(&out, PROG_OUT);
 
   defaultProgram(&plgoal,     "$welcome");
   defaultProgram(&pltoplevel, "prolog");
   defaultProgram(&plinitfile, "none");
+  defaultProgram(&plsysinit,  "none");
 
 #ifdef WIN32
-  sprintf(tmp, "%s/bin", plbase);
+  sprintf(tmp, "%s/lib", plbase);
 #else
   sprintf(tmp, "%s/runtime/%s", plbase, plarch);
 #endif
@@ -712,7 +882,7 @@ getPrologOptions()
 
 	  while(*e && *e != '"')
 	    e++;
-	  while(e>v && isspace(e[-1]))
+	  while(e>v && isspace(CTOI(e[-1])))
 	    e--;
 	  *e = '\0';
 	}
@@ -722,11 +892,17 @@ getPrologOptions()
 	  defaultPath(&plbase, v);
 	else if ( streq(name, "PLARCH") )
 	  defaultPath(&plarch, v);
-	else if ( streq(name, "PLLIBS") )
-	  addLibs(v, &libs);
-	else if ( streq(name, "PLLDFLAGS" ) )
+	else if ( streq(name, "PLLIBS") && !shared )
+	  addOptionString(v);
+	else if ( streq(name, "PLLDFLAGS") && !shared )
 	  appendArgList(&ldoptions, v);
-	else
+	else if ( streq(name, "PLSOEXT") )
+	  soext = strdup(v);
+	else if ( streq(name, "PLTHREADS") && streq(v, "yes") )
+	{ ensureOption(&coptions, "-D_REENTRANT");
+	  ensureOption(&cppoptions, "-D_REENTRANT");
+	  pllib = LIB_PLMT;
+	} else
 	  continue;
 
 	if ( verbose )
@@ -819,7 +995,8 @@ compileFile(const char *compiler, arglist *options, const char *cfile)
 
   prependArgList(args, "-c");
   appendArgList(args, "-D__SWI_PROLOG__");
-  appendArgList(args, "-D__SWI_EMBEDDED__");
+  if ( !shared )
+    appendArgList(args, "-D__SWI_EMBEDDED__");
   concatArgList(args, "-I", &includedirs);
 
   appendArgList(args, "-o");
@@ -890,25 +1067,18 @@ linkBaseExecutable()
   sprintf(tmp, "/out:%s", cout);
   prependArgList(&ldoptions, tmp);
 }
-#else
+  concatArgList(&ldoptions, "", &ofiles);	/* object files */
+  exportlibdirs();
+  appendArgList(&ldoptions, pllib);		/* -lpl */
+  concatArgList(&ldoptions, "", &libs);		/* libraries */
+  concatArgList(&ldoptions, "", &lastlibs);	/* libraries */
+#else /*WIN32*/
   prependArgList(&ldoptions, cout);
   prependArgList(&ldoptions, "-o");		/* -o ctmp */
-#endif
   concatArgList(&ldoptions, "", &ofiles);	/* object files */
-#ifdef WIN32
-  exportlibdirs();
-#else
   concatArgList(&ldoptions, "-L", &libdirs);    /* library directories */
-#endif
-  appendArgList(&ldoptions, LIB_PL);		/* -lpl */
-#ifdef WIN32
-  concatArgList(&ldoptions, "", &libs);		/* libraries */
-#else
+  appendArgList(&ldoptions, pllib);		/* -lpl */
   concatArgList(&ldoptions, "-l", &libs);	/* libraries */
-#endif
-#ifdef WIN32
-  concatArgList(&ldoptions, "", &lastlibs);	/* libraries */
-#else
   concatArgList(&ldoptions, "-l", &lastlibs);	/* libraries */
 #endif
 
@@ -927,14 +1097,63 @@ linkBaseExecutable()
 }
 
 
+void
+linkSharedObject()
+{ char soname[MAXPATHLEN];
+  char *soout;
+
+  if ( !soext )
+    soext = "so";			/* or give error? */
+
+  if ( file_name_extension(out) )
+  { soout = out;
+  } else
+  { soout = replaceExtension(out, soext, soname);
+  }
+
+#ifdef WIN32
+  prependArgList(&ldoptions, "/dll");
+{ char tmp[MAXPATHLEN];
+  sprintf(tmp, "/out:%s", soout);
+  prependArgList(&ldoptions, tmp);
+}
+  concatArgList(&ldoptions, "", &ofiles);	/* object files */
+  exportlibdirs();
+  appendArgList(&ldoptions, pllib);		/* libpl.lib */
+  concatArgList(&ldoptions, "", &libs);		/* libraries */
+  concatArgList(&ldoptions, "", &lastlibs);	/* libraries */
+#else /*WIN32*/
+#ifdef __CYGWIN32__
+  prependArgList(&ldoptions, soout);
+  prependArgList(&ldoptions, "-o");		/* -o ctmp */
+  concatArgList(&ldoptions, "", &ofiles);	/* object files */
+  appendArgList(&ldoptions, "-lplimp");		/* kernel import library */
+  concatArgList(&ldoptions, "-L", &libdirs);    /* library directories */
+  concatArgList(&ldoptions, "-l", &libs);	/* libraries */
+  concatArgList(&ldoptions, "-l", &lastlibs);	/* libraries */
+#else /*__CYGWIN32__*/
+  prependArgList(&ldoptions, SO_LDFLAGS);
+  prependArgList(&ldoptions, soout);
+  prependArgList(&ldoptions, "-o");		/* -o ctmp */
+  concatArgList(&ldoptions, "", &ofiles);	/* object files */
+  concatArgList(&ldoptions, "-L", &libdirs);    /* library directories */
+  concatArgList(&ldoptions, "-l", &libs);	/* libraries */
+  concatArgList(&ldoptions, "-l", &lastlibs);	/* libraries */
+#endif /*__CYGWIN32__*/
+#endif /*WIN32*/
+
+  callprog(ld, &ldoptions);
+}
+
+
 static void
 quoted_name(const char *name, char *plname)
 { int needquote = TRUE;
 
-  if ( islower(name[0]) )
+  if ( islower(CTOI(name[0])) )
   { const char *s = name+1;
 
-    for( ; *s && (isalnum(*s) || *s == '_'); s++)
+    for( ; *s && (isalnum(CTOI(*s)) || *s == '_'); s++)
       ;
     if ( *s == '\0' )
       needquote = FALSE;
@@ -994,19 +1213,22 @@ createSavedState()
   e = put_pl_option(e, "toplevel", pltoplevel);
   *e++ = ',';
   e = put_pl_option(e, "initfile", plinitfile);
+  if ( plclass )
+  { *e++ = ',';
+    e = put_pl_option(e, "class", plclass);
+  }
   strcpy(e, "])");
   e += strlen(e);
 
   appendArgList(&ploptions, "-f");
   appendArgList(&ploptions, "none");
   appendArgList(&ploptions, "-F");
-  appendArgList(&ploptions, "none");
+  appendArgList(&ploptions, plsysinit);
   appendArgList(&ploptions, "-g");
   appendArgList(&ploptions, "true");
   appendArgList(&ploptions, "-t");
   appendArgList(&ploptions, buf);
-
-  appendArgList(&tmpfiles, pltmp);		/* register for deletion */
+  appendArgList(&tmpfiles, pltmp);
 
   callprog(pl, &ploptions);
 }
@@ -1082,7 +1304,12 @@ createOutput()
       printf("\tchmod %03o %s\n", 0777 & ~mask, out);
 
     if ( !fake )
-    { if ( fchmod(ofd, 0777 & ~mask) != 0 )
+    {
+#ifdef HAVE_FCHMOD
+      if ( fchmod(ofd, 0777 & ~mask) != 0 )
+#else
+      if ( chmod(out, 0777 & ~mask) != 0 )
+#endif
       { fprintf(stderr, "Could make %s executable: %s\n", out, oserror());
 	error(1);
       }
@@ -1100,9 +1327,10 @@ removeTempFiles()
 { int n;
 
   for(n = 0; n < tmpfiles.size; n++)
-  { if ( verbose )
-      printf("\trm -f %s\n", tmpfiles.list[n]);
-    remove(tmpfiles.list[n]);
+  { if ( remove(tmpfiles.list[n]) == 0 )
+    { if ( verbose )
+	printf("\trm %s\n", tmpfiles.list[n]);
+    }
   }
 }
 
@@ -1138,18 +1366,36 @@ main(int argc, char **argv)
   putenv("PLLD=true");			/* for subprograms */
 
   verbose = FALSE;
+					/* behave as cpp */
+  if ( argc > 0 && streq(argv[0], "-E") )
+  { arglist cppoptions;
+
+    memset(&cppoptions, 0, sizeof(cppoptions));
+    for(argc--, argv++ ; argc > 0; argc--, argv++)
+      appendArgList(&cppoptions, argv[0]);
+
+    callprog(PROG_CPP, &cppoptions);
+
+    return 0;
+  }
+
   parseOptions(argc, argv);
   defaultProgram(&pl, PROG_PL);
   getPrologOptions();
   fillDefaultOptions();
 
   compileObjectFiles();
-  if ( !nolink )
-  { linkBaseExecutable();
 
-    if ( !nostate )
-    { createSavedState();
-      createOutput();
+  if ( !nolink )
+  { if ( shared )
+      linkSharedObject();
+    else
+    { linkBaseExecutable();
+      
+      if ( !nostate )
+      { createSavedState();
+	createOutput();
+      }
     }
   }
 
@@ -1157,3 +1403,5 @@ main(int argc, char **argv)
 
   return 0;
 }
+
+
